/*
 * debug.c - 
 */

/*
 * $Id: debug.c,v 1.4 2014/10/28 02:22:55 komiya Exp $
 *
 *   debugger
 *
 *   written by M.Yamamoto. (1990)
 *
 *
 */

static char *RCSid="$Id: debug.c,v 1.4 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * */
/*
 *    set-trace-function
 *
 */
/* * * * * * * * * * * * * * * * * * * * * * * * * */

void Sset_trace_function()
{
    void change_bcproc_to_tracer(int, int, int);
    void change_sysproc_to_tracer(int, int, int);
    SOBJ object_procedure;
    SOBJ prologue;
    SOBJ epilogue;
    SOBJ copied_procedure;
    SOBJ env;
#ifdef TOUCH_ARG
    object_procedure.obj = touch_obj(getarg(1));
    prologue.obj = touch_obj(getarg(2));
    epilogue.obj = touch_obj(getarg(3));
#else
    object_procedure.obj = getarg(1);
    prologue.obj = getarg(2);
    epilogue.obj = getarg(3);
#endif
    if (!FUNp(object_procedure.obj) || !FUNp(prologue.obj)
        || !FUNp(epilogue.obj)) {
        err_illegal_argument("All arguments must be procedures.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    if (GETtag(object_procedure.obj) == LTAG_BCLOSURE) {
        copied_procedure.obj = copy_procedure(object_procedure.obj);
        s_push(copied_procedure.obj);
        env.obj = mk_cons(copied_procedure.obj, LTAG_NIL);/*                 */
#ifdef TOUCH_ARG
        env.obj = mk_cons(touch_obj(getarg(2)), env.obj); /* make the environment */
        env.obj = mk_cons(touch_obj(getarg(3)), env.obj);    /*              */
        change_bcproc_to_tracer(touch_obj(getarg(1)), s_pop(), env.obj);
#else
        env.obj = mk_cons(getarg(2), env.obj);       /* make the environment */
        env.obj = mk_cons(getarg(3), env.obj);      /*              */
        change_bcproc_to_tracer(getarg(1), s_pop(), env.obj);
#endif
    } else if (GETtag(object_procedure.obj) == LTAG_SFUNCTION) {
        copied_procedure.obj = copy_procedure(object_procedure.obj);
        s_push(copied_procedure.obj);
        env.obj = mk_cons(copied_procedure.obj, LTAG_NIL);  /*                  */
#ifdef TOUCH_ARG
        env.obj = mk_cons(touch_obj(getarg(2)), env.obj);  /* make the environment */
        env.obj = mk_cons(touch_obj(getarg(3)), env.obj);     /*             */
        change_sysproc_to_tracer(touch_obj(getarg(1)), s_pop(), env.obj);
#else
        env.obj = mk_cons(getarg(2), env.obj);    /* make the environment */
        env.obj = mk_cons(getarg(3), env.obj);      /*              */
        change_sysproc_to_tracer(getarg(1), s_pop(), env.obj);
#endif
    } else {
        err_illegal_argument("1st argument must be a byte-code or a system procedure.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    set_reg(LTAG_T);
}


get_procedure_name(proc)
int proc;
{
    ENTITYP proc_entity;
    proc_entity = (ENTITYP)ADDRESS(proc);
    switch (GETtag(proc)) {
    case LTAG_BCLOSURE  :
    case LTAG_SFUNCTION :
        return(proc_entity->bcc.fun_name.obj);
    default :
        bokkan("system error: at get_procedure_name\n");
    }
}


copy_procedure(proc)
int proc;
{
    ENTITYP procedure;
    procedure = (ENTITYP)ADDRESS(proc);
    switch (GETtag(proc)) {
    case LTAG_BCLOSURE :
    case LTAG_SFUNCTION :
        return(mk_bcc(GETtag(proc),  procedure->bcc.const_vec.obj,
            procedure->bcc.fun_name.obj, procedure->bcc.source.obj,
            procedure->bcc.env.obj, procedure->bcc.body.obj,
            procedure->bcc.arg_info, procedure->bcc.start_addr ));
    default :
        bokkan("internal error: in copy_procedure (unexpected procedure)\n");
    }
}


void change_bcproc_to_tracer(target, copied_proc, env)
int target;
int copied_proc;
int env;
{
    ENTITYP target_proc;
    ENTITYP trace_proc;
    SYMBOL  * Pbc_trace_proc_sym;
    unsigned *aux_vec;

    target_proc = (ENTITYP)ADDRESS(target);
    Pbc_trace_proc_sym = (SYMBOL * )ADDRESS(bc_trace_proc_sym.obj);
    trace_proc = (ENTITYP)ADDRESS(Pbc_trace_proc_sym->sym_val.obj);

    target_proc->bcc.const_vec.obj = trace_proc->bcc.const_vec.obj;
    target_proc->bcc.fun_name.obj = get_procedure_name(copied_proc);
    target_proc->bcc.env.obj = env;
    target_proc->bcc.body.obj = trace_proc->bcc.body.obj;
    target_proc->bcc.arg_info = trace_proc->bcc.arg_info;
    target_proc->bcc.start_addr = trace_proc->bcc.start_addr;

    s_push(target);
    s_push(copied_proc);
    aux_vec = (unsigned *)Pmkvec_nf(2);
    *(aux_vec + 1) = SCOBJ_2;
    *(aux_vec + 2) = s_pop();           /* pop copied_proc */
    target_proc = (ENTITYP)ADDRESS(s_pop());    /* pop target      */
    target_proc->bcc.source.obj = LTAG_VECTOR | SSERDDA(aux_vec);
}


void change_sysproc_to_tracer(target, copied_proc, env)
int target, copied_proc, env;
{
    ENTITYP target_proc;
    unsigned *aux_vector;

    target_proc = (ENTITYP)ADDRESS(target);

    target_proc->bcc.fun_name.obj = get_procedure_name(copied_proc);
    target_proc->bcc.arg_info = 7;
    target_proc->bcc.body.obj = sysfun_tracer_body.obj;

    s_push(target);
    aux_vector = (unsigned *)make_bc_tracer(copied_proc, env);
    s_push(aux_vector);
    aux_vector = (unsigned *)Pmkvec_nf(2);
    *(aux_vector + 1) = SCOBJ_2;
    *(aux_vector + 2) = s_pop();            /* pop bc_tracer   */
    target_proc = (ENTITYP)ADDRESS(s_pop());    /* pop target      */
    target_proc->bcc.source.obj = LTAG_VECTOR | SSERDDA(aux_vector);
}


make_bc_tracer(copied_proc, env)
int copied_proc, env;
{
    ENTITYP bc_trace_proc;
    SOBJ trace_proc;
    unsigned *aux_vector;

    bc_trace_proc = (ENTITYP)ADDRESS(  ((SYMBOL * )ADDRESS(bc_trace_proc_sym.obj))->sym_val.obj  );
    s_push(copied_proc);
    trace_proc.obj = mk_bcc(LTAG_BCLOSURE,  bc_trace_proc->bcc.const_vec.obj,
        get_procedure_name(copied_proc), LTAG_NIL, env, bc_trace_proc->bcc.body.obj,
        bc_trace_proc->bcc.arg_info, bc_trace_proc->bcc.start_addr );
    s_push(trace_proc.obj);
    aux_vector = (unsigned *)Pmkvec_nf(2);
    *(aux_vector + 1) = SCOBJ_2;
    trace_proc.obj = s_pop();
    *(aux_vector + 2) = s_pop();  /* pop copied_proc */
    ((ENTITYP)ADDRESS(trace_proc.obj))->bcc.source.obj
	= LTAG_VECTOR | SSERDDA(aux_vector);
    return(trace_proc.obj);
}


void Ssysfun_tracer()
{
    ENTITYP proc;
    unsigned *aux_vector;
    proc = (ENTITYP)ADDRESS(getarg(0));
    aux_vector = (unsigned *)ADDRESS(proc->bcc.source.obj);
    setsv(-1, *(aux_vector + 2));
    set_up_stack();
}


/* * * * * * * * * * * * * * * * * * * * * * * * * */
/*
 *      detouch-trace
 *
 */
/* * * * * * * * * * * * * * * * * * * * * * * * * */

void Sdetouch_trace_function()
{
    void restore_bc_proc(int, int);
    SOBJ proc;
    ENTITYP proc_entity;
    ENTITYP proc_entity2;
    unsigned *aux_vector;
    unsigned *aux_vector2;

#ifdef TOUCH_ARG
    proc.obj = touch_obj(getarg(1));
#else
    proc.obj = getarg(1);
#endif
    if ( !FUNp(proc.obj) ){
        err_illegal_argument("Argument must be a procedure.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    if (GETtag(proc.obj) == LTAG_BCLOSURE) {
        proc_entity = (ENTITYP)ADDRESS(proc.obj);
        if (proc_entity->bcc.source.obj == LTAG_NIL)
            set_reg(LTAG_FALSE);
        else if (GETtag(proc_entity->bcc.source.obj) == LTAG_VECTOR) {
            aux_vector = (unsigned *)ADDRESS(proc_entity->bcc.source.obj);
            if (*(aux_vector + 1)  == SCOBJ_2) {
                restore_bc_proc(proc.obj, *(aux_vector + 2));
                set_reg(LTAG_T);
            } else
                set_reg(LTAG_FALSE);
        } else
            bokkan("system error: at Sdetouch_trace(byte-code)\n");
    } else if ( GETtag(proc.obj) == LTAG_SFUNCTION ) {
        proc_entity = (ENTITYP)ADDRESS(proc.obj);
        if (proc_entity->bcc.source.obj == LTAG_NIL)
            set_reg(LTAG_FALSE);
        else if ( GETtag(proc_entity->bcc.source.obj) == LTAG_VECTOR) {
            aux_vector = (unsigned *)ADDRESS(proc_entity->bcc.source.obj);
            if (*(aux_vector + 1) == SCOBJ_2) {
                proc_entity2 = (ENTITYP)ADDRESS(*(aux_vector + 2));
                aux_vector2 =(unsigned *)ADDRESS(proc_entity2->bcc.source.obj);
                restore_bc_proc(proc.obj, *(aux_vector2 + 2));
                set_reg(LTAG_T);
            } else
                set_reg(LTAG_FALSE);
        } else
            bokkan("system error: at Sdetouch_trace(system-func)\n");
    } else
        set_reg(LTAG_FALSE);
}


void restore_bc_proc(proc, copied_proc)
int proc, copied_proc;
{
    ENTITYP entity;
    ENTITYP copied_entity;

    entity = (ENTITYP)ADDRESS(proc);
    copied_entity = (ENTITYP)ADDRESS(copied_proc);

    entity->bcc.const_vec.obj = copied_entity->bcc.const_vec.obj;
    entity->bcc.fun_name.obj = copied_entity->bcc.fun_name.obj;
    entity->bcc.source.obj = copied_entity->bcc.source.obj;
    entity->bcc.env.obj = copied_entity->bcc.env.obj;
    entity->bcc.body.obj = copied_entity->bcc.body.obj;
    entity->bcc.arg_info = copied_entity->bcc.arg_info;
    entity->bcc.start_addr = copied_entity->bcc.start_addr;
}


/*
restore_sfn_proc(proc, copied_proc)
int proc, copied_proc;
{
  ENTITYP entity;
  ENTITYP copied_entity;

  entity = (ENTITYP)ADDRESS(proc);
  copied_entity = (ENTITYP)ADDRESS(copied_proc);

  entity->sfn.const_vec.obj = copied_entity->sfn.const_vec.obj;
  entity->sfn.fun_name.obj = copied_entity->sfn.fun_name.obj;
  entity->sfn.arg_info = copied_entity->sfn.arg_info;
  entity->sfn.entry_addr = copied_entity->sfn.entry_addr;
}
*/


/* * * * * * * * * * * * * * * * * * * * * * * * * */
/*
 *   back-trace
 *
 */
/* * * * * * * * * * * * * * * * * * * * * * * * * */

#define  INFINITE_DEPTH     -1
#define  NO_EXIST_AFUN      -2
#define  REACH_END      -3

void back_trace()
{
    int depth;
    SOBJ sc_max_depth;
    int max_depth;

    print_con("\nBacktrace: "); /* Modefied by N.Watanabe Apr.9, 1994 */
    sc_max_depth.obj = get_sval(max_back_trace_depth_sym.obj);
    if (GETfix_tag(sc_max_depth.obj) == LTAG_FIXNUM)
        max_depth = fix_to_int(sc_max_depth.obj);
    else
        max_depth = INFINITE_DEPTH;
    depth = real_stack_back_trace(max_depth);
    if (depth > 0) {
        depth = heap_stack_back_trace(more.obj, depth, OFF);
    } else if (depth == NO_EXIST_AFUN) {
        depth = heap_stack_back_trace(more.obj, max_depth, ON);
    } else if (depth == INFINITE_DEPTH) {
        depth = heap_stack_back_trace(more.obj, INFINITE_DEPTH, OFF);
    }
    if (depth == 0) {
        print_con(" ...");
    }
    print_con("\n");
}

void print_proc_name(unsigned);

real_stack_back_trace(depth)
int depth;
{
    unsigned *active;
    int first_flag;
    SOBJ proc;

    first_flag = ON;
    active = active_val();
    while ( ( (depth > 0) || (depth < 0) ) &&  (active != 0 )   ) {
        if (depth > 0)
            depth--;
        proc.obj = *(active - 1);
        if (first_flag == ON) {
            print_proc_name(proc.obj);
            first_flag = OFF;
        } else {
            print_con(" < ");
            print_proc_name(proc.obj);
        }
        active = (unsigned *)(*(active + 1));
    }
    if (first_flag == ON)
        return(NO_EXIST_AFUN);
    else
        return(depth);
}


heap_stack_back_trace(cont, depth, first_flag)
unsigned cont;
int depth;
int first_flag;
{
    SOBJ proc;

    while ( ( (depth > 0) || (depth < 0) ) &&  ( cont != LTAG_NIL )) {
        if (depth > 0)
            depth--;
#ifdef CALLIOC
        proc.obj = *(((int *)ADDRESS(CARobj(CARobj(CARobj(cont))))) + 1);
#else
        proc.obj = *(((int *)ADDRESS(CARobj(CARobj(cont)))) + 1);
#endif
        if (first_flag == ON) {
            print_proc_name(proc.obj);
            first_flag = OFF;
        } else {
            print_con(" < ");
            print_proc_name(proc.obj);
        }
        cont = CDRobj(cont);
    }
    if (first_flag == ON)
        return(NO_EXIST_AFUN);
    else {
        if (cont == LTAG_NIL)
            return(REACH_END);
        else
            return(depth);
    }
}


void print_proc_name(proc)
unsigned proc;
{
    ENTITYP proc_entity;
    SOBJ fname;

    proc_entity = (ENTITYP)ADDRESS(proc);

    switch (GETtag(proc)) {
    case LTAG_BCLOSURE :
    case LTAG_MBCLOSURE :
    case LTAG_SFUNCTION :
    case LTAG_MSFUNCTION :
        fname.obj = proc_entity->bcc.fun_name.obj;
        if (fname.obj != LTAG_NIL)
            printobj_con(fname.obj, PRINC, 0);
        else
            print_con("#<function>");
        break;
    case LTAG_NCLOSURE :
    case LTAG_MNCLOSURE :
        print_con("#<native-code-closure>(not implemented yet ???)");
        break;
    default :
        print_con("<not-procedure>");
    }
}


void Sget_proc_name()
{
    ENTITYP proc_entity;
    SOBJ proc;

#ifdef TOUCH_ARG
    proc.obj = touch_obj(getarg(1));
#else
    proc.obj = getarg(1);
#endif
    if ( !FUNp(proc.obj) ) {
        err_illegal_argument("Argument must be a procedure.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    if (GETtag(proc.obj) == LTAG_CONTINUATION) {
        err_illegal_argument("Argument must be a procedure, other than continuation.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    proc_entity = (ENTITYP)ADDRESS(proc.obj);
    switch (GETtag(proc.obj)) {
    case LTAG_BCLOSURE :
    case LTAG_MBCLOSURE :
    case LTAG_SFUNCTION :
    case LTAG_MSFUNCTION :
        set_reg(proc_entity->bcc.fun_name.obj);
        break;
    default :
        bokkan("system error : in Sget_proc_name.");
    }
}


void Scopy_procedure()
{
    SOBJ proc;

#ifdef TOUCH_ARG
    proc.obj = touch_obj(getarg(1));
#else
    proc.obj = getarg(1);
#endif
    if ( !FUNp(proc.obj) ) {
        err_illegal_argument("Argument must be a procedure.");
	return;
    }
    if (GETtag(proc.obj) == LTAG_CONTINUATION) {
        err_illegal_argument("Argument must be a procedure, other than continuation.");
	return;
    }

    switch (GETtag(proc.obj)) {
    case LTAG_BCLOSURE :
    case LTAG_MBCLOSURE :
    case LTAG_SFUNCTION :
    case LTAG_MSFUNCTION :
        set_reg(copy_procedure(proc.obj));	    
        break;
    default :
        bokkan("system error : in Scopy_procedure.");
    }
}
