/*
 * ltc.c - LTC(Lazy Task Creation) scheduling
 */
/*
 * $Id: ltc.c,v 1.9 1999/10/04 08:25:02 komiya Exp $
 *
 * Created:  Masahiko Takahashi         (Nov 5, 1996)
 * Modified: shige@kuis.kyoto-u.ac.jp   (Aug 6, 1997)
 *
 */

static char *RCSid="$Id: ltc.c,v 1.9 1999/10/04 08:25:02 komiya Exp $";

#undef FOO

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

#ifdef SUNSOL2
#include <unistd.h>
#endif


#include "scheme.h"

#ifdef THREAD

#ifdef LTC

#include "ltc.h"

/* `extern'ed functions and variables */
extern int iretaddrx ();
extern int iretaddrl ();
extern int xreturn ();
extern int lreturn ();
#if defined(PTHREAD) && defined(sparc) && defined(USE_REG)
extern pthread_attr_t pattr;
#endif

/* prototype declarations */
void init_ltc_scheduler();
void threaddata_alloc(thread_data_t **, int);
void ltcstack_alloc();
void ltc_scheduler(int);
int  get_thread_number(thread_data_t **);
void primary_steal();
int  steal_lt(thread_data_t **);
void rewrite_retaddr();
void stret();
void lfretx();
void lfretl();
void open_ltframe();
void open_lltframe();
void open_lltframe_1();
#ifdef CONDITIONAL_FUTURE
void waitthr_scheduler(int);
void secondary_steal();
void condstack_alloc(thread_data_t **, int);
int switch_to_context(thread_data_t *); /* new */
void switch_cond_context(thread_data_t **);
void switch_main_context(thread_data_t **);
void put_waitthr(waitthr_list_t *, thread_data_t **);
int  get_waitthr(waitthr_list_t *, thread_data_t **);
void cstack_exec_scheduler(int); /* new */
void open_ltframe_c();
void open_ltframe_c_aux();
void open_condframe(thread_data_t **, int);
void iretaddrl_cond_primary();
void iretaddrl_cond_secondary();
unsigned int inherit_result(thread_data_t *thr);
#endif
#ifdef LTC_DEBUG
void stack_dumper ();
#endif

/*
 * variables
 */

int processors;
int sprocessors;
thread_data_t **pthrs; /* primary thread list */
thread_data_t **sthrs; /* secondary thread list */

#ifdef CONDITIONAL_FUTURE
/* unsigned int *thread_counter; /* context changing counter */
waitthr_list_t *twlp;
waitthr_list_t *fwlp;
/* queue of cstacks (thrs) to be checked */
act_cstack_queue_t act_cstack_q;
thread_data_t *current_cond_thr;
#endif

/* mutex lock variables */
int rotator = -1;
int pthr_count = 0;
#ifdef CONDITIONAL_FUTURE
int sthr_count = 0;
#endif
condition_t thr_count_threshold;
mutex_t thr_count_lock;


#ifdef CONDITIONAL_FUTURE

/*
 * add_child() adds a child to COND_CHILD_LIST
 */
void add_child(unsigned int **ccstack, unsigned int **pcstack)
{
    chldlst_t *cl;

    if ((cl = (chldlst_t *)malloc(sizeof(chldlst_t))) == NULL)
        bokkan("memory allocation error at add_child():"
               "Cannot allocate memory.");

    cl->stk = (unsigned int *)ccstack;
    cl->next = (chldlst_t *)pcstack[COND_CHILD_LIST];
    pcstack[COND_CHILD_LIST] = (unsigned int *)cl;
}

/*
 * rm_child() removes a child from COND_CHILD_LIST
 */
void rm_child(unsigned int **ccstack, unsigned int **pcstack)
{
    chldlst_t *cl, *ncl, *pcl;

    cl = (chldlst_t *)(pcstack[COND_CHILD_LIST]);
    if (cl == NULL)
        return;
    else if (cl->stk == (unsigned int *)ccstack) {
        ncl = cl->next;
        free(cl);
        (chldlst_t *)(pcstack[COND_CHILD_LIST]) = ncl;
        return;
    } else
        for (;;) {
            pcl = cl;
            cl = cl->next;
            if (cl == NULL)
                return;
            else if (cl->stk == (unsigned int *)ccstack) {
                ncl = cl->next;
                free(cl);
                pcl->next = ncl;
                return;
            }
        }
}

/*
 * set_chldlst() changes COND_PARENT of each child's stack in COND_CHILD_LIST
 * to an object
 */
void set_chldlst(unsigned int **pcstack, unsigned int obj)
{
    chldlst_t *cl;

    cl = (chldlst_t *)(pcstack[COND_CHILD_LIST]);
    while (cl != NULL) {
        (cl->stk)[COND_PARENT] = obj;
        cl = cl->next;
    }
    return;
}

void set_chldlst_if_not(unsigned int **pcstack,
                        unsigned int **stk, unsigned int obj2)
{
    chldlst_t *cl;

    cl = (chldlst_t *)(pcstack[COND_CHILD_LIST]);
    while (cl != NULL) {
        if (((cl->stk)[COND_PARENT] != obj2) && ((cl->stk)[COND_PARENT] != NULL)) {
            (cl->stk)[COND_PARENT] = stk;
            if (stk != (unsigned int **)LTAG_T)
                add_child(cl->stk, stk);
        }
        cl = cl->next;
    }
    return;
}

/*
 * set_allchldlst() changes COND_RESULT of all child's stack in COND_CHILD_LIST
 * to an object
 */
void set_allchldlst(unsigned int **pcstack, unsigned int obj)
{
    chldlst_t *cl;

    if (pcstack == NULL)
        return;
    cl = (chldlst_t *)(pcstack[COND_CHILD_LIST]);
    while (cl != NULL) {
        (cl->stk)[COND_RESULT] = obj;
        set_allchldlst((unsigned int **)cl->stk, obj);
        cl = cl->next;
    }
    return;
}

/*
 * free_chldlst() frees COND_CHILD_LIST
 */
void free_chldlst(unsigned int **cstack)
{
    chldlst_t *cl, *ncl;

    cl = (chldlst_t *)(cstack[COND_CHILD_LIST]);
    while (cl != NULL) {
        ncl = cl->next;
        free(cl);
        cl = ncl;
    }
    cstack[COND_CHILD_LIST] = NULL;
    return;
}

mutex_t act_cstack_lock;

int find_act_cstack_q(thread_data_t *thr)
{
    cstack_lst_t *cl;

    cl = act_cstack_q.tail;
    while (cl != NULL && cl->thr != thr)
        cl = (cstack_lst_t *)(cl->next);
    if (cl == NULL)
        return 0;
    else
        return 1;
}

/*
 * activate cstack of thr
 */
void enq_act_cstack_q(thread_data_t *thr)
{
    cstack_lst_t *cl;

    if ((cl = (cstack_lst_t *)malloc(sizeof(cstack_lst_t))) == NULL)
        bokkan("memory allocation error at enq_act_cstackq():"
               "Cannot allocate memory.");

    mutex_lock(act_cstack_lock);
    cl->thr = thr;
    cl->next = NULL;
    if (act_cstack_q.tail == NULL)
        act_cstack_q.head = cl;
    else
        act_cstack_q.tail->next = cl;
    act_cstack_q.tail = cl;
#ifdef FOO
    fprintf(stderr, "cstack enqueued.\n");
#endif
    mutex_unlock(act_cstack_lock);
    return;
}

thread_data_t *deq_act_cstack_q()
{
    cstack_lst_t *cl;

    if (act_cstack_q.head == NULL)
        return (thread_data_t *)NULL;
    else {
        thread_data_t *thr;

        mutex_lock(act_cstack_lock);
        if (act_cstack_q.head == NULL)
          return (thread_data_t *)NULL;
        cl = act_cstack_q.head;
        if (cl->next == NULL)
            act_cstack_q.head = act_cstack_q.tail = NULL;
        else
            act_cstack_q.head = cl->next;
#ifdef FOO
        fprintf(stderr, "cstack dequeued.\n");
#endif
        mutex_unlock(act_cstack_lock);
        thr = cl->thr;
        free(cl);
        return thr;
    }
}

int is_empty_act_cstack_q()
{
    if (act_cstack_q.head == NULL)
        return 1;
    else
        return 0;
}

#endif

/*
 * initializing lazy task scheduling
 */
void init_ltc_scheduler()
{
    int i, j;
    char *pp;

    if ((pp = getenv("PROS")) != NULL) {
        processors = atoi(pp);
    } else {
#ifdef SUNSOL2
        /*
         * get current processor numbers
         */
        processors = (int)sysconf(_SC_NPROCESSORS_ONLN);
#else
        processors = 2;
#endif
    }
    
#ifdef SUNSOL2
    /*
     * setting concurrency level to the number of processors
     */
    thr_setconcurrency(processors);
#endif
    
    thr_count_threshold = (condition_t)condition_alloc();
    thr_count_lock = (mutex_t)mutex_alloc();

    /*
     * allocating primary threads' data (the number of processors)
     */
    if ((pthrs = (thread_data_t **)malloc(processors * sizeof(thread_data_t *))) == NULL)
        bokkan("memory allocation error at init_ltc_scheduler(): "
               "Can't allocate memory.");
#ifdef CONDITIONAL_FUTURE
    /*
     * allocation secondary threads' data (the number of processors)
     */
    if ((pp = getenv ("SPROS")) != NULL) {
        sprocessors = atoi(pp);
    } else {
        sprocessors = processors;
    }
    if ((sthrs = (thread_data_t **)malloc(sprocessors * sizeof(thread_data_t *))) == NULL)
        bokkan("memory allocation error at init_ltc_scheduler(): "
               "Can't allocate memory.");

    /* initialize act_cstack_q */
    act_cstack_q.head = NULL;
    act_cstack_q.tail = NULL;
    act_cstack_lock = (mutex_t)mutex_alloc();
#endif

#ifdef CONDITIONAL_FUTURE
    /*
     * allocating waiting threads list
     */
    if ((twlp = (waitthr_list_t *)malloc(sizeof(waitthr_list_t))) == NULL)
        bokkan("memory allocation error at init_ltc_scheduler(): "
               "Can't allocate memory.");
    twlp->mutex = (mutex_t)mutex_alloc();
    twlp->firstp = NULL;
    twlp->lastp = NULL;
    if ((fwlp = (waitthr_list_t *)malloc(sizeof(waitthr_list_t))) == NULL)
        bokkan("memory allocation error at init_ltc_scheduler(): "
               "Can't allocate memory.");
    fwlp->mutex = (mutex_t)mutex_alloc();
    twlp->firstp = NULL;
    fwlp->firstp = NULL;
    fwlp->lastp = NULL;
#endif
    
    /*
     * preparing primary threads
     */
    printf("\n");
    for (i = 1; i < processors; i++) {
        j = (int)cthread_fork(ltc_scheduler, i);
        printf("preparing PROC[%d]: #%x\n", i, (unsigned int) j );
    }

    /*
     * allocate and resister thread's data
     * --- by T.Komiya and shige
     */
    if ((pthrs[0] = (thread_data_t *)malloc(sizeof(thread_data_t))) == NULL)
        bokkan("memory allocation error at init_ltc_scheduler(): "
               "Can't allocate memory.");
    pthrs[0]->mutex = (mutex_t)mutex_alloc();
    pthrs[0]->stack = (unsigned int **)stack_top; /* pros' stack list */
    pthrs[0]->thr   = cthread_self();             /* pros' threads list */
#ifdef CONDITIONAL_FUTURE
    pthrs[0]->action = MODE_MAIN;
    pthrs[0]->mstack = pthrs[0]->stack;
# if defined(sparc) && defined(USE_REG)
    pthrs[0]->regs = main_thread_regs;
# else
    pthrs[0]->regs = (REGS *)cthread_data(cthread_self()); 
# endif
    /* allocate condition stack */
    condstack_alloc(pthrs, 0);
#endif
    
    /*
     * barrier for threads synchronization
     */
    while (pthr_count < (processors-1))
        sleep(1);

    mutex_lock(thr_count_lock);
    pthr_count++;
    condition_broadcast(thr_count_threshold);
    mutex_unlock(thr_count_lock);
    
#ifdef CONDITIONAL_FUTURE
    /*
     * preparing secondary threads
     */
    if (processors == 1)
        sprocessors = 0;

    printf("\n");
    for (i = 0; i < sprocessors; i++) {
        j = (int)mtus_thread_fork_secondary(waitthr_scheduler, i);
        printf("preparing secondary PROC[%d]: #%x\n", i, (unsigned int)j);
    }

    j = (int)cthread_fork(cstack_exec_scheduler, NULL);
    printf("preparing CSTACK_EXEC_SCHEDULER PROC: #%x\n", (unsigned int)j);

    /*
     * barrier for threads synchronization
     */
    while (sthr_count < (sprocessors+1))
        sleep(1);
    mutex_lock(thr_count_lock);
    condition_broadcast(thr_count_threshold);
    mutex_unlock(thr_count_lock);
#endif

    return;
}


/*
 * allocating and setting thread data
 */
void threaddata_alloc(thread_data_t **thrs, int thr_num)
{
    REGS *regsp;
    mutex_t m;
    thread_data_t *thrd;

#if defined(sparc) && defined(USE_REG)
    if ((main_thread_regs = (REGS *)malloc(sizeof(REGS))) == NULL)
        bokkan("memory allocation error at threaddata_alloc(): "
               "Can't allocate memory.");
#else
    if ((regsp = (REGS *)malloc(sizeof(REGS))) == NULL)
        bokkan("memory allocation error at threaddata_alloc(): "
               "Can't allocate memory.");
    cthread_set_data(cthread_self(), (any_t)regsp);
#endif

    /* allocating stack */
    ltcstack_alloc(thrs);

    /*
     * register thread's id and stack
     */
    if ((thrd = (thread_data_t *)malloc(sizeof(thread_data_t))) == NULL)
        bokkan("memory allocation error at threaddata_alloc(): "
               "Can't allocate memory.");
#if defined(CONDITIONAL_FUTURE) && defined(DEBUG_MESG)
    if (thrs == pthrs)
        fprintf(stderr, "new p%d (thr=#%x) is created.\n", thr_num, thrs[thr_num]);
    else
        fprintf(stderr, "new s%d (thr=#%x) is created.\n", thr_num, thrs[thr_num]);
#endif
    m = (mutex_t)mutex_alloc();
    mutex_lock(m);
    thrd->mutex = m;
    thrs[thr_num] = thrd;
    thrs[thr_num]->stack = (unsigned int **)stack_top; /* pros' stack list */
    thrs[thr_num]->thr   = cthread_self();           /* pros' threads list */
#ifdef CONDITIONAL_FUTURE
    thrs[thr_num]->action = MODE_MAIN;
    thrs[thr_num]->mstack = thrs[thr_num]->stack;
# if defined(sparc) && defined(USE_REG)
    thrs[thr_num]->regs = main_thread_regs;
# else
    thrs[thr_num]->regs = regsp;
# endif
    condstack_alloc(thrs, thr_num);
#endif
    mutex_unlock(m);
    return;
}

/*
 * allocating and setting stack for LTC(Lazy Task Creation)
 */
void ltcstack_alloc(thread_data_t **thrs)
{

    /*
     */
    if ((stack_top = (char *)malloc(CHILD_STACK_SIZE)) == NULL)
        bokkan("memory allocation error at ltcstack_alloc(): "
               "Can't allocate memory.");
    stack_bottom = stack_top + CHILD_STACK_SIZE - 28;
    stack_top = stack_top + CHILD_GETA_OF_STACK;

    *(((unsigned int **)stack_bottom) + 1) = (unsigned int *)LTAG_NIL;
    *(((unsigned int **)stack_bottom) + 2) = 0;
    *(((unsigned int **)stack_bottom) + 3) = 0;

    stack_area_bottom = stack_bottom;

    /* set sp_reg */
    set_top(stack_bottom);

    if (thrs == pthrs)
        s_push(primary_steal);
#ifdef CONDITIONAL_FUTURE
    else if (thrs == sthrs)
        s_push(secondary_steal);
#endif
    else
        bokkan("fatal error at ltcstack_alloc(): Unknown thread data list.\n");
    s_push(0);
    s_push(0);

    /* set base & active */
    set_base(top_val());
    set_active(top_val());
    
    /*
     * initialize stack top area
     */
    ((unsigned int **)stack_top)[STACK_LOCK] = (unsigned int *)mutex_alloc();
    ((unsigned int **)stack_top)[STACK_BASE] = (unsigned int *)sp_reg;
    ((unsigned int **)stack_top)[FUTURE_PH] = (unsigned int *)LTAG_NIL;
    ((unsigned int **)stack_top)[STACK_INFO_END] = 0;
    ((unsigned int **)stack_top)[LTQ_HEAD]
        = ((unsigned int **)stack_top)[LTQ_TAIL]
        = (unsigned int *)((unsigned int **)stack_top + (STACK_INFO_END + 1));

    return;
}


/*
 * lazy task scheduling (primary processors' scheduler)
 */
void ltc_scheduler(thr_num)
    int thr_num;
{
#ifdef CONDITIONAL_FUTURE
    unsigned int **cstk;
#endif
    int counter = 0;
    int interval;
    char *env;
    int policy;
    struct sched_param param;

    if ((env = getenv ( "PRI_BC_INTERVAL" )) != NULL) {
        interval = atoi(env);
    } else {
        interval = PRI_BC_INTERVAL;
    }

    /* allocating and setting thread data */
    threaddata_alloc(pthrs, thr_num);

    mutex_lock(thr_count_lock);
    pthr_count++;
    condition_wait(thr_count_threshold, thr_count_lock);
    mutex_unlock(thr_count_lock);

    /*
     * steal task from other processor's ltc stack
     */
    primary_steal();

    /*
     * interpret byte code
     */
    while (1) {
#ifdef CONDITIONAL_FUTURE
        /*
         * executing main
         */

        while (1) {
            ((int(*)())(*( bcbody_reg + bcpc_reg )))();
            counter++;

            if (counter > interval) {
                counter = 0;
                break;
            }
        }

        /*
         * executing condition
         */
        cstk = pthrs[thr_num]->cstack;
        if (cstk[COND_RESULT] == NULL) {
            mutex_t m = (mutex_t)(pthrs[thr_num]->mutex);
            mutex_lock(m);
            pthrs[thr_num]->condition = INVALID;
            pthrs[thr_num]->cstack[COND_PARENT] = (unsigned int *)LTAG_T;
            pthrs[thr_num]->cstack[COND_CHILD_LIST] = 0;
#ifdef DEBUG_MESG
            fprintf(stderr, "task canceled.\n");
#endif
            mutex_unlock(m);
            primary_steal();
            cstk[COND_RESULT] = (unsigned *)LTAG_T;
        } else if (pthrs[thr_num]->condition == VALID) {
            if ((cstk[COND_PARENT] == (unsigned *)LTAG_T)
                && (cstk[COND_BCBODY] != (unsigned int *)LTAG_NIL)
                && (! find_act_cstack_q(pthrs[thr_num]))) {
                enq_act_cstack_q(pthrs[thr_num]);
/*                goto gogo;    *//* *** */
            } else {
            gogo:
                while (cstk[COND_PARENT] != (unsigned int *)LTAG_T)
                    cstk = (unsigned int **)cstk[COND_PARENT];
                if (cstk[COND_RESULT] == (unsigned int *)LTAG_FALSE) {
                    mutex_t m = (mutex_t)(pthrs[thr_num]->mutex);
#ifdef DEBUG_MESG
                    fprintf(stderr, "result = FALSE\n");
#endif
                    mutex_lock(m);
                    pthrs[thr_num]->cstack[COND_RESULT] = (unsigned int *)LTAG_FALSE;
                    if (*((unsigned int *)stack_bottom - 1) != (unsigned int)primary_steal)
                        fprintf(stderr, "???????\n");
                    *((unsigned int *)stack_bottom - 1)
                        = (unsigned int)secondary_steal;
                    if (pthrs[thr_num]->cstack[COND_BCBODY] != LTAG_NIL) {
                        /* the task is not ordinary task in conditional task
                           group, but conditional task */
                        if (((unsigned int **)(pthrs[thr_num]->cregs->_stack_top))[COND_SPREG][2] != (unsigned int)iretaddrl_cond_primary)
                            fprintf(stderr, "????**\n");
                        ((unsigned int **)(pthrs[thr_num]->cregs->_stack_top))[COND_SPREG][2] = (unsigned int)iretaddrl_cond_secondary;
                    }

                    put_waitthr(fwlp, pthrs);
                    threaddata_alloc(pthrs, thr_num);
                    mutex_unlock(m);
                    primary_steal();
                }
            }
        }
#else
        ((int(*)())(*( bcbody_reg + bcpc_reg )))();
#endif
    }

    
    /* not reach! */
    return;
}

/*
 * get thread number
 */
int get_thread_number(thread_data_t **thrs)
{
    int i;
    int pros;
    mtus_thread_t thread_id;

    /* get my thread id */
    thread_id = mtus_thread_self();

    if (thrs == pthrs)
        pros = processors;
#ifdef CONDITIONAL_FUTURE
    else if (thrs == sthrs)
        pros = sprocessors;
#endif
    /* compare thread id */
    for(i = 0 ; i < pros ; i++) {
        if (thrs[i] != NULL && thread_id == (thrs[i]->thr))
            return i;
    }

    /*
     * paranoia check (threaddata list check)
     */
    /* compare thread id */
    for(i = 0 ; i < processors ; i++) {
        if (thread_id == (pthrs[i]->thr))
            return i;
    }
#ifdef CONDITIONAL_FUTURE
    for(i = 0 ; i < sprocessors ; i++) {
        if (thread_id == (sthrs[i]->thr))
            return i;
    }
#endif

    /*
     * paranoia check (stack_top check)
     */
#ifdef CONDITIONAL_FUTURE
    /* first, check main stack */
    for(i = 0 ; i < processors ; i++)
        if (thrs[i]->mstack == (unsigned int **)stack_top)
            return i;
    /* second, check condition stack */
    for(i = 0 ; i < processors ; i++)
        if (thrs[i]->cstack == (unsigned int **)stack_top)
            return i;
#else
    /* check stack */
    for(i = 0 ; i < processors ; i++)
        if (thrs[i]->stack == (unsigned int **) stack_top)
            return i;
#endif
    
    return -1;
}

/*
 * steal function for primary processors
 *   This is a wrapper function.
 *   --- shige
 */
void primary_steal()
{
    int thr_num;
    int result;
    struct timespec t, r;
#ifdef CONDITIONAL_FUTURE
    extern struct sched_param pschedparam;
#endif

    t.tv_sec = 0;
    t.tv_nsec = 100000000;

    if((thr_num = get_thread_number(pthrs)) < 0)
        bokkan("fatal error at primary_steal(): "
               "this thread is an unknown thread.\n");

    result = 1; /* default: steal action failed */


/*
#ifdef CONDITIONAL_FUTURE
    if (pthrs[thr_num]->condition == VALID) {
        pthrs[thr_num]->condition = INVALID;
        set_chldlst(pthrs[thr_num]->cstack, NULL);
        free_chldlst(pthrs[thr_num]->cstack);
    }
#endif
*/

#ifdef CONDITIONAL_FUTURE
    /*
     * up my thread priority
     */
#if 1
    if (pthread_setschedparam(mtus_thread_self(), SCHED_OTHER, &pschedparam))
        fprintf(stderr, "Cannot chage the priority at primary_steal()\n");
#endif
#endif

    while(result != 0) {
#ifdef DEBUG_CFUTURE
        fprintf(stderr, "DEBUG [ppro%d]: primary_steal()\n", thr_num);
#endif
    
#ifdef CONDITIONAL_FUTURE
        result = get_waitthr(twlp, pthrs);
#endif
        if (result != 0)
            result = steal_lt(pthrs);
        else
            break;

        /* if steal action successed */
        if (result == 0)
            break;

        /* if cannot get task, wait for a while */
        nanosleep(&t, &r);
    }

    return;
}

/*
 * steal tasks from other processor's ltc
 *   --- T.Komiya
 */
int steal_lt(thread_data_t **thrs)
{
    int i, j, target_pro_no;
    int pros;
    int pro_no;
    unsigned int **target_top;
    unsigned int target_sp, target_act, target_base;
    unsigned int target_cact, target_csp;
    unsigned int *base, *act;
#ifdef CONDITIONAL_FUTURE
    unsigned int *cond_bcbody, cond_bcpc;
    mutex_t m;
#endif
    FUTURE *fut;
    ENTITYP proc_entity;

#if 0
    static int n = 0;
#endif

    if((pro_no = get_thread_number(thrs)) < 0)
        bokkan("fatal error at steal_lt(): "
               "this thread is an unknown thread.\n");

#if 0
    if (n++ > 100) {
        n = 0;
        printf("* %d", pthr_count);
    }
#endif
    if (thrs == pthrs)
        pros = processors;
#ifdef CONDITIONAL_FUTURE
    else if (thrs == sthrs)
        pros = sprocessors;
#endif

    for (i = 1; i < pros; i++) {
        target_pro_no = (pro_no + i) % pros;
#ifdef CONDITIONAL_FUTURE
        m = thrs[target_pro_no]->mutex;
        mutex_lock(m);          /* thrs[] may be enqueued */
        if (thrs[target_pro_no]->cstack[COND_RESULT] == NULL) {
            mutex_unlock(m);
            continue;
        }
/*
        {
            unsigned *obj;
            unsigned int **cstk;

            if (thrs == pthrs)
                obj = (unsigned int *)LTAG_FALSE;
            else
                obj = (unsigned int *)LTAG_T;
            cstk = thrs[target_pro_no]->cstack;
            while (cstk[COND_PARENT] != (unsigned int *)LTAG_T)
                cstk = (unsigned int **)cstk[COND_PARENT];
            if (cstk[COND_RESULT] == obj) {
                mutex_unlock(m);
                continue;
            }
        }
 */
        mutex_unlock(m);
#endif
        target_top = thrs[target_pro_no]->stack;
        if (target_top[LTQ_TAIL] > target_top[LTQ_HEAD]) {
#ifdef CONDITIONAL_FUTURE
            mutex_lock((mutex_t)(thrs[target_pro_no]->mutex));
#endif
            mutex_lock((mutex_t)((thrs[target_pro_no]->stack)[STACK_LOCK]));

            target_top = thrs[target_pro_no]->stack;
            if (!(target_top[LTQ_TAIL] > target_top[LTQ_HEAD])) {
                mutex_unlock((mutex_t)target_top[STACK_LOCK]);

#ifdef CONDITIONAL_FUTURE
                mutex_unlock((mutex_t)(thrs[target_pro_no]->mutex));
#endif

                continue;
            }

            /* move LTQ head */
            target_top[LTQ_HEAD] += LTINFO_SIZE;

#ifdef RMO
# if defined(sparc)
            asm("membar #StoreLoad"); /* for V9 */
# else
            mutex_lock(dummy);
            mutex_unlock(dummy);
# endif
#endif

            if (!(target_top[LTQ_TAIL] >= target_top[LTQ_HEAD])) {
                target_top[LTQ_HEAD] -= LTINFO_SIZE;
#ifdef RMO
# if defined(sparc)
                asm("membar #StoreLoad"); /* for V9 */
# else
                mutex_lock(dummy);
                mutex_unlock(dummy);
# endif
#endif
                mutex_unlock((mutex_t)target_top[STACK_LOCK]);

#ifdef CONDITIONAL_FUTURE
                mutex_unlock((mutex_t)(thrs[target_pro_no]->mutex));
#endif

                continue;
            }

            /* retrieve previous lazy task info */
            target_base = (unsigned int)target_top[STACK_BASE];
            target_act = 
             (unsigned int)(target_top[LTQ_HEAD] - LTINFO_SIZE)[LTINFO_BASEREG];
            target_sp = 
               (unsigned int)(target_top[LTQ_HEAD] - LTINFO_SIZE)[LTINFO_SPREG];
            target_top[STACK_BASE] =
           (unsigned int *)(target_top[LTQ_HEAD] - LTINFO_SIZE)[LTINFO_BASEREG];
            
#ifdef CONDITIONAL_FUTURE
            cond_bcbody =
           (unsigned int *)(target_top[LTQ_HEAD] - LTINFO_SIZE)[LTINFO_CBCBODY];
            cond_bcpc   = (target_top[LTQ_HEAD] - LTINFO_SIZE)[LTINFO_CBCPC];
#endif

            /*
             * copying task from target stack to own stack
             *   (from taget_sp to target_base)
             *   if target_act > taget_base  ==> from target_sp to target_act
             */
            active_reg = (unsigned int *)((unsigned int **)stack_bottom - 3);
            if (target_act > target_base)
                target_base = target_act;
            sp_reg     = (unsigned int *)
                ((unsigned int)active_reg - (target_base - target_sp));
            memcpy((char *)sp_reg, (char *)target_sp, target_base - target_sp);


#ifdef LTC_DEBUG
            printf ("targetact 0x%x, ownact 0x%x\n",
                    target_act, active_reg);
            printf ("targetsp 0x%x, ownsp 0x%x\n",
                    (unsigned int)target_sp,
                    (unsigned int)sp_reg);
            for ( i = (target_act - target_sp)/sizeof(unsigned int) ; i >= 0 ; i-- ) {
                printf (" ownsp:    %-8x : %-8x"
                        "   targetsp: %-8x : %-8x\n",
                        (unsigned int *)sp_reg + i,
                            *((unsigned int *)sp_reg + i),
                        (unsigned int *)target_sp + i,
                        *((unsigned int *) target_sp + i));
                }
#endif

#ifdef CONDITIONAL_FUTURE
            thrs[pro_no]->cstack[STAT] = (unsigned int *)LTAG_FALSE;
            /*
             * prepare target condition stack
             */
            if (cond_bcbody != (unsigned int *)LTAG_NIL
               && cond_bcpc != LTAG_NIL) {
                thrs[pro_no]->cstack[COND_RESULT] = (unsigned int *)LTAG_FALSE; /* *** */
                /*
                 * target thread has a condtional future task,
                 * (so child (current) becomes conditional)
                 * own thread must also steals both its condition stack
                 * its condition register.
                 */
                if (thrs[target_pro_no]->condition == VALID) {
                    /* case 3 and 4 */
                    unsigned int **tmp_cstack;
                    REGS *tmp_cregs;

                    /* swapping */
                    tmp_cstack = thrs[pro_no]->cstack;
                    tmp_cregs = thrs[pro_no]->cregs;
                    if (thrs[target_pro_no]->cstack == NULL)
                        fprintf(stderr, "^^^^^^^^^^^^^^\n");
                    thrs[pro_no]->cstack = thrs[target_pro_no]->cstack;
                    thrs[pro_no]->cregs = thrs[target_pro_no]->cregs;
                    thrs[target_pro_no]->cstack = tmp_cstack;
                    thrs[target_pro_no]->cregs = tmp_cregs;
                    /*
                     * inheritance
                     *   parent: own thread condition
                     *    child: terget thread condition
                     */
                    add_child(thrs[target_pro_no]->cstack,
                              (thrs[pro_no]->cstack));
                    (thrs[target_pro_no]->cstack)[COND_PARENT]
                        = (unsigned int *)thrs[pro_no]->cstack;
                    (thrs[target_pro_no]->cstack)[COND_CHILD_LIST] = NULL;
                    thrs[pro_no]->condition = VALID;
                } else {        /* case 5 */
                    if (thrs[target_pro_no]->cstack[COND_PARENT]
                        != (unsigned int *)LTAG_T)
                        fprintf(stderr, "error?\n");
                    thrs[target_pro_no]->condition = VALID;
                    thrs[target_pro_no]->cstack[COND_CHILD_LIST] = NULL;
                    thrs[pro_no]->condition = INVALID;
                }
                
                /*
                 * copying environment for executing condition
                 * to target thread condition stack
                 */
                target_csp = target_sp;
                target_cact = target_act;
                thrs[target_pro_no]->cregs->_active_reg
                    = (unsigned int *)(thrs[target_pro_no]->cregs->_stack_bottom) - 3;
                ((unsigned int **)thrs[target_pro_no]->cregs->_stack_top)[STACK_BASE] = (unsigned int *)(thrs[target_pro_no]->cregs->_active_reg);
                thrs[target_pro_no]->cregs->_sp_reg = (unsigned int *)
                    ((unsigned int)(thrs[target_pro_no]->cregs->_active_reg)
                     - (target_cact - target_csp));
                memcpy((char *)(thrs[target_pro_no]->cregs->_sp_reg),
                       (char *)target_sp, target_cact - target_csp);

                /*
                 * save condition bytcode information
                 * from target stack to target condition stack
                 */
                thrs[target_pro_no]->cregs->_bcbody_reg = cond_bcbody;
                thrs[target_pro_no]->cregs->_bcpc_reg = cond_bcpc;
                (thrs[target_pro_no]->cstack)[COND_BCBODY] = cond_bcbody;
                (thrs[target_pro_no]->cstack)[COND_BCPC] = (unsigned int *)cond_bcpc;


                /* 
                 * adapting target cond stack
                 *   rewriting base & active in the region of copying stack
                 */
                i = ((unsigned int)(thrs[target_pro_no]->cregs->_sp_reg) - target_sp);
                base = thrs[target_pro_no]->cregs->_sp_reg;   /* base */
                act  = base + 1;                     /* active */
                do {
                    *base += i;
                    *act  += i;
                    base = (unsigned int *)*base;
                    act = base + 1;
                } while (*base != 0);


                /*
                 * rewind stack and open condition frame
                 *   because too much stack area was copied
                 *  --- shige
                 *
                 * copied stack                right stack   open frame by
                 *  |   :   |                   |   :   |     open_condframe()
                 *  +-------+                   +-------+
                 *  |       |<- target_act      |       |
                 *  +-------+                   +-------+
                 *  |   :   |                   |   :   |
                 *  |   :   |                   |   :   |
                 *  |   :   |                   |   :   |
                 *  +-------+                   +-------+
                 *  |bcbody |                   | bcpc  | condition label
                 *  +-------+                   +-------+
                 *  | bcpc  |        ======>    | iret  | for condition hook
                 *  +-------+                   +-------+
                 *  |lfret* |                   |active | cregs->_base_reg
                 *  +-------+                   +-------+
                 *  |active |                   | base  | cregs->_base_reg
                 *  +-------+                   +-------+
                 *  | base  | cregs->_base_reg  |   :   |
                 *  +-------+
                 *  |   :   |
                 */
                /* save base_reg */
                thrs[target_pro_no]->cregs->_base_reg
                   = (unsigned int *)*(thrs[target_pro_no]->cregs->_sp_reg);
                /* rewind stack */
                thrs[target_pro_no]->cregs->_sp_reg += 5;
                /* open condition frame */
                open_condframe(thrs, target_pro_no);

            } else if (cond_bcbody != (unsigned int *)LTAG_NIL
                       || cond_bcpc != LTAG_NIL) {
                bokkan("steal_lt: Unknown error.\n"); /* Why bokkan? :P */
            } else if (thrs[target_pro_no]->condition == VALID) { /* case 1 */
                /*  parent is any task in conditional task group,
                    child(current) is ordinary task in conditional task group
                    we refer to ordinary task in conditional task group as
                    inherited task */
                add_child(thrs[pro_no]->cstack, thrs[target_pro_no]->cstack);
                thrs[pro_no]->cstack[COND_PARENT]
                    = thrs[target_pro_no]->cstack;
                thrs[pro_no]->cstack[COND_CHILD_LIST] = NULL;
                thrs[pro_no]->condition = VALID;
                /* ordinary task in conditional task group senses the parent
                   condition but it has not condition associated with it */
                (thrs[pro_no]->cstack)[COND_BCBODY] = (unsigned int *)LTAG_NIL;
            } else {            /* case 2 */
                /*  both parent and child(current) are ordinary
                    (not in conditional task group) */
                thrs[pro_no]->condition = INVALID;
            }
#endif
            
#ifdef CONDITIONAL_FUTURE
            /* unlocking target thread_data */
            mutex_unlock((mutex_t)(thrs[target_pro_no]->mutex));
#endif
            
            /*
             * adapting target stack
             */
            fut = (FUTURE *)mk_plchld();
            fut->status = UNDETERMINED;
            ((unsigned int **)stack_top)[FUTURE_PH] = target_top[FUTURE_PH];
            target_top[FUTURE_PH] = (unsigned int *)(LTAG_FUTURE | (unsigned int)fut);
            reg_reg = LTAG_FUTURE | (unsigned int)fut;
            /* rewrite return-function */
            *(((unsigned int *)target_sp) + 2) = (unsigned int)stret;
            mutex_unlock((mutex_t)target_top[STACK_LOCK]);

            /*
             * re-initializing own stack
             */
            ((unsigned int **)stack_top)[STACK_BASE] = (unsigned int *)active_reg;
            ((unsigned int **)stack_top)[LTQ_HEAD] 
                = ((unsigned int **)stack_top)[LTQ_TAIL]
                = (unsigned int *)((unsigned int **)stack_top + (STACK_INFO_END + 1));

            /*
             * adapting own stack
             *   rewriting base & active in the region of copying stack
             */
            i = ((unsigned int)sp_reg - target_sp);
            base = sp_reg;   /* base */
            act  = base + 1; /* active */
            do {
                *base += i;
                *act  += i;
                base = (unsigned int *)*base;
                act = base + 1;
            } while (*base != 0);
            
#ifdef LTC_DEBUG
            /* gettimeofday(&tp, &tzp); */
            /* printf (  "%d %d\n", tp.tv_sec, tp.tv_usec ); */
            if (thrs == pthrs)
                fprintf (stderr, "Thread[p%d] steals from Thread[p%d].\n",
                        pro_no , target_pro_no);
#if defined(CONDITIONAL_FUTURE) && defined(DEBUG_MESG)
            else if (thrs == sthrs)
                fprintf (stderr, "Thread[s%d] %d steals from Thread[s%d].\n",
                        pro_no , target_pro_no);
#endif
#endif
#if defined(CONDITIONAL_FUTURE) && defined(DEBUG_MESG)
            if (thrs == pthrs)
                fprintf (stderr,"Thread[p%d %s] steals from Thread[p%d %s].\n",
                         pro_no,
                         thrs[pro_no]->condition ? "v" : "i",
                         target_pro_no,
                         thrs[target_pro_no]->condition ? "v" : "i");
            else if (thrs == sthrs)
                fprintf (stderr,"Thread[s%d %s] steals from Thread[s%d %s].\n",
                         pro_no,
                         thrs[pro_no]->condition ? "v" : "i",
                         target_pro_no,
                         thrs[target_pro_no]->condition ? "v" : "i");
#endif
            
            /*
             * rewriting steal task's return address
             */
            rewrite_retaddr();

            return 0;
        }
    }

    return 1;
}


/*
 * rewriting steal task's return address
 *   --- takahashi & komiya
 */
void rewrite_retaddr()
{
  ENTITYP proc_entity;

  if (*(sp_reg + 2) == (unsigned int)lfretx) {
    *(sp_reg + 2) = (unsigned int)iretaddrx;
    active_reg = sp_reg;
    xreturn ();
    
  } else if (*(sp_reg + 2) == (unsigned int) lfretl) {
    *(sp_reg + 2) = (unsigned int) iretaddrl;
    base_reg = sp_reg;
    active_reg = (unsigned int *)*(sp_reg + 1);
    proc_entity = (ENTITYP)ADDRESS(*(active_reg - 1));
    
    switch ( GETtag( *( active_reg - 1 ))) {
    case LTAG_BCLOSURE :
    case LTAG_MBCLOSURE :
    case LTAG_SFUNCTION :
    case LTAG_MSFUNCTION :
    case LTAG_NCLOSURE :
    case LTAG_MNCLOSURE :
      bcbody_reg =
        (unsigned int *)ADDRESS(proc_entity->bcc.body.obj);
      break;
    default :
      bokkan("system error : in Sfunc_def.");
    }
    
    lreturn ();
    
  } else {
    printf("*(sp_reg+2) = %x\n", *(sp_reg+2));
    printf("sp_reg = %x\n", sp_reg);
    printf("stack_bottom = %x\n", stack_bottom);
    bokkan ("frame_pc is not lfretx nor lfretl");
  }
  
  return;
}
            

void stret()
{
    FUTURE *fut;
    unsigned int *top;
    mtus_thread_t thr_id;
    int i, thrs;

    top = ((unsigned int **)stack_top)[FUTURE_PH];

    if (GETtag(top) == LTAG_FUTURE) {
        fut = (FUTURE *)ADDRESS(top);
        mutex_lock(fut->lock);

        if (fut->status == DETERMINED) {
            fprintf (stderr, "why FUTURE is DETERMINED ? (in stret())\n");
        }

        fut->value.obj = reg_reg;
        fut->status = DETERMINED;
        condition_broadcast(fut->done);
        mutex_unlock(fut->lock);

#ifdef LTC_DEBUG
        printf ( "%x : RESOLVED.\n", (unsigned int)cthread_self());
#endif
        thrs = 0;
        thr_id = mtus_thread_self();
        for (i = 0; i < processors; i++) {
            if (pthrs[i]->thr == thr_id) {
#ifdef CONDITIONAL_FUTURE
                pthrs[i]->condition = INVALID;
                set_chldlst_if_not(pthrs[i]->cstack,
                                   pthrs[i]->cstack[COND_PARENT],
                                   LTAG_T);
                if ((pthrs[i]->cstack[COND_PARENT] != (unsigned int *)LTAG_T)
                    && (pthrs[i]->cstack[COND_PARENT] != NULL))
                    rm_child(pthrs[i]->cstack, pthrs[i]->cstack[COND_PARENT]);
                pthrs[i]->cstack[COND_PARENT] = (unsigned int *)LTAG_T;
#endif
                thrs = 1; /* this is a primary thread */
                break;
            }
        }
#ifdef CONDITIONAL_FUTURE
        for (i = 0; i < sprocessors; i++) {
            if (sthrs[i]->thr == thr_id) {
                sthrs[i]->condition = INVALID;
                set_chldlst_if_not(sthrs[i]->cstack,
                                   sthrs[i]->cstack[COND_PARENT],
                                   LTAG_T);
                if ((sthrs[i]->cstack[COND_PARENT] != (unsigned int *)LTAG_T)
                    && (sthrs[i]->cstack[COND_PARENT] != NULL))
                    rm_child(sthrs[i]->cstack, sthrs[i]->cstack[COND_PARENT]);
                sthrs[i]->cstack[COND_PARENT] = (unsigned int *)LTAG_T;
                thrs = 2; /* this is a secondary thread */
                break;
            }
        }
#endif

        if (thrs == 1)
            primary_steal();
#ifdef CONDITIONAL_FUTURE
        else if (thrs == 2)
            secondary_steal();
#endif
        else
            bokkan("fatal error at stret(): This thread is an unknown thread.");
    } else {
        bokkan ( "It was stealed, but why not FUTURE ?" );
    }

    return;
}

void lfretx()
{
    ((unsigned int **)stack_top)[LTQ_TAIL] -= LTINFO_SIZE;

#ifdef RMO
# if defined(sparc)
    asm("membar #StoreLoad");          /* for V9 */
# else
    mutex_lock(dummy);
    mutex_unlock(dummy);
# endif
#endif

    if (((unsigned int **)stack_top)[LTQ_TAIL]
        >= ((unsigned int **)stack_top)[LTQ_HEAD]) {
        bcpc_reg = SCPOP();
        bcbody_reg = (unsigned int *)ADDRESS(SCPOP());
    } else {
        mutex_lock((mutex_t)(((unsigned int **)stack_top)[STACK_LOCK]));
        if (((unsigned int **)stack_top)[LTQ_TAIL]
            >= ((unsigned int **)stack_top)[LTQ_HEAD]) {
            bcpc_reg = SCPOP();
            bcbody_reg = (unsigned int *)ADDRESS(SCPOP());
            mutex_unlock((mutex_t)(((unsigned int **)stack_top)[STACK_LOCK]));
        } else {
            mutex_unlock((mutex_t)(((unsigned int **)stack_top)[STACK_LOCK]));
            stret();
        }
    }

    return;
}



void lfretl()
{
    ((unsigned int **)stack_top)[LTQ_TAIL] -= LTINFO_SIZE;
#ifdef RMO
# ifdef SUN
        asm("membar #StoreLoad");          /* for V9 */
# else
        mutex_lock(dummy);
        mutex_unlock(dummy);
# endif
#endif
    if (((unsigned int **)stack_top)[LTQ_TAIL]
        >= ((unsigned int **)stack_top)[LTQ_HEAD]) {
        bcpc_reg = SCPOP();
    } else {
        mutex_lock((mutex_t)(((unsigned int **)stack_top)[STACK_LOCK]));
        if (((unsigned int **)stack_top)[LTQ_TAIL]
            >= ((unsigned int **)stack_top)[LTQ_HEAD]) {
            bcpc_reg = SCPOP();
            mutex_unlock((mutex_t)(((unsigned int **)stack_top)[STACK_LOCK]));
        } else {
            mutex_unlock((mutex_t)(((unsigned int **)stack_top)[STACK_LOCK]));
            printf ( "%x : conflict\n", (unsigned int)cthread_self());
            stret();
        }
    }

    return;

}



void open_ltframe()
{
    unsigned int **tail;
    unsigned int *ltqinfo;
    
    tail = ((unsigned int **)stack_top) + LTQ_TAIL;
    ltqinfo = *tail;

    /*
     * push LTQ base register
     */
    ltqinfo[LTINFO_BASEREG] = (unsigned int)active_reg;

    if (*tail + CHILD_GETA_OF_STACK > sp_reg) {
        bokkan_stack_overflow ();
    }

    sp_reg[-1] = LTAG_STRING | SSERDDA(bcbody_reg);
    sp_reg[-2] = GETbcarg(1);
    sp_reg[-3] = (unsigned int)lfretx;
    sp_reg[-4] = (unsigned int)active_reg;
    sp_reg[-5] = (unsigned int)base_reg;
    sp_reg -= 5;
    base_reg = sp_reg;
    PCUP(2);

    /*
     * push LTQ sp register
     */
    ltqinfo[LTINFO_SPREG] = (unsigned int)sp_reg;

#ifdef CONDITIONAL_FUTURE
    ltqinfo[LTINFO_CBCBODY] = LTAG_NIL;
    ltqinfo[LTINFO_CBCPC] = LTAG_NIL;
#endif
    
    *tail += LTINFO_SIZE;

/*
 *  --- T.Komiya
   if (pthr_count > 0 ) {
   condition_signal(thr_count_threshold);
   }
*/
    return;
}


/*
 * open_lltframe
 */
void open_lltframe()
{
    unsigned int **tail;
    unsigned int *ltqinfo;
    
    tail = ((unsigned int **)stack_top) + LTQ_TAIL;
    ltqinfo = *tail;

    if (*tail + CHILD_GETA_OF_STACK > sp_reg) {
        bokkan_stack_overflow ();
    }
    
    sp_reg[-1] = GETbcarg(3);
    sp_reg[-2] = (unsigned int)lfretl;
    if (GETbcarg(2) == 0)
        sp_reg[-3] = FETCH0(-GETbcarg(1) + 1);
    else
        sp_reg[-3] = fetch(GETbcarg(1), GETbcarg(2), 1);
    sp_reg[-4] = (unsigned int)base_reg;
    sp_reg -= 4;
    base_reg = sp_reg;
    PCUP(4);
    
    ltqinfo[LTINFO_SPREG] = (unsigned int)sp_reg;
    ltqinfo[LTINFO_BASEREG] = (unsigned int)active_reg;

    *tail += LTINFO_SIZE;
    
/*
 *  --- T.Komiya
   if (pthr_count > 0 ) {
   condition_signal(thr_count_threshold);
   }
*/
    return;
}


/*
 * open_lltframe_1 (auxuility)
 */
void open_lltframe_1()
{
    unsigned int **tail;
    unsigned int *ltqinfo;
    
    tail = ((unsigned int **)stack_top) + LTQ_TAIL;
    ltqinfo = *tail;

    if (*tail + CHILD_GETA_OF_STACK > sp_reg) {
        bokkan_stack_overflow ();
    }
    
    sp_reg[-1] = GETbcarg (2);
    sp_reg[-2] = (unsigned int)lfretl;
    sp_reg[-3] = (unsigned int)(base_reg - GETbcarg(1));
    sp_reg[-4] = (unsigned int)base_reg;
    sp_reg -= 4;
    base_reg = sp_reg;
    PCUP(3);
    
    ltqinfo[LTINFO_SPREG] = (unsigned int)sp_reg;
    ltqinfo[LTINFO_BASEREG] = (unsigned int)active_reg;

    *tail += LTINFO_SIZE;
    
/*
 *  --- T.Komiya
   if (pthr_count > 0 ) {
   condition_signal(thr_count_threshold);
   }
*/

    return;
}

#ifdef CONDITIONAL_FUTURE
/*
 * switching context --- new
 */
int switch_to_context(thread_data_t *thr)
{
    if ((thr->condition == INVALID) || (thr->cstack[COND_RESULT] == NULL)
        || (thr->cstack[COND_BCBODY] == (unsigned int *)LTAG_NIL))
        return 0;

    thr->action = MODE_COND;

#if defined(sparc) && defined(USE_REG)
    main_thread_regs    = thr->cregs;
#else
    cthread_set_data(cthread_self(), (any_t)(thr->cregs));
#endif

    return 1;
}

/*
 * allocating and initializing stack for condition
 */
void condstack_alloc(thread_data_t **thrs, int thr_no)
{
    REGS *regsp;
    unsigned int **cstack;

    if ((regsp = (REGS *)malloc(sizeof(REGS))) == NULL)
        bokkan("memory allocation error at condstack_alloc(): "
               "Can't allocate memory.");

    /*
     * allocating ltc stack, and setting stack top and bottom
     */
    if ((regsp->_stack_top = (char *)malloc(CHILD_STACK_SIZE)) == NULL)
        bokkan("memory allocation error at condstack_alloc(): "
               "Can't allocate memory.");
    regsp->_stack_bottom = regsp->_stack_top + CHILD_STACK_SIZE - 28;
    regsp->_stack_top    = regsp->_stack_top + CHILD_GETA_OF_STACK;

    *((unsigned int **)(regsp->_stack_bottom) + 1) = (unsigned int *)LTAG_NIL;
    *((unsigned int **)(regsp->_stack_bottom) + 2) = 0;
    *((unsigned int **)(regsp->_stack_bottom) + 3) = 0;

    regsp->_stack_area_bottom = regsp->_stack_bottom;

    cstack = (unsigned int **)(regsp->_stack_top);

    /*
     * set sp_reg
     *  == set_top(stack_bottom);
     */
    regsp->_sp_reg = (unsigned int *)(regsp->_stack_bottom);
    /*
     * bci does not reach!
     */
    *(--(regsp->_sp_reg)) = LTAG_NIL; /* == s_push(LTAG_NIL); */
    *(--(regsp->_sp_reg)) = 0;        /* == s_push(0); */
    *(--(regsp->_sp_reg)) = 0;        /* == s_push(0); */
    /*
     * set base
     *  == set_base(top_val());
     */
    regsp->_base_reg = regsp->_sp_reg;
    /*
     * set base & active
     *  == set_active(top_val());
     */
    regsp->_active_reg = regsp->_sp_reg;

    /*
     * stack info area
     */
    cstack[STACK_LOCK]     = (unsigned int *)mutex_alloc();
    cstack[STACK_BASE]     = (unsigned int *)(regsp->_active_reg);
    cstack[FUTURE_PH]      = (unsigned int *)LTAG_NIL;
    cstack[STACK_INFO_END] = 0;
    cstack[LTQ_HEAD]       =
        cstack[LTQ_TAIL]   = (unsigned int *)(cstack + (STACK_INFO_END + 1));
    cstack[COND_SPREG]     = 0;
/*    cstack[COND_RESULT]    = (unsigned int *)LTAG_NIL; */
/*    cstack[COND_RESULT]    =  thrs == pthrs ? (unsigned int *)LTAG_T
        :(unsigned int *)LTAG_FALSE; */
    cstack[COND_RESULT]    =  (unsigned int *)LTAG_FALSE; /* *** */
    cstack[COND_PARENT]    = (unsigned int *)LTAG_T;
    cstack[COND_CHILD_LIST] = 0;
    cstack[STAT] = (unsigned int *)LTAG_FALSE;

    /* save to thread_data */
    thrs[thr_no]->cregs = regsp;
    thrs[thr_no]->cstack = cstack;
    thrs[thr_no]->condition = INVALID;

    return;
}

/*
 * wait thread stealing scheduling (secondary processor's scheduler)
 *   --- shige
 */
void waitthr_scheduler(int thr_num)
{
#ifdef CONDITIONAL_FUTURE
    unsigned int **cstk;
#endif
    int counter = 0;
    int interval;
    char *env;
    struct timespec t, r;

    t.tv_sec = 0;
    t.tv_nsec = 1000000000;

    if ((env = getenv ( "SEC_BC_INTERVAL" )) != NULL) {
        interval = atoi(env);
    } else {
        interval = SEC_BC_INTERVAL;
    }

    /* allocating and setting thread data */
    threaddata_alloc(sthrs, thr_num);

    mutex_lock(thr_count_lock);
    sthr_count++;
    condition_wait(thr_count_threshold, thr_count_lock);
    mutex_unlock(thr_count_lock);

    /*
     * steal thread data or lazy tasks
     */
    secondary_steal();
    
    /*
     * interpret byte code
     */
    while (1) {
        /*
         * executing main
         */
        while (1) {
            ((int(*)())(*( bcbody_reg + bcpc_reg )))();
            counter++;
            /* execute condition hook */
            if (counter > interval) {
                counter = 0;
                break;
            }
        }

        /*
         * executing condition
         */
        cstk = sthrs[thr_num]->cstack;
        if (cstk[COND_RESULT] == NULL) {
            mutex_t m = (mutex_t)(sthrs[thr_num]->mutex);
            mutex_lock(m);
            sthrs[thr_num]->condition = INVALID;
            sthrs[thr_num]->cstack[COND_PARENT] = (unsigned int *)LTAG_T;
            sthrs[thr_num]->cstack[COND_CHILD_LIST] = 0;
#ifdef DEBUG_MESG
            fprintf(stderr, "task canceled.\n");
#endif
            mutex_unlock(m);
            secondary_steal();
            cstk[COND_RESULT] = (unsigned *)LTAG_FALSE;
        } else if (sthrs[thr_num]->condition == VALID) {
            if ((cstk[COND_PARENT] == (unsigned *)LTAG_T)
                && (cstk[COND_BCBODY] != (unsigned int *)LTAG_NIL)
                && (! find_act_cstack_q(sthrs[thr_num]))) {
                enq_act_cstack_q(sthrs[thr_num]);
            } else {
                while (cstk[COND_PARENT] != (unsigned int *)LTAG_T)
                    cstk = (unsigned int **)cstk[COND_PARENT];
                if (cstk[COND_RESULT] == (unsigned int *)LTAG_T) {
                    mutex_t m = (mutex_t)(sthrs[thr_num]->mutex);
#ifdef DEBUG_MESG
                    fprintf(stderr, "result = T\n");
#endif
                    mutex_lock(m);
                    sthrs[thr_num]->cstack[COND_RESULT] = (unsigned int *)LTAG_T;
                    if (*((unsigned int *)stack_bottom - 1) != (unsigned int)secondary_steal)
                      fprintf(stderr, "???????\n");
                    *((unsigned int *)stack_bottom - 1)
                        = (unsigned int)primary_steal;
                    if (sthrs[thr_num]->cstack[COND_BCBODY] != LTAG_NIL) {
                        /* the task is not ordinary task in conditional task
                           group, but conditional task */
                        if (((unsigned int **)(sthrs[thr_num]->cregs->_stack_top))[COND_SPREG][2] != (unsigned int)iretaddrl_cond_secondary)
                            fprintf(stderr, "?????***\n");
                        ((unsigned int **)(sthrs[thr_num]->cregs->_stack_top))[COND_SPREG][2] = (unsigned int)iretaddrl_cond_primary;
                    }
                    sthrs[thr_num]->condition = INVALID;

                    put_waitthr(twlp, sthrs);
                    threaddata_alloc(sthrs, thr_num);
                    mutex_unlock(m);
                    secondary_steal();
                }
            }
        }
    }

    /* not reach! */
    return;
}

/*
 * steal function for secondary processors
 *   This is a wrapper function.
 *   --- shige
 */
void secondary_steal()
{
    int thr_num;
    int result;
    struct timespec t, r;

    t.tv_sec = 0;
    t.tv_nsec = 100000000;

    if((thr_num = get_thread_number(sthrs)) < 0)
        bokkan("fatal error at secondary_steal(): "
               "this thread is an unknown thread.\n");

    result = 1; /* default: steal action failed */

    while(result != 0) {
#ifdef DEBUG_CFUTURE
        fprintf(stderr, "DEBUG [spro%d]: secondary_steal()\n", thr_num);
#endif
        result = get_waitthr(fwlp, sthrs);

        if (result != 0)
            result = steal_lt(sthrs);
        else
            break;

        /* if steal action successed */
        if (result == 0)
            break;

        /* if cannot get task, wait for a while */
        nanosleep(&t, &r);
    }

    return;
}

/*
 * add stack of wating task to waitthr_list
 *   --- shige
 */
void put_waitthr(waitthr_list_t *wlp, thread_data_t **thrs)
{
    /*
     * FIFO policy is only implemented!
     *  put thread-data to the last of wiatthr_list.
     *  --- shige (Nov 2, 1998)
     */

    int thr_num;
    waitthr_t *new, *tmp;
    thread_data_t *thrdata;
    
    /* get my thread number */
    if((thr_num = get_thread_number(thrs)) < 0)
        bokkan("fatal error at put_waitthr(): "
               "this thread is an unknown thread.\n");

#ifdef DEBUG_CFUTURE
        fprintf(stderr,
                "DEBUG [%spro%d]: put_waithr_list start\n",
                ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num);
#endif

    /* allocate new waitthr data */
    if((new = (waitthr_t *)malloc(sizeof(waitthr_t))) == NULL)
        bokkan("memory allocation error at put_waitthr(): "
               "Can't allocate memory.");
    new->priority = 0; /* (XXX) not implemented! --- shige (Nov 2, 1998) */
    new->thrdata = thrs[thr_num];

    /* locking waitthr_list operations */
    mutex_lock(wlp->mutex);

#ifdef DEBUG_CFUTURE
        fprintf(stderr,
                "DEBUG [%spro%d]: put_waithr_list   locked 0x%x\n",
                ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num, wlp);
#endif

#ifdef DEBUG_CFUTURE
        fprintf(stderr,
                "DEBUG [%spro%d]: put_waitthr "
                "base=0x%x, active=0x%x, sp=0x%x, body=0x%x, pc=0x%x, stack=0x%x\n"
                ,
                ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num,
                base_reg, active_reg, sp_reg,
                bcbody_reg, bcpc_reg,
                (unsigned int **)stack_top);
#endif

    /* inserting stack of waitting task to last element */
    if (wlp->firstp == NULL) {
        new->prevp = NULL;
        new->nextp = NULL;
        wlp->firstp = new;
        wlp->lastp = new;
    } else {
        wlp->lastp->nextp = new;
        new->prevp = wlp->lastp;
        new->nextp = NULL;
        wlp->lastp = new;
    }
    
    /* unlock waitthr_list operations */
    mutex_unlock(wlp->mutex);
#ifdef DEBUG_CFUTURE
    fprintf(stderr,
            "DEBUG [%spro%d]: put_waithr_list unlocked 0x%x\n",
            ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num, wlp);
#endif

#ifdef DEBUG_MESG
    fprintf(stderr,
            "Thread[%s%d %s] put to queue.\n",
            ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num,
            thrs[thr_num]->condition ? "v" : "i");
#endif
#ifdef DEBUG_CFUTURE
    fprintf(stderr,
            "DEBUG [%spro%d]: put_waithr_list end\n",
            ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num);
#endif

    return;
}

/*
 * delete stack of wating task from waitthr_list's `firstp'
 *   --- shige
 */
int get_waitthr(waitthr_list_t *wlp, thread_data_t **thrs)
{
    /*
     * FIFO policy is only implemented!
     *  get thread-data from the first of waitthr_list.
     *  --- shige (Nov 2, 1998)
     */

    int thr_num;
    waitthr_t *tmp;
    
    /* get my thread number */
    if((thr_num = get_thread_number(thrs)) < 0)
        bokkan("fatal error at get_waitthr(): "
               "this thread is an unknown thread.\n");

    /*
     * list has no waitthr data (non-locking check)
     */
    if (wlp->firstp == NULL) {
        return 1;
    }

#ifdef DEBUG_CFUTURE
        fprintf(stderr,
                "DEBUG [%spro%d]: get_waithr_list start\n",
                ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num);
#endif

    /* locking waitthr_list operations */
    mutex_lock(wlp->mutex);
#ifdef DEBUG_CFUTURE
        fprintf(stderr,
                "DEBUG [%spro%d]: get_waithr_list   locked 0x%x\n",
                ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num, wlp);
#endif

    /*
     * ***after locking***, list has no waitthr data
     */
    if (wlp->firstp == NULL) {
        mutex_unlock(wlp->mutex);
#ifdef DEBUG_CFUTURE
        fprintf(stderr,
                "DEBUG [%spro%d]: get_waithr_list unlocked 0x%x FAILED\n",
                ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num, wlp);
#endif
        return 1;
    }
    
    /* deleting stack of waitting task from first element */
    tmp = wlp->firstp;
    if (wlp->firstp->nextp == NULL) { /* only one element is exist */
        wlp->firstp = NULL;
        wlp->lastp = NULL;
    } else {
        wlp->firstp = tmp->nextp;
        wlp->firstp->prevp = NULL;
    }
    
    /* unlock waitthr_list operations */
    mutex_unlock(wlp->mutex);
#ifdef DEBUG_CFUTURE
    fprintf(stderr,
            "DEBUG [%spro%d]: get_waithr_list unlocked 0x%x\n",
            ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num, wlp);
#endif

    if (tmp->thrdata->cstack[COND_RESULT] == NULL) {
        free(((char *)(tmp->thrdata->mstack) - CHILD_GETA_OF_STACK));
        free(tmp->thrdata->regs);
        if (thrs[thr_num]->cstack == NULL)
            fprintf(stderr, "???\n");
        free(((char *)(tmp->thrdata->cstack) - CHILD_GETA_OF_STACK));
        free(tmp->thrdata->cregs);
        free(tmp);
        return 1;
    }
    /* switching context */
    mutex_lock(thrs[thr_num]->mutex);
    free(((char *)(thrs[thr_num]->mstack) - CHILD_GETA_OF_STACK));
    free(thrs[thr_num]->regs);
    if (thrs[thr_num]->cstack != NULL) {
        free(((char *)(thrs[thr_num]->cstack) - CHILD_GETA_OF_STACK));
        free(thrs[thr_num]->cregs);
    }

#if 0  /**********************************/
    thrs[thr_num]->mstack = tmp->thrdata->mstack;
    if (tmp->thrdata->cstack == NULL)
        fprintf(stderr, "vvvvvvvvvvvvvv\n");
    thrs[thr_num]->cstack = tmp->thrdata->cstack;
    thrs[thr_num]->regs = tmp->thrdata->regs;
    thrs[thr_num]->cregs = tmp->thrdata->cregs;
#if defined(sparc) && defined(USE_REG)
    main_thread_regs = thrs[thr_num]->regs;
#else 
    cthread_set_data(cthread_self(), (any_t)(thrs[thr_num]->regs));
#endif
    thrs[thr_num]->stack = thrs[thr_num]->mstack;
    thrs[thr_num]->thr = mtus_thread_self();
    thrs[thr_num]->condition = tmp->thrdata->condition; /* new */
#endif /**********************************/

    thrs[thr_num]->condition = INVALID; /* new */
    mutex_unlock(thrs[thr_num]->mutex);
/* new */
    thrs[thr_num] = tmp->thrdata;
#if defined(sparc) && defined(USE_REG)
    main_thread_regs = thrs[thr_num]->regs;
#else 
    cthread_set_data(cthread_self(), (any_t)(thrs[thr_num]->regs));
#endif
    thrs[thr_num]->stack = thrs[thr_num]->mstack;
    thrs[thr_num]->thr = mtus_thread_self();
/* new end */
    
#ifdef DEBUG_CFUTURE
        fprintf(stderr,
                "DEBUG [%spro%d]: get_waitthr "
                "base=0x%x, active=0x%x, sp=0x%x, body=0x%x, pc=0x%x, stack=0x%x\n"
                ,
                ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num,
                base_reg, active_reg, sp_reg,
                bcbody_reg, bcpc_reg,
                (unsigned int **)stack_top);
#endif
/* new
    free(tmp);
 */

#ifdef DEBUG_MESG
    fprintf(stderr,
            "Thread[%s%d %s] steals from queue.\n",
            ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num,
            thrs[thr_num]->condition ? "v" : "i");
#endif
#ifdef DEBUG_CFUTURE
        fprintf(stderr,
                "DEBUG [%spro%d]: get_waithr_list end\n",
                ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num);
#endif

    return 0;
}

/*
 * scheduler for cond check
 */
void cstack_exec_scheduler(int dum)
{
    thread_data_t *thr;
    struct timespec t, r;
    REGS *regsp;
    mutex_t m;

    t.tv_sec = 0;
    t.tv_nsec = 50000000;      /* 50ms */

#if defined(sparc) && defined(USE_REG)
    if ((main_thread_regs = (REGS *)malloc(sizeof(REGS))) == NULL)
        bokkan("memory allocation error at cstack_exec_scheduler(): "
               "Can't allocate memory.");
#else
    if ((regsp = (REGS *)malloc(sizeof(REGS))) == NULL)
        bokkan("memory allocation error at cstack_exec_scheduler(): "
               "Can't allocate memory.");
    cthread_set_data(cthread_self(), (any_t)regsp);
#endif

    mutex_lock(thr_count_lock);
    sthr_count++;
    condition_wait(thr_count_threshold, thr_count_lock);
    mutex_unlock(thr_count_lock);

    while (1) {
        if (! is_empty_act_cstack_q()) {
            mutex_t m;

            if (thr = current_cond_thr = deq_act_cstack_q()) {
                m = thr->mutex;
                mutex_lock(m);
                if (switch_to_context(thr)) {
#ifdef FOO
                    fprintf(stderr, "cond check start...\n");
#endif
                    while (thr->action == MODE_COND)
                        ((int(*)())(*( bcbody_reg + bcpc_reg )))();
#ifdef FOO
                    fprintf(stderr, "cond check done.\n");
#endif
                    enq_act_cstack_q(thr);
                }
                mutex_unlock(m);
            }
        }
        nanosleep(&t, &r);
    }
}


/*
 * open_ltframe_c (args: [label])
 *   --- shige
 */
void open_ltframe_c()
{
    int policy;
    unsigned int **tail;
    unsigned int *ltqinfo;
    mtus_thread_t thr_id;
    extern struct sched_param sschedparam;

/*
    struct sched_param param;
*/
    
    tail = ((unsigned int **)stack_top) + LTQ_TAIL;
    ltqinfo = *tail;

    /*
     * push LTQ base register
     */
    ltqinfo[LTINFO_BASEREG] = (unsigned int)active_reg;

    if (*tail + CHILD_GETA_OF_STACK > sp_reg) {
        bokkan_stack_overflow ();
    }

    sp_reg[-1] = LTAG_STRING | SSERDDA(bcbody_reg);
    sp_reg[-2] = GETbcarg(1);
    sp_reg[-3] = (unsigned int)lfretx;
    sp_reg[-4] = (unsigned int)active_reg;
    sp_reg[-5] = (unsigned int)base_reg;
    sp_reg -= 5;
    base_reg = sp_reg;
    PCUP(2);

    /*
     * push LTQ sp register
     */
    ltqinfo[LTINFO_SPREG] = (unsigned int)sp_reg;

    /*
     * The following operand must be executed in auxility function.
     *   *tail += LTINFO_SIZE;
     * --- shige
     */

    /*
     * down my thread priority
     */
#if 1
    if (pthread_setschedparam(mtus_thread_self(), SCHED_OTHER, &sschedparam))
        fprintf(stderr, "Cannot chage the priority at open_frame_c()\n");
#endif
    return;
}

/*
 * open_ltframe_c_aux (arg: [condition start address] [offset])
 *   --- shige
 */
void open_ltframe_c_aux()
{
    unsigned int **tail;
    unsigned int *ltqinfo;
    
    tail = ((unsigned int **)stack_top) + LTQ_TAIL;
    ltqinfo = *tail;

    /* condition information */
    ltqinfo[LTINFO_CBCBODY] = (unsigned int)bcbody_reg;
    ltqinfo[LTINFO_CBCPC]   = (unsigned int)GETbcarg(1);

    *tail += LTINFO_SIZE;

    PCUP(3);

    return;
}

/*
 * open condition's frame
 *  This function is called from steal_lt().
 *  *** Caution ***: A steal thread setup target thread's condition stack.
 *   --- shige
 */
void open_condframe(thread_data_t **thrs, int thr_num)
{
    thread_data_t *target;

    target = thrs[thr_num];

#ifdef DEBUG_CFUTURE
    fprintf(stderr, "DEBUG [%spro%d]: open_condframe\n",
            ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")), thr_num);
#endif
    
    (target->cregs->_sp_reg)[-1] = target->cregs->_bcpc_reg;
    /* switching iretaddrl_cond_* */
    if (thrs == pthrs)
        (target->cregs->_sp_reg)[-2] = (unsigned int)iretaddrl_cond_primary;
    else if (thrs == sthrs)
        (target->cregs->_sp_reg)[-2] = (unsigned int)iretaddrl_cond_secondary;
    else
        bokkan("fatal error at open_condframe(): "
               "This is an unknown thread data list.");

    (target->cregs->_sp_reg)[-3]
        = (unsigned int)(target->cregs->_base_reg);
    (target->cregs->_sp_reg)[-4]
        = (unsigned int)(target->cregs->_base_reg);
    target->cregs->_sp_reg -= 4;
    target->cregs->_base_reg = target->cregs->_sp_reg;

    /*
     * this function is not called from bci
     * therefore don't increse bc_pc offset 
     * PCUP(2);
     */

    /*
     * If the result of this condition is not true,
     * this stack is put the side for a while.
     * Hence, we need sp_reg infomation, when this stack is
     * resumed (executed again). 
     */
    (target->cstack)[COND_SPREG] = (unsigned int *)(target->cregs->_sp_reg);

#ifdef DEBUG_CFUTURE
    fprintf(stderr,
            "DEBUG [%spro%d]: [store] bcbody=0x%x, bcpc=0x%x, sp=0x%x\n",
            ((thrs == pthrs) ? "p" : ((thrs == sthrs) ? "s" : "?")),
            thr_num,
            target->cregs->_bcbody_reg,
            target->cregs->_bcpc_reg,
            target->cregs->_sp_reg);
#endif
    
    return;
}

/*
 * iretaddrl_cond_primary
 *
 * This function is called ***in condition stack context***.
 * Therefore the following *_reg belongs to condition stack's regs.
 * And this function is for ***primary*** processors.
 *   --- shige
 */
void iretaddrl_cond_primary()
{
    int thr_num;
    int policy;
    mtus_thread_t thr_id;
    
    current_cond_thr->action = MODE_MAIN;
    if (reg_reg != LTAG_FALSE) {
        if (((unsigned int **)stack_top)[COND_RESULT] != NULL) {
            ((unsigned int **)stack_top)[COND_RESULT] = (unsigned int *)LTAG_T;
            set_chldlst(current_cond_thr->cstack, LTAG_T);
        }
#ifdef DEBUG_CFUTURE
        fprintf(stderr, "DEBUG [ppro%d]: cond result = TRUE\n", thr_num);
#endif
        /*
         * free condition stack and regs
         */
        current_cond_thr->condition = INVALID;
    }
    else {
        if (((unsigned int **)stack_top)[COND_RESULT] != NULL) {
            ((unsigned int **)stack_top)[COND_RESULT] = (unsigned int *)LTAG_FALSE;

#ifdef DEBUG_CFUTURE
            fprintf(stderr, "DEBUG [ppro%d]: cond result = FALSE\n", thr_num);
#endif
            /* store bytecode address into thread's registers */
            bcbody_reg = ((unsigned int **)stack_top)[COND_BCBODY];
            if (bcbody_reg == (unsigned int *)LTAG_NIL)
                fprintf(stderr, "xxxx\n");
            bcpc_reg   = (unsigned int)((unsigned int **)stack_top)[COND_BCPC];
            /* store stack pointer address into thread's registers */
            sp_reg   = ((unsigned int **)stack_top)[COND_SPREG];
            base_reg = sp_reg;

#ifdef DEBUG_CFUTURE
            fprintf(stderr,
                    "DEBUG [ppro%d]: [load] bcbody=0x%x, bcpc=0x%x, sp=0x%x\n",
                    thr_num, bcbody_reg, bcpc_reg, sp_reg);
#endif
        }
    }

#ifdef DEBUG_CFUTURE
    fprintf(stderr, "DEBUG [ppro%d]: iretaddrl_cond_primary: end\n", thr_num);
#endif
    
    return;
}

/*
 * iretaddrl_cond_secondary
 *
 * This function is called ***in condition stack context***.
 * Therefore the following *_reg belongs to condition stack's regs.
 * And this function is for ***secondary*** processors.
 *   --- shige
 */
void iretaddrl_cond_secondary()
{
    int thr_num;
    
    current_cond_thr->action = MODE_MAIN;
    if (reg_reg != LTAG_FALSE) {
        if (((unsigned int **)stack_top)[COND_RESULT] != NULL) {
            ((unsigned int **)stack_top)[COND_RESULT] = (unsigned int *)LTAG_T;
            set_chldlst(current_cond_thr->cstack, LTAG_T);
        }
#ifdef DEBUG_CFUTURE
        fprintf(stderr, "DEBUG [spro%d]: cond result = TRUE\n", thr_num);
#endif
/*        current_cond_thr->condition = INVALID; */
    }
    else {
        if (((unsigned int **)stack_top)[COND_RESULT] != NULL) {
            ((unsigned int **)stack_top)[COND_RESULT] = (unsigned int *)LTAG_FALSE;
#ifdef DEBUG_CFUTURE
            fprintf(stderr, "DEBUG [spro%d]: cond result = FALSE\n", thr_num);
#endif
#ifdef DEBUG_CFUTURE
            fprintf(stderr,
                    "DEBUG [spro%d]: [load] bcbody=0x%x, bcpc=0x%x, sp=0x%x\n",
                    thr_num, bcbody_reg, bcpc_reg, sp_reg);
#endif
        }
    }

    /* store bytecode address into thread's registers */
    bcbody_reg = ((unsigned int **)stack_top)[COND_BCBODY];
    if (bcbody_reg == (unsigned int *)LTAG_NIL)
        fprintf(stderr, "yyy\n");
    bcpc_reg   = (unsigned int)((unsigned int **)stack_top)[COND_BCPC];
    /* store stack pointer address into thread's registers */
    sp_reg   = ((unsigned int **)stack_top)[COND_SPREG];
    base_reg = sp_reg;

#ifdef DEBUG_CFUTURE
    fprintf(stderr, "DEBUG [spro%d]: iretaddrl_cond_secondary: end\n", thr_num);
#endif
    
    return;
}

/*
 * inherit parent result
 *   --- shige
 */
unsigned int inherit_result(thread_data_t *thr)
{
    unsigned int **condstack;
    unsigned int *parent;
    unsigned int result;

    condstack = thr->cstack;
    parent = condstack[COND_PARENT];
    result = LTAG_NIL;

    for (parent = condstack[COND_PARENT];
         parent != NULL; parent = (unsigned int *)parent[COND_PARENT]) {
        if (parent[COND_RESULT] == LTAG_FALSE)
            result = LTAG_FALSE;
    }
    
    return result;
}

/*
 * suicide --- Scheme Function
 */
int Ssuicide()              /* this can be used only in future condition */
{
    int thr_num;
    FUTURE *fut;
    unsigned int *futaddr;
    mtus_thread_t thr_id;
    int i, thr_type;
    thread_data_t **thrs;

    futaddr = ((unsigned int **)current_cond_thr->mstack)[FUTURE_PH];

    if (GETtag(futaddr) == LTAG_FUTURE) {
        fut = (FUTURE *)ADDRESS(futaddr);
        mutex_lock(fut->lock);

        if (fut->status == DETERMINED) {
            fprintf (stderr, "why FUTURE is DETERMINED ? (in suicide())\n");
        }
        fut->value.obj = getarg(1);
        fut->status = DETERMINED;
        condition_broadcast(fut->done);
        mutex_unlock(fut->lock);
    } else {
        bokkan ( "It was stealed, but why not FUTURE ?" );
    }

    set_allchldlst(current_cond_thr->cstack, NULL);
    current_cond_thr->cstack[COND_RESULT] = NULL;
    current_cond_thr->action = MODE_MAIN;

    /*
     * free condition stack and regs
     */
    current_cond_thr->condition = INVALID;

    fprintf(stderr, "suicide.\n");
    return 0;
}

int Scurrent_stat()             /* this can be used only in future condition */
{
    int thr_num;
    unsigned int **cstk;

    if (((unsigned int **)stack_top)[STAT] == (unsigned int *)LTAG_T)
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
    return 0;
}

int Spromote()
{
    SOBJ obj;
    ENTITYP fut;
    int i, thr_num;
    unsigned int **cstk;
    mtus_thread_t id;
    thread_data_t **thrs;

    obj.obj = getarg(1);
    if ((GETtag(obj.obj) != LTAG_FUTURE)) {
        set_reg(LTAG_FALSE);
        return 0;
    }
    fut = (ENTITYP)ADDRESS(obj.obj);
    if (fut->future.status == DETERMINED) {
        set_reg(LTAG_FALSE);
        return 0;
    }

    for (i = 0; i < sprocessors; i++) {
        if ((sthrs[i] != NULL) && ((sthrs[i]->mstack)[FUTURE_PH] == obj.obj)) {
            thr_num = i;
            thrs = sthrs;
            break;
        }
    }
    if (i == sprocessors) {
        for (i = 0; i < processors; i++) {
            if ((pthrs[i] != NULL) && ((pthrs[i]->mstack)[FUTURE_PH] == obj.obj)){
                thr_num = i;
                thrs = pthrs;
                break;
            }
        }
        if (i == processors) {
            set_reg(LTAG_FALSE);
            return 0;
        }
    }
    cstk = thrs[thr_num]->cstack;
    cstk[STAT] = (unsigned int *)LTAG_T;
    cstk[COND_RESULT] = (unsigned int *)LTAG_T;
#ifdef DEBUG_MESG
    fprintf(stderr, "promoted from %s.\n", thrs == pthrs ? "P" : "S");
#endif
    set_reg(LTAG_T);
    return 0;
}

int Scancel()
{
    SOBJ obj;
    ENTITYP fut;
    int i, thr_num;
    unsigned int **cstk;
    mtus_thread_t id;
    thread_data_t **thrs;

    obj.obj = getarg(1);
    if ((GETtag(obj.obj) != LTAG_FUTURE)) {
        set_reg(LTAG_FALSE);
        return 0;
    }
    fut = (ENTITYP)ADDRESS(obj.obj);
    if (fut->future.status == DETERMINED) {
        set_reg(LTAG_FALSE);
        return 0;
    }

    for (i = 0; i < sprocessors; i++) {
        if ((sthrs[i] != NULL) && ((sthrs[i]->mstack)[FUTURE_PH] == obj.obj)) {
            thr_num = i;
            thrs = sthrs;
            break;
        }
    }
    if (i == sprocessors) {
        for (i = 0; i < processors; i++) {
            if ((pthrs[i] != NULL) && ((pthrs[i]->mstack)[FUTURE_PH] == obj.obj)){
                thr_num = i;
                thrs = pthrs;
                break;
            }
        }
        if (i == processors) {
            set_reg(LTAG_FALSE);
            return 0;
        }
    }
    cstk = thrs[thr_num]->cstack;
    cstk[STAT] = (unsigned int *)LTAG_FALSE;
    set_allchldlst(cstk, NULL);
    cstk[COND_RESULT] = NULL;
    thrs[thr_num]->action = MODE_MAIN;
    thrs[thr_num]->condition = INVALID;
#ifdef DEBUG_MESG
    fprintf(stderr, "canceling  %s.\n", thrs == pthrs ? "P" : "S");
#endif
    set_reg(LTAG_T);
    return 0;
}

#endif /* CONDITIONAL_FUTURE */


#ifdef LTC_DEBUG
void stack_dumper ();

void stack_dumper ()
{
    int     i, j;
    char    *pp;

    for ( j = 0 ; j < CHILD_STACK_SIZE ; j += 4 ) {
        for ( i = 0 ; i < processors ; i++ ) {
            pp = (char *) pthrs[i]->stack;
            printf ( "%8xh : %02x %02x %02x %02x\t",
                     (unsigned) (pp + j),
                     (unsigned char) *( pp + j ), (unsigned char) *( pp + j + 1 ),
                     (unsigned char) *( pp + j + 2 ), (unsigned char) *( pp + j + 3 ));
        }
        printf ("\n");
    }
}
#endif  /* LTC_DEBUG */


#endif  /* LTC */
#endif  /* THREAD */
