/*
 * ltc.h - lazy task creation scheduling
 */
/*
 * $Id: ltc.h,v 1.2 1999/07/15 05:39:25 komiya Exp $
 *
 * Created: shige@kuis.kyoto-u.ac.jp   (Jan 11, 1999)
 *
 */

#ifndef __LTC_H__
#define __LTC_H__

/*
 * Stack info for Lazy Task Creation:
 *
 *  top ==>+----------+
 *     [0] |stack_lock|
 *         +----------+
 *     [1] |stack_base|
 *         +----------+
 *     [2] |  future  |
 *         +----------+
 *     [3] |   tail   |
 *         +----------+
 *     [4] |   head   |
 *         +----------+
 *     [5] | cbcbody  |  #ifdef CONDITIONAL_FUTURE
 *         +----------+
 *     [6] |  cbcpc   |  #ifdef CONDITIONAL_FUTURE
 *         +----------+
 *     [7] |  cspreg  |  #ifdef CONDITIONAL_FUTURE
 *         +----------+
 *     [8] | info_end |  == 0  (stop of stack infos)
 *         +----------+
 *         |    :     |
 *         |(LT-infos)|
 *         |    :     |
 */
#define STACK_LOCK       0
#define STACK_BASE       1
#define FUTURE_PH        2
#define LTQ_TAIL         3
#define LTQ_HEAD         4
#ifdef CONDITIONAL_FUTURE
#define COND_BCBODY      5   /* if condition == VALID && COND_BCBODY == NIL
                                then this task is ordinary task in conditional
                                task group */
#define COND_BCPC        6
#define COND_SPREG       7
#define COND_RESULT      8   /* if NULL then the current task should be
                                canceled */
#define COND_PARENT      9   /* T -> the current task is a root of conditional
                                task group */
#define COND_CHILD_LIST 10
#define STAT 11
#endif
#ifdef CONDITIONAL_FUTURE
#define STACK_INFO_END  12
#else
#define STACK_INFO_END   5
#endif

/*
 * LT(Lazy Task) info:
 *
 *       +---------+
 *   [0] | sp_reg  |
 *       +---------+
 *   [1] |base_reg |
 *       +---------+
 *   [2] |cond_body|  #ifdef CONDITIONAL_FUTURE
 *       +---------+
 *   [3] | cond_pc |  #ifdef CONDITIONAL_FUTURE
 *       +---------+
 */
#define LTINFO_SPREG      0
#define LTINFO_BASEREG    1
#ifdef CONDITIONAL_FUTURE
#define LTINFO_CBCBODY    2
#define LTINFO_CBCPC      3
#endif

#ifdef CONDITIONAL_FUTURE
#define LTINFO_SIZE       4
#else
#define LTINFO_SIZE       2
#endif

/*
 * action type
 *   thread_data->action
 */
#define MODE_MAIN  0
#define MODE_COND  1

/*
 * status of condition stack
 *   thread_data->condition
 */
#define VALID    1
#define INVALID  0


/*
 * variable type
 */
typedef struct thread_data {
    mutex_t mutex;
    mtus_thread_t thr;
    mtus_thread_t bthr;
    unsigned int **stack;
#ifdef CONDITIONAL_FUTURE
    int action;
    int condition;
    unsigned int **mstack;
    unsigned int **cstack;
    REGS *regs;
    REGS *cregs;
#if 0
    unsigned int bcbody, bcpc;
#endif
#endif
} thread_data_t;

#ifdef CONDITIONAL_FUTURE
/* stack list of waiting tasks */
typedef struct waitthr {
    int priority; /* (XXX) not implemented! --- shige (Nov 2, 1998) */
    thread_data_t *thrdata;
    struct waitthr *nextp;
    struct waitthr *prevp;
} waitthr_t;

typedef struct waitthr_list {
    struct waitthr *firstp;
    struct waitthr *lastp;
    mutex_t mutex;
} waitthr_list_t;

/* child list */
typedef struct chldlst {
    unsigned int *stk;
    struct chldlst *next;
} chldlst_t;

/* cond stack list */
typedef struct cstack_lst {
    thread_data_t *thr;
    struct cstack_lst *next;
} cstack_lst_t;

typedef struct act_cstack_queue {
    cstack_lst_t *head;
    cstack_lst_t *tail;
} act_cstack_queue_t;
#endif

/*
 * experimental definition
 */
/* #define SECONDARY_THREAD_NUM  6 */
#define PRI_BC_INTERVAL       100
#define SEC_BC_INTERVAL       100

#endif /* __LTC_H__ */
