/*
 * mine.c - 
 */

/*
 * $Id: mine.c,v 1.7 2014/10/28 02:22:55 komiya Exp $
 *
	mine.c	-- The Official Editor of YUASA Lab.
	
	(C) Copyright Yuasa Lab., 1990.  All rights reserved.
*/

static char *RCSid="$Id: mine.c,v 1.7 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"

#ifdef KCL

#ifndef ED_MINE
#include "include.h"
#endif

//#include <varargs.h>
#include <stdarg.h>

#ifdef __CYGWIN32__
#include <stdlib.h>
#endif

#define SPRINTF1(s,f,x) SPRINTF((s),(f),(x))
#define SPRINTF2(s,f,x,y) SPRINTF((s),(f),(x),(y))
#define SPRINTF3(s,f,x,y,z) SPRINTF((s),(f),(x),(y),(z))
char *
SPRINTD(s,n)
char *s; int n;
{
	if (n != 0) {
		s = SPRINTD(s, n/10);
		*s++ = n%10 + '0';
	}
	return(s);
}

SPRINTF(char *s, ...)
{
	va_list ap;
	char *p, *sval;
	int ival;

	va_start(ap, s);
        p = va_arg(ap, char *);

	for (; *p; p++) {
		if (*p != '%') {
			*s++ = *p;
			continue;
		}
		switch(*++p) {
		case 'd':
			ival = va_arg(ap, int);
			if (ival == 0)
				*s++ = '0';
			else if (ival < 0) {
				*s++ = '-';
				s = SPRINTD(s, -ival);
			} else
				s = SPRINTD(s, ival);
			break;
		case 's':
			for (sval = va_arg(ap, char *); *sval; sval++)
				*s++ = *sval;
			break;
		default:
			*s++ = '%';
			*s++ = *p;
			break;
		}
	}
	*s = '\0';
	va_end(ap);
}

/* old C version
SPRINTF(va_alist) va_dcl
{
	va_list ap;
	char *s, *p, *sval;
	int ival;

	va_start(ap);
	s = va_arg(ap, char *);
	p = va_arg(ap, char *);

	for (; *p; p++) {
		if (*p != '%') {
			*s++ = *p;
			continue;
		}
		switch(*++p) {
		case 'd':
			ival = va_arg(ap, int);
			if (ival == 0)
				*s++ = '0';
			else if (ival < 0) {
				*s++ = '-';
				s = SPRINTD(s, -ival);
			} else
				s = SPRINTD(s, ival);
			break;
		case 's':
			for (sval = va_arg(ap, char *); *sval; sval++)
				*s++ = *sval;
			break;
		default:
			*s++ = '%';
			*s++ = *p;
			break;
		}
	}
	*s = '\0';
	va_end(ap);
}
*/

#else

#include <stdio.h>
#define SPRINTF1(s,f,x) sprintf((s),(f),(x))
#define SPRINTF2(s,f,x,y) sprintf((s),(f),(x),(y))
#define SPRINTF3(s,f,x,y,z) sprintf((s),(f),(x),(y),(z))

#endif

#ifdef unix
#include <signal.h>

#ifdef MACHTEN
#include <curses.h>
#endif

#ifdef LINUX_OLD
#include <termio.h>
#include <curses.h>
#define SAVE 0
#define LOAD 1
#define MINE 2
#endif

#endif

typedef int character;

#define TRUE 1
#define FALSE 0
int TABSPACE;

#ifdef LINUX_OLD
struct termio tin, tio;
#else
int COLS;
int LINES;
#endif

int fixed_window;

#ifdef TUS
#define EMBED
#else
#ifdef KCL
#define EMBED
#else
#undef EMBED
#endif
#endif

#ifdef unix
#define BUFlen 1048576
#define KBUFlen 65536
extern char *tgetstr();
extern char *getenv();

char termcap_bp[1024];
char *tty_name;
char tgetstr_area[1024];
#endif

#ifdef pc
#include <dos.h>
#define BUFlen 30000
#define KBUFlen 16384
#endif

int MAXCOL;
int MAXROW;
int TEXTROW;
int BANNERROW;
int MSGROW;
int CENTERROW;

#define BEND (BUFlen-1)
#define BSTART 0
#define GETA 16
#define HYAKU 128
#define KLOGlen 1024
#define JUU 10

#define CtrlA	1
#define CtrlB	2
#define CtrlC	3
#define CtrlD	4
#define CtrlE	5
#define CtrlF	6
#define CtrlG	7
#define CtrlH	8
#define CtrlI	9
#define CtrlJ	10
#define CtrlK	11
#define CtrlL	12
#define CtrlM	13
#define CtrlN	14
#define CtrlO	15
#define CtrlP	16
#define CtrlQ	17
#define CtrlR	18
#define CtrlS	19
#define CtrlT	20
#define CtrlU	21
#define CtrlV	22
#define CtrlW	23
#define CtrlX	24
#define CtrlY	25
#define CtrlZ	26

#define LS	((char)-1)
#define DEL	CtrlH
#define TAB	CtrlI
#define NL	CtrlJ
#define CR	CtrlM
#define SP	' '
#define ESC	27
#define Rubout	127

#define SIG_CONT	0
#define SIG_EXIT	1
#define SIG_LOAD	2

char tmpstr[HYAKU];

#undef MALLOCED_BUF
#ifndef MALLOCED_BUF
char buf[BUFlen];
#else
char *buf;
#endif
char kbuf[KBUFlen];

int prev;
int next;
int ccol;
int ccol1;
int pcol;
int prow;
int crow;
int modp;
int mark;
int Ginc;
int Gdec;
int kbufp;
int kbufleft;
int lastcom;
int lastcom_valid;
int nrepeat;
char msgstr[HYAKU];
char msgbuf[HYAKU];
int msgcol;

/* lastcom takes as its value 0 or one of the following */
#define KILLAFT 1
#define KILLBFR 2
#define MOVELINE 3

#define EUC 0
#define SJIS 1
#define JISASCII 2
#define SJIS1 4
#define JISKANJI 5
#define JISESC 6

char tyibuf[10];
int tyiptr;
int tyimode;
int tyomode;
unsigned int tyoSJISbuf;

char finbuf[10];
int finptr;
int finmode;
int foutmode;
unsigned int foutSJISbuf;

char klog[KLOGlen];
int klogindex[10];
int klogp;
int kloglimit;
int cklog;
int replayp;
int replayindex;
int nreplay;

char cfile[HYAKU];
char fname[HYAKU];
char internal_cfile[HYAKU];
char internal_fname[HYAKU];
int fnamemode;
int readonly;
#define readonly_check if(readonly){warn("Buffer is readonly."); return;}
#define readonly_check_int if(readonly){warn("Buffer is readonly."); return 0;}

char cdir[HYAKU];

int marklist[JUU];
int markp;

/* basic display operations */

#ifdef unix

char CSthrow_cursor[] = "\033[%2+1%;%1+1%H";
char CSclear_display[] = "\033[2J";
char CSclear_to_eol[] = "\033[K";
char CSring_bell[] = "\007";
char CSbegin_reverse_video[] = "\033[7m";
char CSend_reverse_video[] = "\033[0m";
char CSset_scroll_region[] = "\033[1;%1+1%r";
#ifdef MACHTEN
char CSinsert_nlines[] = "";
char CSdelete_nlines[] = "";
#else
char CSinsert_nlines[] = "\033[%1%L";
char CSdelete_nlines[] = "\033[%1%M";
#endif
char CSinsert_line[] = "";
char CSdelete_line[] = "";

#else

#define CSSIZ 32

char CSthrow_cursor[CSSIZ];
char CSclear_display[CSSIZ];
char CSclear_to_eol[CSSIZ];
char CSring_bell[CSSIZ];
char CSbegin_reverse_video[CSSIZ];
char CSend_reverse_video[CSSIZ];
char CSset_scroll_region[CSSIZ];
char CSinsert_nlines[CSSIZ];
char CSdelete_nlines[CSSIZ];
char CSinsert_line[CSSIZ];
char CSdelete_line[CSSIZ];

#endif

#define nonnil(s) ((s)[0] != 0)

void tyo(character);

format0(s)
char *s;
{
	character c;
	while ((c = *(s++)) != 0) {
		if (c == '%') {
			tyo('%');
			if ((c = *(s++)) != '%')
				tyo(c);
		} else
			tyo(c);
	}
}
format1(s,x)
char *s; int x;
{
	character c;
	while ((c = *(s++)) != 0) {
		if (c == '%') {
			if ((c = *s++) == '%')
				tyo('%');
			else {
				int codep = FALSE, plusp = TRUE, d = 0;
				if (c == 'c') {
					codep = TRUE; s++;
				}
				if ((c = *(s++)) == '%');
				else {
					if (c == '-') plusp = FALSE;
					while ((c = *(s++)) != '%')
						d = d*10 + c - '0';
				}
				if (codep) tyo((plusp ? x + d : x - d));
				else tyoint((plusp ? x + d : x - d));
			}
		} else
			tyo(c);
	}
}
format2(s,x,y)
char *s; int x,y;
{
	character c;
	while ((c = *(s++)) != 0) {
		if (c == '%') {
			if ((c = *(s++)) == '%')
				tyo('%');
			else {
				int codep = FALSE, z, plusp = TRUE, d = 0;
				if (c == 'c') {
					codep = TRUE; c = *(s++);
				}
				z = (c == '1' ? x : y);
				if ((c = *(s++)) == '%');
				else {
					if (c == '-') plusp = FALSE;
					while ((c = *(s++)) != '%')
						d = d*10 + c - '0';
				}
				if (codep) tyo((plusp ? z + d : z - d));
				else tyoint((plusp ? z + d : z - d));
			}
		} else
			tyo(c);
	}
}

throw_cursor1(col,row) int col,row; {
	/*
	Move the cursor to the specified position (col,row),
	where the upper-left corner of the screen is (0,0) and
	the lower-right corner is (COLS-1, LINES-1).
	*/
	format2(CSthrow_cursor,col,row);
	pcol = col;
	prow = row;
}
#define throw_cursor(col,row) throw_cursor1((col), (row)+TEXTROW)
#define throw_MSGcursor(col) throw_cursor1((col), MSGROW)
#define throw_BANNERcursor(col) throw_cursor1((col), BANNERROW)

#define kanji(c) (((c) & 0x80) && ((c) != LS))

#ifdef TUS
#include "tus.inc"
#endif

#define clear() format0(CSclear_display)

#define clrtoeol() format0(CSclear_to_eol)

#define ring_bell() format0(CSring_bell);

insert_lines(row,n) int row,n; {
	throw_cursor(0,row);
	if (nonnil(CSinsert_nlines))
		format1(CSinsert_nlines,n);
	else if (nonnil(CSinsert_line))
		while (--n >= 0)
			format0(CSinsert_line);
#ifdef MACHTEN
        else
            while (--n >= 0)
                insertln();
#endif
}
delete_lines(row,n) int row,n; {
	throw_cursor(0,row);
	if (nonnil(CSdelete_nlines))
		format1(CSdelete_nlines,n);
	else if (nonnil(CSdelete_line))
		while (--n >= 0)
			format0(CSdelete_line);
#ifdef MACHTEN
        else
            while (--n >= 0)
                deleteln();
#endif
}

#ifdef unix
#ifdef NOusleep
#define sleep_a_little() sleep(1)
#else
#define sleep_a_little() usleep(300000)
#endif
#else
#ifdef pc
sleep_a_little() {
	int n=1000;
	while(!kbhit() && n--);
}
#else
sleep_a_little() {
	int n=1000;
	while(n--);
}
#endif
#endif

/* OS-dependent operations */

#ifdef unix
#define	raw_getchar() getchar()
#endif

#ifdef pc
/*
#define	raw_getchar() getchar()
*/
character raw_getchar() {
	union REGS inregs;
	union REGS outregs;
	do {
		inregs.h.ah = 6;
		inregs.h.dl = 0xff;
		intdos(&inregs, &outregs);
	} while (outregs.h.al == 0);
	return(outregs.h.al);
}
#endif

/* the rest of the program is independent of display/OS */
min(x,y) int x,y; {
	if (x > y) return(y);
	else return(x);
}

JISputc(c,fp,modep) unsigned int c; FILE *fp; int *modep;
{
	if (*modep == JISKANJI) {
		if (kanji(c)) putc((c & 0x7f),fp);
		else {
			putc(ESC,fp); putc('(',fp); putc('J',fp);
			putc(c,fp);
			*modep = JISASCII;
		}
	} else {
		if (kanji(c)) {
			putc(ESC,fp); putc('$',fp); putc('B',fp);
			putc((c & 0x7f),fp);
			*modep = JISKANJI;
		} else	
			putc(c,fp);
	}
}
SJISputc(c1,c2,fp) unsigned int c1,c2; FILE *fp;
{
	c1 &= 0x7f;
	c2 &= 0x7f;
	if ((c1 & 1) == 0) c2 += 0x9f-0x21;
	else {
		c2 += 0x40-0x21;
		if (c2 > 0x7e) c2++;
	}
	if (c1 < 0x5f) c1 = ((c1-1) >> 1) + 0x81-0x10;
	else c1 = ((c1-1) >> 1) + 0xe0-0x10;
	putc(c1,fp);
	putc(c2,fp);
}
#define nextcode (fp == NULL ? (raw_getchar() & 0x7f) : getc(fp))
character JISgetc(fp, modep, buf, ptr)
FILE *fp; int *modep; char buf[]; int *ptr;
{
	character c;
	if (*modep == JISESC) c = ESC; else c = nextcode;
	for(;; c= nextcode) {
		if (c != ESC) {
			if (*modep == JISKANJI) {
				buf[(*ptr)++] = nextcode | 0x80;
				return(c | 0x80);
			} else	
				return(c);
		}
		*modep = JISASCII; /* no Kanji code has ESC as the first byte */
		if ((c = nextcode) == '(') {
		    if ((c = nextcode) != 'H' && c != 'J' && c != 'B') {
				if (c == ESC) *modep = JISESC; else buf[(*ptr)++] = c;
				buf[(*ptr)++] = '(';
				return(ESC);
			}
		} else if (c == '$') {
			if ((c = nextcode) == 'B' || c == '@')
				*modep = JISKANJI;
			else {
				if (c == ESC) *modep = JISESC; else buf[(*ptr)++] = c;
				buf[(*ptr)++] = '$';
				return(ESC);
			}
		} else {
			if (c == ESC) *modep = JISESC; else buf[(*ptr)++] = c;
			return(ESC);
		}
	}
}
#undef nextcode
#define nextcode (fp == NULL ? raw_getchar() : getc(fp))
character SJISgetc(fp, buf, ptr) FILE *fp; char buf[]; int *ptr; {
	character c;
	unsigned int c1, c2;
	if ((c = nextcode) == EOF) return(c);
	c1 = c;
	if (!kanji(c1)) return(c1);
	c2 = nextcode;
	c1 = ((c1 - 0x81) << 1) + 0x21;
	if (c2 >= 0x9f) {
		c1++;
		c2 -= 0x9f-0x21;
	} else if (c2 >= 0x80) c2 -= 0x80-0x60;
	else c2 -= 0x40-0x21;
	buf[(*ptr)++] = c2 | 0x80;
	return(c1 | 0x80);
}
#undef nextcode

write_fname(to,from)
char *to, *from;
{
	unsigned int c,c1;	

	if (fnamemode == EUC)
		strcpy(to,from);
	else if (fnamemode == SJIS) {
		for (c = *from++; c != 0; c = *from++) {
			if (c & 0x80) {
				c &= 0x7f;
				c1 = (*from++) & 0x7f;
				if ((c & 1) == 0)
					c1 += 0x9f-0x21;
				else {
					c1 += 0x40-0x21;
					if (c1 > 0x7e) c1++;
				}
				if (c < 0x5f) c = ((c-1) >> 1) + 0x81-0x10;
				else c = ((c-1) >> 1) + 0xe0-0x10;
				*to++ = c;
				*to++ = c1;
				}
			else
				*to++ = c;
		}
		*to = 0;
	} else {
		for (c = *from++; c != 0; c = *from++) {
			if (c & 0x80) {
				*to++ = ESC; *to++ = '('; *to++ = 'J';
				do {
					*to++ = c & 0x7f;
					*to++ = (*from++) & 0x7f;
					c = *from++;
				} while (c & 0x80);
				*to++ = ESC; *to++ = '$'; *to++ = 'B';
				if (c == 0) break;
			}
			*to++ = c;
		}
		*to = 0;
	}
}

read_fname(to,from)
char *to, *from;
{
	unsigned int c,c1;

	if (fnamemode == EUC)
		strcpy(to,from);
	else if (fnamemode == SJIS) {
		for (c = *from++; c != 0; c = *from++) {
			if (c & 0x80) {
				c = ((c - 0x81) << 1) + 0x21;				
				if ((c1 = *from++) == 0)
					break;
				if (c1 >= 0x9f) {
					c++;
					c1 -= 0x9f-0x21;
				} else if (c1 >= 0x80)
					c1 -= 0x80-0x60;
				else
					c1 -= 0x40-0x21;
				*to++ = c | 0x80;
				*to++ = c1 | 0x80;
			} else
				*to++ = c;
		}
		*to = 0;
	} else {
		for (c = *from++; c != 0; c = *from++)
			if (c==ESC && *from=='$' && (from[1]=='B'||from[1]=='@')) {
				from += 2;
				for (c = *from++; c != 0; c = *from++)
					if (c==ESC && *from == '('
					           && (from[1]=='H'||from[1]=='J'||from[1]=='B'))
					{
						from += 2;
						break;
					} else if (*from == 0) {
						c = 0;
						break;
					} else {
						*to++ = c | 0x80;
						*to++ = (*from++) | 0x80;
					}
				if (c == 0) break;
			} else
				*to++ = c;
		*to = 0;
	}
}

close_output_file(fp) FILE *fp;
{
	if (foutmode == JISKANJI) {
		putc(ESC,fp); putc('(',fp); putc('J',fp);
		foutmode = JISASCII;
	}
	fclose(fp);
}

close_input_file(fp) FILE *fp;
{
	if (foutmode == JISKANJI || foutmode == JISESC)
		foutmode = JISASCII;
	fclose(fp);
}

void fout(c,fp) character c; FILE *fp; {
	switch (foutmode) {
	case EUC: putc(c,fp); return;
	case SJIS:
		if (!kanji(c)) {
			putc(c,fp); return;
		} else {
			foutSJISbuf = (c & 0x7f); foutmode = SJIS1; return;
		}
	case SJIS1:
		SJISputc(foutSJISbuf, (unsigned int) c, fp);
		foutmode = SJIS; return;
	default: JISputc((unsigned int) c, fp, &foutmode);
	}
}
void tyo(c) character c; {
	switch (tyomode) {
	case EUC: putchar(c); return;
	case SJIS:
		if (!kanji(c)) {
			putchar(c); return;
		} else {
			tyoSJISbuf = (c & 0x7f); tyomode = SJIS1; return;
		}
	case SJIS1:
		SJISputc(tyoSJISbuf, (unsigned int) c, stdout);
		tyomode = SJIS; return;
	default: JISputc((unsigned int) c, stdout, &tyomode);
	}
}
tyostr(s) char *s; {
	character c;
	while ((c = *(s++)) != 0) tyo(c);
}

char tyodigits[10];
tyoint(n) int n; {
	int i = 0;
	if (n < 0) {
		tyo('-'); n = -n;
	}
	while (n >= 10) {
		tyodigits[i++] = (n % 10) + '0';
		n = n / 10;
	}
	tyo(n + '0');
	while (i > 0) tyo(tyodigits[--i]);
}

character fin(fp) FILE *fp; {
	if (finptr > 0) return(finbuf[--finptr]);
	else {
		switch (finmode) {
		case EUC: return(getc(fp));
		case SJIS: return(SJISgetc(fp, finbuf, &finptr));
		default: return(JISgetc(fp, &finmode, finbuf, &finptr));
		}
	}
}
#ifdef kana
#include "kana.c"
#else
character tyi() {
	if (tyiptr > 0) return(tyibuf[--tyiptr]);
	else {
		switch (tyimode) {
		case EUC: return(raw_getchar());
		case SJIS: return(SJISgetc(NULL, tyibuf, &tyiptr));
		default: return(JISgetc(NULL, &tyimode, tyibuf, &tyiptr));
		}
	}
}
#endif
/* inserts a Kanji char */
ins2(c,c1) character c,c1; {
	if (ccol >= MAXCOL-1) {
		buf[++prev] = LS;
		ccol = 2;
		crow++;
	} else ccol += 2;
	buf[++prev] = c;
	buf[++prev] = c1;
}
/* inserts an ASCII char */
ins1(c) character c; {
	if (SP <= c && c < Rubout) {
		if (ccol >= MAXCOL) {
			buf[++prev] = LS;
			ccol = 1;
			crow++;
		} else ccol++;
	} else if (c == NL || c == LS) {
		ccol = 0;
		crow++;
	} else if (c == TAB) {
		if (ccol >= MAXCOL) {
			buf[++prev] = LS;
			ccol = TABSPACE;
			crow++;
		} else {
			ccol = ccol+TABSPACE-(ccol%TABSPACE);
			if (ccol > MAXCOL) ccol = MAXCOL;
		}
	} else {
		if (ccol >= MAXCOL-1) {
			buf[++prev] = LS;
			ccol = 2;
			crow++;
		} else ccol += 2;
	}
	buf[++prev] = c;
}
no_space_to_ins(c) character c; {
	if (kanji(c))
		return (ccol >= MAXCOL-1);
	else if ((SP <= c && c < Rubout) || c == NL || c == TAB)
		return (ccol >= MAXCOL);
	else
		return (ccol >= MAXCOL-1);
}
adjust_LS() {
	if (buf[next] == LS) {
		buf[++prev] = LS;
		next++;
		crow++; ccol = 0;
	}
}	
bwdcol(col,c) int col; character c; {
	if (kanji(c)) return(col-1);
	else if (c < SP) return(col-2);
	else if (c < Rubout) return(col-1);
	else return(col-2);
}
wt(c) character c; {
	int col;
	if (kanji(c)) {
		tyo(c); pcol++;
	}
	else if (c < SP) {
		if (c == TAB) {
			for(col = (pcol+TABSPACE-(pcol%TABSPACE)); pcol < col; pcol++)
				tyo(SP);
		} else {
			tyo('^'); tyo('A'-1+c); pcol += 2;
		}
	} else if (c < Rubout) {
		tyo(c); pcol++;
	} else {
		tyostr("^?"); pcol += 2;
	}
}
findcol() {
	character c;
	int next0 = next;
	for (c = buf[prev]; c != NL && c != LS; c = buf[--prev])
		buf[--next] = c;
	ccol = 0;
	while (next < next0) {
		c = buf[next++];
		if (kanji(c)) ins2(c,buf[next++]);
		else ins1(c);
	}
}
int valid_fwd_pos(i) int i; {
	int c = buf[i];
	if (!kanji(c) || c == LS) return(TRUE);
	else if (!kanji(c = buf[i-1]) || c == LS) return(TRUE);
	else
		for (;;) {
			if (++i >= BEND || !kanji(c = buf[i]) || c == LS) return(FALSE);
			if (++i >= BEND || !kanji(c = buf[i]) || c == LS) return(TRUE);
		}
}
int valid_bwd_pos(i) int i; {
	int c = buf[i];
	if (!kanji(c) || c == LS) return(TRUE);
	else if (!kanji(c = buf[i+1]) || c == LS) return(TRUE);
	else
		for (;;) {
			if (--i<=BSTART || !kanji(c = buf[i]) || c == LS) return(FALSE);
			if (--i<=BSTART || !kanji(c = buf[i]) || c == LS) return(TRUE);
		}
}
int simple_transpose() {
	void modified();
	character c1 = buf[prev], c2 = buf[next];
	readonly_check_int
	if (kanji(c2)) {
		if (kanji(c1)) {
			if (buf[prev-1] == c2 && c1 == buf[next+1]) return(TRUE);
			else {
				buf[prev] = buf[next+1];
				buf[next] = buf[prev-1];
				buf[next+1] = c1;
				buf[prev-1] = c2;
				throw_cursor(ccol-2,crow);
				tyo(c2); tyo(buf[prev]);
				tyo(buf[next]); tyo(c1);
				throw_cursor(ccol,crow);
				modified();
				return(TRUE);
			}
		} else if (c1 == LS &&
		           buf[prev-2] == c2 && buf[prev-1] == buf[next+1])
			return(TRUE);
		else return(FALSE);
	} else if (kanji(c1)) return(FALSE);
	else if (c1 == c2) return(TRUE);
	else if (c1 == LS && buf[prev-1] == c2) return(TRUE);
	else if (c1 >= SP && c2 >= SP && c1 < Rubout && c2 < Rubout) {
		buf[prev] = c2; buf[next] = c1;
		throw_cursor(ccol-1,crow);
		tyo(c2); tyo(c1);
		throw_cursor(ccol,crow);
		modified();
		return(TRUE);
	} else return(FALSE);
}
refresh_screen() {
	fflush(stdout);
}
erase_line(col,row) int col,row; {
	throw_cursor(col,row);
	clrtoeol();
}
writechar(c) character c; {
	if (kanji(c)) {
		tyo(c); msgcol++;
	} else if (c < SP) {
		tyo('^'); tyo('A'-1+c); msgcol += 2;
	} else if (c < Rubout) {
		tyo(c); msgcol++;
	} else {
		tyostr("^?"); msgcol += 2;
	}
}
writestr(s) char *s; {
	character c;
	int i = 0;
	while ((c = s[i++]) != 0) writechar(c);
}
int writedel(p) int p; {
	character c = msgbuf[--p];

	if (kanji(c)) {
		p--; msgcol -= 2;
	} else if (c < SP) msgcol -= 2;
	else if (c < Rubout) msgcol--;
	else msgcol -= 2;

	throw_MSGcursor(msgcol);
	clrtoeol();
	return(p);
}
msg(s) char *s; {
	throw_MSGcursor(0);
	if (msgcol != 0) clrtoeol();
	msgcol = 0;
	writestr(s);
	throw_cursor(ccol,crow);
	refresh_screen();
}
addmsg(s) char *s; {
	throw_MSGcursor(msgcol);
	writestr(s);
	throw_cursor(ccol,crow);
	refresh_screen();
}
msgchar(c) character c; {
	throw_MSGcursor(msgcol);
	writechar(c);
	throw_cursor(ccol,crow);
	refresh_screen();
}
#define msg1(s,x) SPRINTF1(msgstr,(s),(x)); msg(msgstr)
#define addmsg1(s,x) SPRINTF1(msgstr,(s),(x)); addmsg(msgstr)
#define msg2(s,x,y) SPRINTF2(msgstr,(s),(x),(y)); msg(msgstr)
clearmsg() {
	throw_MSGcursor(0);
	clrtoeol();
	msgcol = 0;
	throw_cursor(ccol,crow);
	refresh_screen();
}
warn(s) char *s; {
	ring_bell();
	replayp = -1;
	msg(s);
}
#define warn1(s,x) SPRINTF1(msgstr,(s),(x)); warn(msgstr)
#define unbound() warn("key unbound")
#define aborted() warn("aborted")
#define noSexpr() warn("no S-expression found")

character keyin() {
	void end_klog1();
	character c;
	if (replayp < 0) c = tyi();
	else if ((c = klog[replayp++]) == 0) {
		if ((--nreplay) > 0) {
			replayp = klogindex[replayindex];
			c = klog[replayp++];
		} else {
			replayp = -1;
			c = tyi();
		}
	}
	if (klogp >= 0) {
		klog[klogp++] = c;
		if (klogp >= kloglimit) {
			warn("Log buffer full.  ");
			klogp--;
			end_klog1();
		}
	}
	return(c);
}
unkeyin(c) character c; {
	if (replayp >= 0) replayp--;
	else tyibuf[tyiptr++] = c;
	if (klogp >= 0) klogp--;
}
void banner();
void begin_klog() {
	int n, i, j;
	if (klogp >= 0) {
		warn("Already logging"); return;
	}
	if (readmsg("Macro number (0-9):")) {
		aborted(); return;
	}
	n = msgbuf[0] - '0';
	if (n < 0 || n > 9 || msgbuf[1] != 0) {
		warn("Bad macro number"); return;
	}
	if ((j = klogindex[0])==0 && klog[klogindex[n]]==0) {
		warn("Log buffer already full"); return;
	}
	for (i = 0; i <= n; i++)
		klogindex[i] -= j;
	klogp = klogindex[n];
	if (n == 9) kloglimit = KLOGlen;
	else kloglimit = klogindex[n+1];
	for (i = 0; i < klogp;)
		klog[i++] = klog[j++];
	cklog = n;
	msg1("Starts logging for macro %d", cklog);
	banner();
}
void end_klog1() {
	int i;
	klog[klogp] = 0;
	for (i = 0; i <= cklog; i++)
		klogindex[i] += kloglimit - klogp - 1;
	do {
		klog[--kloglimit] = klog[klogp--];
	} while (klogp >= 0);
	if (klog[klogindex[cklog]] == 0) {
		addmsg1("Macro %d undefined", cklog);
	} else {
		addmsg1("Ends logging for macro %d", cklog);
	}
	cklog = -1;
	banner();
}
void end_klog() {
	if (klogp < 0) {
		warn("Not logging"); return;
	}
	klogp -= 2;
	end_klog1();
}
void begin_replay(n) int n; {
	if (klogp >= 0) {
		klogp -= 2;
		if (cklog == n) {
			warn("Can't execute the macro while logging for it.");
			return;
		}
	}
	if (klog[klogindex[n]] == 0) {
		unbound(); return;
	}
	nreplay = nrepeat;
	replayindex = n;
	replayp = klogindex[n];
}
void show_macro() {
	int n;
	if (readmsg("Macro number (0-9):")) {
		aborted(); return;
	}
	n = msgbuf[0] - '0';
	if (n < 0 || n > 9 || msgbuf[1] != 0) {
		warn("Bad macro number");
	} else if (n == cklog) {
		warn1("Can't show macro %d while logging for it", n);
	} else if (klog[klogindex[n]] == 0) {
		warn1("Macro %d not defined", n);
	} else {
		msg2("Macro %d: \"%s\"", n, klog+klogindex[n]);
	}
}
int readstr(s,deflt,dc) char *s, *deflt; character dc; {
	character c, c1;
	int p = 0, msgcol0;
	throw_MSGcursor(0);
	msgcol = 0;
	writestr(s); clrtoeol();
	msgcol0 = msgcol;
	refresh_screen();
	while ((c = keyin()) != ESC) {
		if (c == DEL || c == Rubout) {
			if (p > 0) p = writedel(p);
		} else if (c == CtrlG) return(1);
		else if (c == dc && p == 0) {
			writestr(deflt); break;
		} else if (c == CtrlU) {
			msgcol = msgcol0; throw_MSGcursor(msgcol); clrtoeol();
			p = 0;
		} else if (c == CtrlY) {
			int from, to, i;
			if (kbufleft) {from = 0; to = kbufp;}
			else {from = kbufp; to = KBUFlen;}
			if (from >= to) {ring_bell();}
			for (i = from; i < to; i++) {
				msgbuf[p++] = c = kbuf[i];
				if (p >= HYAKU) {
					ring_bell(); p--; break;
				} else if (kanji(c)) {
					msgbuf[p++] = c1 = kbuf[++i];
					if (p >= HYAKU) {
						ring_bell(); p -= 2; break;
					} else {
						writechar(c); writechar(c1);
					}
				} else writechar(c);
			}
		} else {
			if (c == CR) c = NL;
			else if (c == CtrlQ) c = keyin();
			msgbuf[p++] = c;
			if (p >= HYAKU) {
				ring_bell(); p--;
				if (kanji(c)) keyin();
			} else if (kanji(c)) {
				msgbuf[p++] = c1 = keyin();
				if (p >= HYAKU) {
					ring_bell(); p -= 2;
				} else {
					writechar(c); writechar(c1);
				}
			} else writechar(c);
		}
		refresh_screen();
	}
	writechar('"');
	msgbuf[p] = 0;
	throw_cursor(ccol,crow);
	refresh_screen();
	if (c == ESC) return(0);
	else return(-1);
}
int readmsg(s) char *s; {
	character c, c1;
	int p = 0, msgcol0;
	throw_MSGcursor(0);
	msgcol = 0;
	writestr(s); clrtoeol();
	msgcol0 = msgcol;
	refresh_screen();
	for (c = keyin(); c != NL && c != CR; c = keyin()) {
		if (c == DEL || c == Rubout) {
			if (p > 0) p = writedel(p);
		} else if (c == CtrlG) return(TRUE);
		else if (c == CtrlU) {
			msgcol = msgcol0; throw_MSGcursor(msgcol); clrtoeol();
			p = 0;
		} else if (kanji(c)) {
			c1 = keyin();
			msgbuf[p++] = c;
			if (p >= HYAKU) {
				ring_bell(); p--;
			} else {
				msgbuf[p++] = c1;
				if (p >= HYAKU) {
					ring_bell(); p -= 2;
				} else {
					writechar(c); writechar(c1);
				}
			}
		} else if (c >= SP && c < Rubout) {
			msgbuf[p++] = c;
			if (p >= HYAKU) {
				ring_bell(); p--;
			}
			else writechar(c);
		}
		refresh_screen();
	}
	msgbuf[p] = 0;
	throw_cursor(ccol,crow);
	refresh_screen();
	return(FALSE);
}
yesp() {
	if ((msgbuf[0] == 'y' || msgbuf[0] == 'Y') &&
		(msgbuf[1] == 'e' || msgbuf[1] == 'E') &&
		(msgbuf[2] == 's' || msgbuf[2] == 'S') &&
		(msgbuf[3] == 0))
		return(TRUE);
	else return(FALSE);
}
int query(s) char *s; {
	SPRINTF1(msgstr,"%s (yes or ^G): ",s);
	if (readmsg(msgstr)) return(FALSE);
	if (yesp()) return(TRUE);
	SPRINTF1(msgstr,"Answer yes or ^G.  %s: ",s);
	while (!readmsg(msgstr))
		if (yesp()) return(TRUE);
	return(FALSE);
}
void banner() {
	int i, n;
	throw_BANNERcursor(0);
	clrtoeol();
	if (nonnil(CSbegin_reverse_video))
		format0(CSbegin_reverse_video);
	tyo(readonly ? 'R' : '-');
	tyo(modp ? '*' : '-');
	tyo('-');

#ifdef kana
#define FNAMEPOS 12
	if (kana_mode) {tyo('\244'); tyo('\242');}
	else {tyo(' '); tyo('A');}
#else
#define FNAMEPOS 10
#endif

	for (n = 9; n >= 0; n--)
		if (n == cklog || klog[klogindex[n]] != 0) break;
	n++;
	for (i = 0; i < n; i++)
		if (i == cklog) {
			if (nonnil(CSend_reverse_video))
				format0(CSend_reverse_video);
			tyo('0'+i);
			if (nonnil(CSbegin_reverse_video))
				format0(CSbegin_reverse_video);
		} else if (klog[klogindex[i]] == 0) tyo('-');
		else tyo('0'+i);
	tyostr(" MINE: ");
	for (i = COLS - FNAMEPOS - n; i > 0; i--) tyo('=');
	throw_BANNERcursor(FNAMEPOS + n);
	if (internal_cfile[0] == 0) tyostr("[no file] ");
	else {
		tyostr(internal_cfile); tyo(SP);
	}
	if (cdir[0] != 0) {
		tyostr(cdir);
#ifdef pc
		tyo('\\');
#else
		tyo('/');
#endif
		tyo(SP);
	}
	if (nonnil(CSend_reverse_video))
		format0(CSend_reverse_video);
	throw_cursor(ccol,crow);
}
void modified() {
	if (readonly)
		warn("Bug in MINE: readonly buffer not protected.  Contact Taiichi.");
	if (!modp) {
		modp = TRUE;
		throw_BANNERcursor(0);
		if (nonnil(CSbegin_reverse_video))
			format0(CSbegin_reverse_video);
		tyostr("-*-");
		if (nonnil(CSend_reverse_video))
			format0(CSend_reverse_video);
		throw_cursor(ccol,crow);
	}
}
void unmodified() {
	if (modp) {
		modp = FALSE;
		throw_BANNERcursor(0);
		if (nonnil(CSbegin_reverse_video))
			format0(CSbegin_reverse_video);
		tyostr("---");
		if (nonnil(CSend_reverse_video))
			format0(CSend_reverse_video);
		throw_cursor(ccol,crow);
	}
}
int draw_line(col,row,i) int col,row,i; {
	character c;
	erase_line(col,row);
	if (i <= prev) {
		for (; i <= prev; i++) {
			if ((c = buf[i]) == NL) return(i);
			else if (c == LS) {
				if (buf[i-1] == TAB) throw_cursor(MAXCOL,row);
				wt('!'); return(i);
			} else wt(c);
		}
		i = next;
	}
	for (c = buf[i]; c != NL && c != LS; c = buf[++i]) wt(c);
	if (c == LS) {
		if (buf[i-1] == TAB) throw_cursor(MAXCOL,row);
		wt('!');
	}
	return(i);
}
void redraw(i) int i; {
	character c;
	int row;
	for (row = 0; row <= MAXROW; row++) {
		if (i >= BEND) erase_line(0,row);
		else i = draw_line(0, row, (i == prev ? next : i+1));
	}
	throw_cursor(ccol,crow);
}
void center() {
	character c;
	int i;
	for (crow = 0, i = tol(prev); crow < CENTERROW && i > BSTART;
	     crow++, i = tol(i-1));
		redraw(i);
}
int
change_line_length() {
	character c;
	int next0 = next;
	int prev0 = BEND;
	while (prev > BSTART) buf[--next] = buf[prev--];

	while (next < BEND) {
		if (next == next0) prev0 = prev;
		if (check_buf1()) return(TRUE);
		if ((c = buf[next++]) != LS) {
			if (kanji(c)) ins2(c,buf[next++]);
			else ins1(c);
		}
	}
	if (prev0 != BEND) {
		while (prev > prev0) buf[--next] = buf[prev--];
		adjust_LS();
	}
	return(FALSE);
}
void set_display_parameters() {
	MAXCOL = COLS - 2;
	MAXROW = LINES - 3;
	TEXTROW = 0;
	BANNERROW = LINES - 2;
	MSGROW = LINES - 1;
	CENTERROW = (int)(LINES/2 - 1);
}

void redisplay() {
#ifdef unix
	if (!fixed_window) {
		int oldLINES = LINES;
		int oldCOLS = COLS;
		int oldMAXCOL = MAXCOL;

# ifdef __CYGWIN32__
                LINES = atoi(getenv("LINES"));
                COLS = atoi(getenv("COLUMNS"));
# else
		tgetent(termcap_bp, tty_name);
		LINES = tgetnum("li");
		COLS = tgetnum("co");
# endif
		set_display_parameters();
		if (MAXROW < 10 || MAXCOL < 10) {
			warn("Too small window.");
			LINES = oldLINES;
			COLS = oldCOLS;
			set_display_parameters();
			return;
		}
		if (MAXCOL != oldMAXCOL)
			if (change_line_length()) {
				LINES = oldLINES;
				COLS = oldCOLS;
				set_display_parameters();
				change_line_length();
				return;
			}
	}
#endif
	center();
	banner();
	clearmsg();
}
void setcsr() {
	if (crow < 0 || crow > MAXROW) center();
	else throw_cursor(ccol,crow);
}
void reform(inc,dec) int inc,dec; {
	character c;
	int ccol0 = ccol, crow0 = crow, prev0 = prev;

	for (c = buf[next]; c != NL; c = buf[++next]) {
		if (c == LS)
			if (no_space_to_ins(buf[next+1]))
				break;
			else dec++;
		else if (kanji(c)) ins2(c,buf[++next]);
		else ins1(c);
	}
	while (prev > prev0) buf[--next] = buf[prev--];
	Ginc = crow - crow0 + inc; Gdec = dec;
	crow = crow0; ccol = ccol0;
	adjust_LS();
}

void redraw1(col,row,i) int col,row,i; {
  if (nonnil(CSinsert_nlines) || nonnil(CSinsert_line)) {
	int n, inc = min(Ginc,MAXROW-row), dec = min(Gdec,MAXROW-row);
	if (i > prev) i = next;
	if (row < 0 || crow > MAXROW) center();
	else {
		i = draw_line(col,row++,i);
		if (inc > dec) {
			if (nonnil(CSset_scroll_region))
				insert_lines(row, inc-dec);
			else {
				erase_line(0,BANNERROW);
				insert_lines(row, inc-dec);
				banner();
				clearmsg();
			}
		} else if (inc < dec) {
			if (nonnil(CSset_scroll_region))
				delete_lines(row, dec-inc);
			else {
				clearmsg();
				erase_line(0,BANNERROW);
				delete_lines(row, dec-inc);
				banner();
			}
		}
		for (n = inc; n > 0; n--) {
			i = draw_line(0, row++, (i == prev ? next : i+1));
		}
		if (inc < dec) {
			for (; row <= MAXROW-(dec-inc); row++) {
				if (i >= BEND) break;
				i = eol(i+1);
			}
			for (; row <= MAXROW; row++) {
				if (i >= BEND) break;
				i = draw_line(0, row, (i == prev ? next : i+1));
			}
		}
		throw_cursor(ccol,crow);
	}
  } else {
	int n, inc = Ginc, dec = Gdec;
	if (i > prev) i = next;
	if (row < 0 || crow > MAXROW) center();
	else {
		i = draw_line(col,row,i);
		if (inc == dec) n = min(inc, MAXROW-row);
		else n = MAXROW-row;
		for (; n > 0; n--) {
			if (i >= BEND) break;
			i = draw_line(0, ++row, (i == prev ? next : i+1));
		}
		if (inc < dec && row < MAXROW)
			for (n = min(dec - inc, MAXROW-row); n > 0; n--)
				erase_line(0, ++row);
		throw_cursor(ccol,crow);
	}
  }
}

int check_buf(n) int n; {
	if (prev+n+(n/MAXCOL)+GETA >= next) {
		warn("buffer full"); return(TRUE);
	} else return(FALSE);
}
int check_buf1() {
	if (prev+GETA+1 >= next) {
		warn("buffer full"); return(TRUE);
	} else return(FALSE);
}
void delaft(i,drawp) int i,drawp; {
	character c;
	int dec = 0;

	if (lastcom != KILLAFT) kbufp = 0;
	if (i-next >= KBUFlen-kbufp) {
		if (query("kill-buffer full.  kill any way?")) {
			while (next < i) {
				c = buf[next++];
				if (c == LS || c == NL) dec++;
			}
		} else {
			aborted(); return;
		}
	} else {
		while (next < i) {
			c = buf[next++];
			if (c == LS) dec++;
			else {
				kbuf[kbufp++] = c;
				if (c == NL) dec++;
			}
		}
	}
	reform(0,dec);
	if (drawp) redraw1(ccol,crow,next);
	kbufleft = TRUE;
	lastcom = KILLAFT; lastcom_valid = TRUE;
	modified();
}
void delbfr(i,drawp) int i,drawp; {
	character c;
	int dec = 0;

	if (lastcom != KILLBFR) kbufp = KBUFlen;
	if (prev-i >= kbufp) {
		if (query("kill-buffer full.  kill any way?")) {
			while (prev > i) {
				c = buf[prev--];
				if (c == LS || c == NL) {
					crow++; dec++;
				}
			}
		} else {
			aborted(); return;
		}
	} else {
		while (prev > i) {
			c = buf[prev--];
			if (c == LS) {
				crow--; dec++;
			} else {
				kbuf[--kbufp] = c;
				if (c == NL) {
					crow--; dec++;
				}
			}
		}
	}
	findcol();
	reform(0,dec);
	if (drawp) redraw1(ccol,crow,next);
	kbufleft = FALSE;
	lastcom = KILLBFR; lastcom_valid = TRUE;
	modified();
}
void load_file(ro) int ro; {
	character c;
	FILE *fp;
	int failed = FALSE;

	prev = mark = BSTART; next = BEND;
	crow = ccol = 0;
	if ((fp = fopen(cfile, "r")) == NULL) msg("New file");
	else {
		msg1("Loading %s ... ", internal_cfile);
		while ((c = fin(fp)) != EOF) {
			if (check_buf1()) {
				failed = TRUE; break;
			}
			if (kanji(c)) ins2(c,fin(fp));
			else ins1(c);
		}
		close_input_file(fp);
		if (failed) {
			prev = BSTART;
			internal_cfile[0] = 0;
		} else {
			mark = prev;
			while (prev > BSTART) buf[--next] = buf[prev--];
			addmsg("Done");
			readonly = ro;
		}
	}
	crow = ccol = ccol1 = 0;
	redraw(BSTART);
	banner();
}
void visit_file(ro) int ro; {
	FILE *fp;
	if (readmsg(ro ? "read-file File name: " : "visit-file File name: ")) {
		aborted(); return;
	}
	strcpy(internal_fname,msgbuf);
	write_fname(fname,msgbuf);
	if (modp) {
		if (query("Discard changes?")) unmodified();
		else {
			aborted(); return;
		}
	}
	if (internal_fname[0] != 0) {
		strcpy(internal_cfile,internal_fname);
		strcpy(cfile,fname);
	}
	load_file(ro);
}
void insert_file() {
	character c;
	FILE *fp;
	int ccol0 = ccol, crow0 = crow, prev0 = prev, failed = FALSE;
	readonly_check
	if (readmsg("insert-file File name: ")) {
		aborted(); return;
	}
	strcpy(internal_fname,msgbuf);
	write_fname(fname,msgbuf);
	if ((fp = fopen(fname, "r")) == NULL) {
		warn1("File %s not found", internal_fname);
		return;
	}
	msg1("Reading %s ... ", internal_fname);
	while ((c = fin(fp)) != EOF) {
		if (check_buf1()) {
			failed = TRUE; break;
		}
		if (kanji(c)) ins2(c,fin(fp));
		else ins1(c);
	}
	close_input_file(fp);
	if (failed) {
		ccol = ccol0; crow = crow0; prev = prev0;
	} else {
		addmsg("Done");
		if (prev > prev0) {
			modified();
			reform(crow-crow0,0);
			redraw1(ccol0, crow0, prev0+1);
		}
	}
}
void set_tab() {
	int n = 0;
	if (readmsg("Tab space: ")) {
		aborted(); return;
	}
	sscanf(msgbuf, "%d", &n);
	if (n <= 0) {
		warn("Bad tab space");
	} else if (n != TABSPACE) {
		int old_tabspace = TABSPACE;
		TABSPACE = n;
		if (change_line_length()) {
			TABSPACE = old_tabspace;
			change_line_length();
		} else
			center();
	}
}
void count_lines() {
	character c = NL;
	int count = 0;
	int i, cursor_line;

	for (i = BSTART+1; i <= prev; i++)
		if ((c = buf[i]) == NL) count++;
	cursor_line = count+1;
	for (i = next; i < BEND; i++)
		if ((c = buf[i]) == NL) count++;
	if (c != NL) count++;

	if (count == 0) { msg("no line"); }
	else if (count == 1) { msg("total 1 line"); }
	else if (cursor_line > count) {
		msg1("total %d lines, cursor at the end", count);
	} else { msg2("total %d lines, cursor at line %d", count, cursor_line); }
}

void count_bytes() {
	character c = NL;
	int count = 0;
	int i, cursor_position;

	for (i = BSTART+1; i <= prev; i++)
		if ((c = buf[i]) == LS) count++;
	cursor_position = prev - BSTART - count;
	count = 0;
	for (i = next; i < BEND; i++)
		if ((c = buf[i]) == LS) count++;
	count = cursor_position + BEND - next - count;

	if (count == 0) { msg("no byte"); }
	else if (count == 1) { msg("total 1 byte"); }
	else if (cursor_position == count) {
		msg1("total %d bytes, cursor at the end", count); }
	else {
		msg2("total %d bytes, cursor at byte %d",
		     count, cursor_position + 1);
	}
}

void dump_buf(from,to,fp) int from,to; FILE *fp; {
	character c;
	for (; from <= to; from++)
		if ((c = buf[from]) != LS) fout(c,fp);
}

save_file(mode) int mode; {
	FILE *fp;
	readonly_check_int
	if (internal_cfile[0] == 0) {
		if (readmsg((mode ? "save-file File name: " :
		                    "send-file File name: "))) {
			aborted(); return(FALSE);
		} else {
			strcpy(internal_fname,msgbuf);
			write_fname(fname, msgbuf);
			if ((fp = fopen(fname, "r")) != NULL) {
				fclose(fp);
				SPRINTF1(tmpstr,"Overwrite %s?", internal_fname);
				if (!query(tmpstr)) {
					aborted(); return(FALSE);
				}
			}
			strcpy(internal_cfile, internal_fname);
			strcpy(cfile, fname);
			banner();
		}
	} else if (!modp) {
		if (mode) warn("no change"); return(TRUE);
	}
	if ((fp = fopen(cfile, "w+")) == NULL) {
		warn1("cannot open %s", internal_cfile);
		return(FALSE);
	}
	msg1("Saving %s ... ", internal_cfile);
	dump_buf(BSTART+1,prev,fp);
	dump_buf(next,BEND-1,fp);
	close_output_file(fp);
	addmsg("Done");
	unmodified();
	return(TRUE);
}
void save_region() {
	character c;
	FILE *fp;
	int i;
	if (readmsg("save-region File name: ")) {
		aborted(); return;
	}
	strcpy(internal_fname,msgbuf);
	write_fname(fname,msgbuf);
	if ((fp = fopen(fname, "r")) != NULL) {
		fclose(fp);
		SPRINTF1(tmpstr,"Overwrite %s?",internal_fname);
		if (!query(tmpstr)) {
			aborted(); return;
		}
	}
	if ((fp = fopen(fname, "w+")) == NULL) {
		warn1("cannot open %s", internal_fname); return;
	}
	msg1("Writing %s ... ", internal_fname);
	if (mark == prev);
	else if (mark < prev) {
		if (!valid_bwd_pos(mark)) mark--;
		dump_buf(mark+1,prev,fp);
	} else {
		i = min(mark - prev + next - 1, BEND);
		if (i != next && !valid_fwd_pos(i)) i++;
		dump_buf(next,i,fp);
	}
	close_output_file(fp);
	addmsg("Done");
}
void insert(c) character c; {
	int ccol0 = ccol, crow0 = crow, prev0 = prev;
	readonly_check
	if (check_buf(nrepeat)) return;
	while ((nrepeat--) > 0) ins1(c);
	reform(crow-crow0,0);
	redraw1(ccol0, crow0, prev0+1);
	modified();
}
void insert2(c,c1) character c,c1; {
	int ccol0 = ccol, crow0 = crow, prev0 = prev;
	if (check_buf(nrepeat*2)) return;
	while ((nrepeat--) > 0) ins2(c,c1);
	reform(crow-crow0,0);
	redraw1(ccol0, crow0, prev0+1);
	modified();
}
void fwd_char() {
	character c;
	while ((nrepeat--) > 0) {
		if (next >= BEND) {
			warn("no next character"); break;
		}
		c = buf[next++];
		if (kanji(c)) ins2(c,buf[next++]);
		else ins1(c);
		adjust_LS();
	}
	setcsr();
}
void bwd_char() {
	character c;
	while ((nrepeat--) > 0) {
		if (prev <= BSTART) {
			warn("no previous character"); break;
		}
		c = buf[--next] = buf[prev--];
		if (c == NL) crow--;
		else if (c == LS) {
			c = buf[--next] = buf[prev--];
			if (kanji(c)) buf[--next] = buf[prev--];
			crow--;
		} else if (kanji(c))
			buf[--next] = buf[prev--];
	}
	findcol();
	setcsr();
}
int alphanumeric(c) character c; {
	if ((c>='a' && c<='z') || (c>='A' && c<='Z') || (c>='0' && c<='9'))
		return(TRUE);
	else
		return(FALSE);
}
void fwdbuf(i) int i; {
	while (next < i) buf[++prev] = buf[next++];
	if (buf[next] == LS) {
		buf[++prev] = LS; next++;
	}
}
void bwdbuf(i) int i; {
	while (prev > i) buf[--next] = buf[prev--];
}
void fwdbufcsr(i) int i; {
	character c;
	while (next < i) {
		c = buf[next++];
		if (kanji(c)) ins2(c,buf[next++]);
		else ins1(c);
	}
	adjust_LS();
}
void bwdbufcsr(i) int i; {
	character c;
	while (prev > i) {
		c = buf[--next] = buf[prev--];
		if (c == NL || c == LS) {
			crow--;
		}
	}
	findcol();
	adjust_LS();
}
void fwd_wd() {
	character c;
	int i = next;
	while ((nrepeat--) > 0) {
		for (c = buf[i]; !alphanumeric(c); c = buf[++i])
			if (i >= BEND) {
				warn("no next word"); setcsr(); return;
			}
		for (c = buf[++i]; alphanumeric(c) || c == LS; c = buf[++i]);
		fwdbufcsr(i);
	}
	setcsr();
}
void bwd_wd() {
	character c;
	int i = prev;
	while ((nrepeat--) > 0) {
		for (c = buf[i]; !alphanumeric(c); c = buf[--i])
			if (i <= BSTART) {
				warn("no previous word"); setcsr(); return;
			}
		for (c = buf[--i]; alphanumeric(c) || c == LS; c = buf[--i]);
		if (buf[i+1] == LS) i++;
		bwdbufcsr(i);
	}
	setcsr();
}
void delete_next_wd() {
	character c;
	int i = next, j = next;
	readonly_check
	while ((nrepeat--) > 0) {
		for (c = buf[i]; !alphanumeric(c); c = buf[++i])
			if (i >= BEND) {
				warn("no next word"); break;
			}
		if (i >= BEND) break;
		for (c = buf[++i]; alphanumeric(c) || c == LS; c = buf[++i]);
		j = i;
	}
	if (j != next) delaft(j,TRUE);
}
void delete_prev_wd() {
	character c;
	int i = prev, j = prev;
	readonly_check
	while ((nrepeat--) > 0) {
		for (c = buf[i]; !alphanumeric(c); c = buf[--i])
			if (i <= BSTART) {
				warn("no previous word"); break;
			}
		if (i <= BSTART) break;
		for (c = buf[--i]; alphanumeric(c) || c == LS; c = buf[--i])
		if (buf[i+1] == LS) i++;
		j = i;
	}
	if (j != prev) delbfr(j,TRUE);
}
void updown_wd(dir) character dir; {
	character c;
	int i = next, flag = FALSE, flagor = FALSE;
	int ccol0, crow0, prev0, first, incdec;
	readonly_check
	while ((nrepeat--) > 0) {
		for (c = buf[i]; !alphanumeric(c); c = buf[++i])
			if (i >= BEND) {
				warn("no next word"); break;
			}
		if (i >= BEND) break;
		fwdbufcsr(i); ccol0 = ccol; crow0 = crow; prev0 = prev;
		first = TRUE; incdec = 0;
		for(;;) {
			if (c >= 'A' && c <= 'Z') {
				if (dir == 'L' || (dir == 'C' && !first)) {
					buf[i] = c - 'A' + 'a'; flag = TRUE;
				}
			} else if (c >= 'a' && c <= 'z') {
				if (dir == 'U' || (dir == 'C' && first)) {
					buf[i] = c - 'a' + 'A'; flag = TRUE;
				}
			}
			first = FALSE;
			if (alphanumeric(c = buf[++i]));
			else if (c == LS)
				if (alphanumeric(c = buf[++i])) incdec++;
				else break;
			else break;
		}
		fwdbufcsr(i);
		if (flag && crow <= MAXROW) {
			Ginc = Gdec = incdec;
			redraw1(ccol0, crow0, prev0+1);
			flagor = TRUE;
		}
	}
	if (crow > MAXROW) center();
	else setcsr();
	if (flagor) modified();
}
whitespace(c) character c; {
	if (c == SP || c == TAB || c == NL || c == LS) return(TRUE);
	else return(FALSE);
}
delimiter(c) character c; {
	if (c == SP || c == TAB || c == NL || c == '(' || c == ')' ||
		c == '{' || c == '}' || c == '[' || c == ']')
		return(TRUE);
	else return(FALSE);
}
fwdSexpr1(c,i) character c; int i; {
	int level;
	character open, close;
	switch(c) {
	case '(': close = ')'; break;
	case '{': close = '}'; break;
	case '[': close = ']'; break;
	case ')': case '}': case ']': return(i+1);
	default:
		for (c = buf[++i]; !delimiter(c); c = buf[++i]);
		return(i);
	}
	open = c; level = 0;
	for (c = buf[++i]; i < BEND; c = buf[++i]) {
		if (c == close) {
			if (level == 0) return(i+1);
			else level--;
		} else if (c == open) level++;
	}
	warn("no closing parenthesis");
	return(-1);
}
bwdSexpr1(c,i) character c; int i; {
	int level;
	character open, close;
	switch(c) {
	case ')': open = '('; break;
	case '}': open = '{'; break;
	case ']': open = '['; break;
	case '(': case '{': case '[': return(i-1);
	default:
		for (c = buf[--i]; !delimiter(c); c = buf[--i]);
		return(i);
	}
	close = c; level = 0;
	for (c = buf[--i]; i > BSTART; c = buf[--i]) {
		if (c == open) {
			if (level == 0) return(i-1);
			else level--;
		} else if (c == close) level++;
	}
	warn("no opening parenthesis");
	return(-1);
}
void insert_paren(c) character c; {
	int ccol0, crow0, prev0, next0, i;
	readonly_check
	if (check_buf(nrepeat)) return;
	while ((nrepeat--) > 0) {
		ccol0 = ccol; crow0 = crow; prev0 = prev;
		ins1(c);
		reform(crow-crow0,0);
		redraw1(ccol0, crow0, prev0+1);
		modified();
		if ((i = bwdSexpr1(c,prev0+1)) >= 0) {
			ccol0 = ccol; crow0 = crow; next0 = next;
			bwdbufcsr(i);
			if (crow < 0) {
				draw_line(0,MSGROW,tol(i)+1);
				throw_MSGcursor(ccol);
				refresh_screen();
				sleep_a_little();
				fwdbuf(next0);
				ccol = ccol0; crow = crow0;
				clearmsg();
			} else {
				setcsr(); refresh_screen();
				sleep_a_little();
				fwdbuf(next0);
				ccol = ccol0; crow = crow0;
				setcsr(); refresh_screen();
			}
		}
	}
}
void fwd_Sexpr() {
	character c;
	int i = next, j = next;
	while ((nrepeat--) > 0) {
		for (c = buf[i]; whitespace(c); c = buf[++i])
			if (i >= BEND) {
				noSexpr(); setcsr(); break;
			}
		if (i >= BEND) break;
		i = fwdSexpr1(c,i);
		if (i < 0) break;
		j = i;
	}
	if (j != next) {
		fwdbufcsr(j);
		setcsr();
	}
}
void bwd_Sexpr() {
	character c;
	int i = prev, j = prev;
	while ((nrepeat--) > 0) {
		for (c = buf[i]; whitespace(c); c = buf[--i])
			if (i <= BSTART) {
				noSexpr(); setcsr(); break;
			}
		if (i >= BEND) break;
		if ((i = bwdSexpr1(c,i)) < 0) break;
		j = i;
	}
	if (j != prev) {
		bwdbufcsr(j);
		setcsr();
	}
}
void delete_next_Sexpr() {
	character c;
	int i = next, j = next;
	readonly_check
	while ((nrepeat--) > 0) {
		for (c = buf[i]; whitespace(c); c = buf[++i])
			if (i >= BEND) {
				noSexpr(); break;
			}
		if (i >= BEND) break;
		if ((i = fwdSexpr1(c,i)) < 0) break;
		j = i;
	}
	if (j != next) delaft(j,TRUE);
}
void delete_prev_Sexpr() {
	character c;
	int i = prev, j = prev;
	readonly_check
	while ((nrepeat--) > 0) {
		for (c = buf[i]; whitespace(c); c = buf[--i])
			if (i <= BSTART) {
				noSexpr(); break;
			}
		if (i <= BSTART) break;
		if ((i = bwdSexpr1(c,i)) < 0) break;
		j = i;
	}
	if (j != prev) delbfr(i,TRUE);
}

void markS(i) {
	if (markp == 0) markp = JUU;
	marklist[--markp] = i;
}
int setmarks(i) int i; {
	character c;
	int level;
	markp = JUU;
	for (c = buf[i]; whitespace(c); c = buf[--i])
		if (i <= BSTART) return(i);
	if (c == '(') {
	     markS(i+1); return(i);
	} else if (c == ')') {
		level = 0;
		for (;;) {
			c = buf[--i];
			if (i <= BSTART) return(i);
			else if (c == '(') {
				if (level == 0) {
					markS(i); c = buf[--i]; break;
				} else level--;
			} else if (c == ')') level++;
		}
	} else {
		for (;;) {
			c = buf[--i];
			if (c == NL) {
				markS(i+1); return(i);
			} else if (c == '(') {
				markS(i+2); return(i);
			} else if (c == ')' || c == SP || c == TAB) {
				markS(i+1); break;
			}
		}
	}
	for (;;) {
		for (c = buf[i]; whitespace(c); c = buf[--i])
		if (c == NL) return(i);
		if (c == '(') return(i);
		else if (c == ')') {
			level = 0;
			for (;;) {
				c = buf[--i];
				if (c == NL) return(i);
				else if (c == '(') {
					if (level == 0) {
					markS(i); c = buf[--i]; break;
					} else level--;
				} else if (c == ')') level++;
			}
		} else {
			for (;;) {
				c = buf[--i];
				if (c == NL) {
					markS(i+1); return(i);
				} else if (c == '(') {
					markS(i+2); return(i);
				} else if (c == ')' || c == SP || c == TAB) {
					markS(i+1); break;
				}
			}
		}
	}
}
void indent(freshlinep) int freshlinep; {
	character c;
	int i = prev, inc = 0, dec = 0;
	int ccol0, crow0, prev0, mark;

	readonly_check
	if (freshlinep) {
		if (check_buf1()) return;
		ccol0 = ccol; crow0 = crow; prev0 = prev;
		buf[++prev] = NL;
		ccol = 0; crow++; inc++; i = prev;
	} else {
		for (i = prev, c = buf[i]; c != NL; c = buf[--i]);
		bwdbufcsr(i);
		ccol0 = ccol; crow0 = crow; prev0 = prev;
	}
	for (c = buf[next]; whitespace(c); c = buf[++next])
		if (c == NL) break;
		else if (c == LS) dec++;
	i = setmarks(i);
	for (c = buf[i]; c != NL; c = buf[--i]);
	i++;
	for (;;) {
		if (markp < JUU) {
			mark = marklist[markp++];
			while (i < mark) {
				if (check_buf1()) break;
				c = buf[i++];
				if (c >= SP && c < Rubout) {
					buf[++prev] = SP; ccol++;
				} else if (c == TAB) {
					buf[++prev] = c; ccol = ccol+TABSPACE-(ccol%TABSPACE);
				} else if (c == LS) {
					buf[++prev] = c; ccol = 0; crow++; inc++;
				} else {
					buf[++prev] = SP; buf[++prev] = SP; ccol += 2;
				}
			}
		}
		reform(inc,dec);
		redraw1(ccol0,crow0,prev0+1);
		refresh_screen();
		modified();
		if (markp >= JUU) return;
		if ((c = keyin()) != TAB) {
			unkeyin(c); return;
		}
		ccol0 = ccol; crow0 = crow; prev0 = prev;
		inc = dec = 0;
	}
}
void adjust_column() {
	character c;
	ccol = 0;
	while (ccol < ccol1) {
		c = buf[next];
		if (c == NL || c == LS) break;
		next++;
		if (kanji(c)) ins2(c,buf[next++]);
		else ins1(c);
	}
	if (buf[next] == LS) {
		c = buf[--next] = buf[prev--];
		if (kanji(c))
			buf[--next] = buf[prev--];
		findcol();
	}
	setcsr();
}
void fwd_real_ln(){
	character c;
	int i = next, j = 0, row = crow;
	if (lastcom != MOVELINE) {
		ccol1 = ccol; lastcom = MOVELINE;
	}
	lastcom_valid = TRUE;

	while ((nrepeat--) > 0) {
		while ((c = buf[i++]) != NL)
			if (c == LS) row++;
		row++;
		if (i >= BEND) {
			warn("no next line");
			fwdbuf(BEND);
			findcol();
			setcsr();
			return;
		}
		crow = row;
		j = i;
	}
	if (j > 0) {
		fwdbuf(j);
		adjust_column();
	}
}
void bwd_real_ln(){
	character c;
	int i = prev, f = FALSE, row = crow;
	if (lastcom != MOVELINE) {
		ccol1 = ccol; lastcom = MOVELINE;
	}
	lastcom_valid = TRUE;

	while ((nrepeat--) >= 0) {
		while((c = buf[i--]) != NL)
			if (c == LS) row--;
		row--;
		if (i <= BSTART && nrepeat >= 0) {
			warn("no previous line");
			bwdbuf(BSTART);
			ccol = 0;
			setcsr();
			return;
		}
		f = TRUE;
	}
	if (f) {
		crow = row+1;
		bwdbuf(i+1);
		adjust_column();
	}
}
void fwd_ln(){
	character c;
	int i, j = next;
	if (lastcom != MOVELINE) {
		ccol1 = ccol; lastcom = MOVELINE;
	}
	lastcom_valid = TRUE;

	while ((nrepeat--) > 0) {
		i = eol(j);
		if (i >= BEND) {
			warn("no next line");
			fwdbuf(BEND);
			findcol();
			setcsr();
			return;
		}
		j = i+1;
		crow++;
	}
	if (j > next) {
		fwdbuf(j);
		adjust_column();
	}
}
void bwd_ln(){
	character c;
	int i = tol(prev), crow0 = crow;
	if (lastcom != MOVELINE) {
		ccol1 = ccol; lastcom = MOVELINE;
	}
	lastcom_valid = TRUE;

	while ((nrepeat--) > 0) {
		if (i <= BSTART) {
			warn("no previous line");
			bwdbuf(BSTART);
			ccol = 0;
			setcsr();
			return;
		}
		i = tol(i-1);
		crow--;
	}
	if (crow != crow0) {
		bwdbuf(i);
		adjust_column();
	}
}
void goto_top() {
	character c;
	while (prev > BSTART) {
		c = buf[--next] = buf[prev--];
		if (c == NL || c == LS) crow--;
	}
	ccol = 0;
	setcsr();
}
void goto_bottom() {
	character c;
	while (next < BEND) {
		c = buf[++prev] = buf[next++];
		if (c == NL || c == LS) crow++;
	}
	findcol();
	setcsr();
}
int tol(i) int i; {
	character c;
	for (c = buf[i]; c != NL && c != LS; c = buf[--i]) ;
	return(i);
}
int eol(i) int i; {
	character c;
	for (c = buf[i]; c != NL && c != LS; c = buf[++i]) ;
	return(i);
}
void goto_real_tol() {
	character c;
	for (c = buf[prev]; c != NL; c = buf[--prev]) {
		buf[--next] = c;
		if (c == LS) crow--;
	}
	ccol = 0;
	setcsr();
}
void goto_real_eol() {
	character c;
	for (c = buf[next]; c != NL; c = buf[++next]) {
		if (kanji(c)) ins2(c,buf[++next]);
		else ins1(c);
	}
	setcsr();
}
void goto_tol() {
	character c;
	for (c = buf[prev]; c != NL && c != LS; c = buf[--prev]) {
		buf[--next] = c;
	}
	ccol = 0;
	setcsr();
}
void goto_eol() {
	character c;
	for (c = buf[next]; c != NL && c != LS; c = buf[++next]) {
		if (kanji(c)) ins2(c,buf[++next]);
		else ins1(c);
	}
	if (c == LS) {
		c = buf[--next] = buf[prev--];
		if (kanji(c)) {
			buf[--next] = buf[prev--];
		}
		findcol();
	}
	setcsr();
}
void fwd_pg() {
	int i = next-1, j = 0, row = crow;
	while ((nrepeat--) > 0) {
		for (; row <= MAXROW; row++)
			if ((i = eol(i+1)) >= BEND) {
				warn("no next page"); break;
			}
		if (i >= BEND) break;
		j = i;
		row = 2;
	}
	if (j != 0) {
		fwdbufcsr(j+1);
		ccol = 0; crow = 2;
		redraw(tol(tol(prev - 1)-1));
	}
	if (i >= BEND) {
		fwdbufcsr(BEND);
		throw_cursor(ccol,crow);
	}
}
void bwd_pg() {
	int i = prev+1, row = crow, f = FALSE;
	while ((nrepeat--) > 0) {
		for (; row >= 0; row--)
			if ((i = tol(i-1)) <= BSTART) {
				warn("no previous page"); break;
			}
		if (i <= BSTART) break;
		row = MAXROW - 2; f = TRUE;
	}
	if (f) {
		if (i <= BSTART) {
			bwdbuf(BSTART); ccol = crow = 0;
		} else {
			for (row = MAXROW - 2; row >= 0; row--)
				if ((i = tol(i-1)) <= BSTART) break;
			if (row <= 0) {
				bwdbuf(eol(eol(i+1)+1));
				ccol = 0; crow = 2;
			} else {
				bwdbuf(BSTART); ccol = crow = 0;
			}
		}
		redraw(i);
	}
}
void delete_prev_char(){
	character c;
	int crow0 = crow;
	readonly_check
	while ((nrepeat--) > 0) {
		if (prev <= BSTART) {
			warn("no previous character"); break;
		}
		c = buf[prev--];
		if (c == NL) crow--;
		else if (c == LS) {
			c = buf[prev--];
			if (kanji(c)) prev--;
			crow--; 
		} else if (kanji(c))
			prev--;
	}
	findcol();
	reform(0,crow0-crow);
	redraw1(ccol,crow,next);
	modified();
}
void delete_next_char() {
	character c;
	int dec = 0, next0 = next;
	readonly_check
	while ((nrepeat--) > 0) {
		if (next >= BEND) {
			warn("no next character"); break;
		}
		c = buf[next++];
		if (c == NL) dec++;
		else {
			if (kanji(c)) next++;
			if (buf[next] == LS) {
				next++; dec++;
			}
		}
	}
	if (next != next0) {
		reform(0,dec);
		redraw1(ccol,crow,next);
		modified();
	}
}
void kill_ln() {
	int i = next;
	readonly_check
	while ((nrepeat--) > 0) {
		if (buf[i] == NL) {
			if (i >= BEND) {
				warn("no next line"); break;
			}
			i++;
		} else while (buf[++i] != NL);
	}
	if (i != next) delaft(i,TRUE);
}
void set_mark() {
	mark = prev;
	msg("Mark set");
}
void exchange_mark() {
	int prev0 = prev, i;
	if (mark == prev) return;
	else if (mark < prev) {
		if (!valid_bwd_pos(mark)) mark--;
		if (buf[mark+1] == LS) mark++;
		bwdbufcsr(mark);
	} else {
		i = min(next + mark - prev, BEND);
		if (!valid_fwd_pos(i)) i++;
		if (buf[i] == LS) i++;
		fwdbufcsr(i);
	}
	setcsr();
	mark = prev0;
}
void kill_region() {
	character c;
	int i;
	readonly_check
	if (mark == prev) return;
	else if (mark < prev) {
		if (!valid_bwd_pos(mark)) mark--;
		delbfr(mark,TRUE);
	} else {
		i = min(next + mark - prev, BEND);
		if (!valid_fwd_pos(i)) i++;
		delaft(i,TRUE);
	}
}
void yank() {
	character c;
	int ccol0 = ccol, crow0 = crow, prev0 = prev;
	int from, to, i;

	readonly_check
	if (kbufleft) {
		from = 0; to = kbufp;
	} else {
		from = kbufp; to = KBUFlen;
	}
	if (from >= to) {
		warn("kill-buffer empty"); return;
	}
	if (check_buf((to-from)*nrepeat)) return;
	while ((nrepeat--) > 0) {
		for (i = from; i < to; i++) {
			c = kbuf[i];
			if (kanji(c)) ins2(c,kbuf[++i]);
			else ins1(c);
		}
	}
	reform(crow-crow0,0);
	redraw1(ccol0, crow0, prev0+1);
	modified();
}
void open_line() {
	int ccol0 = ccol, crow0 = crow, inc = 0;
	readonly_check
	while ((nrepeat--) > 0) {
		if (check_buf(inc+1)) break;
		inc++;
	}
	if (inc != 0) {
		ccol = 0; crow += inc;
		reform(inc,0);
		while ((inc--) > 0) buf[--next] = NL;
		ccol = ccol0; crow = crow0;
		redraw1(ccol, crow, next);
		modified();
	}
}
void append_line() {
	character c, c0;
	int ccol0, crow0, prev0, dec = 0, i;
	readonly_check
	while ((nrepeat--) > 0) {
		for (i = tol(prev); buf[i] != NL; i = tol(i-1));
		if (i <= BSTART) {
			warn("no previous line"); break;
		}
		bwdbufcsr(i);
		crow--;
		dec++;
		for (c=buf[--prev]; c == SP || c == TAB || c == LS; c=buf[--prev])
			if (c == LS) {
				dec++; crow--;
			}
		c0 = c;
		findcol();
		for (c=buf[next]; c == SP || c == TAB || c == LS; c=buf[++next])
			if (c == LS) dec++;
		ccol0 = ccol; crow0 = crow; prev0 = prev;
		if (c0 != NL && c0 != '(' && c0 != '{' && c0 != '['
		    && c != NL && c != ')' && c != '}' && c != ']') {
			ins1(SP);
			dec -= crow-crow0;
		}
	}
	reform(0,dec);
	redraw1(ccol0,crow0,prev0+1);
	modified();
}

char schstr[HYAKU];

void search_fwd() {
	character c, sc, sc1;
	int i, si, bi, x;
	x = readstr("search \"", schstr, CtrlS);
	if (x == 1) {
		aborted(); return;
	} else if (x == 0) {
		/*
		if (msgbuf[0] == 0) {
			warn("nothing to search for"); return;
		}
		strcpy(schstr,msgbuf);
		*/
		if (msgbuf[0] != 0) strcpy(schstr,msgbuf);
		else if (schstr[0] == 0) {
			warn("nothing to search for"); return;
		}
	} else if (schstr[0] == 0) {
		warn("nothing to search for"); return;
	}
	sc = schstr[0];
	for (i = next; i < BEND; i++) {
		if (buf[i] == sc) {
			for (si=1,bi=i+1; (sc1 = schstr[si]) != 0; si++,bi++) {
				if (bi >= BEND) break;
				if ((c = buf[bi]) == LS) c = buf[++bi];
				if (sc1 != c) break;
			}
			if (sc1 == 0 && valid_fwd_pos(bi)) {
				fwdbufcsr(bi); setcsr(); return;
			}
		}
	}
	warn1("Failed search \"%s\"", schstr);
}
void search_bwd() {
	character c, sc;
	int i, si, bi, n, x;
	x = readstr("R-search \"", schstr, CtrlR);
	if (x == 1) {
		aborted(); return;
	} else if (x == 0) {
		if (msgbuf[0] == 0) {
			warn("nothing to search for"); return;
		}
		strcpy(schstr,msgbuf);
	} else if (schstr[0] == 0) {
		warn("nothing to search for"); return;
	}
	n = strlen(schstr);
	sc = schstr[n-1];
	for (i = prev; i > BSTART; i--) {
		if (buf[i] == sc) {
			for (si=n-2,bi=i-1; si>=0; si--,bi--) {
				if (bi <= BSTART) break;
				if ((c = buf[bi]) == LS) c = buf[--bi];
				if (schstr[si] != c) break;
			}
			if (si < 0 && valid_bwd_pos(bi)) {
				bwdbufcsr(bi); setcsr(); return;
			}
		}
	}
	warn1("Failed reverse-search \"%s\"", schstr);
}
char repstr[HYAKU];
char repfrom[HYAKU];
char repfrom1[HYAKU];
char repto[HYAKU];

replace_strings(qp) int qp; {
	SPRINTF1(repstr, "%sreplace \"", (qp ? "Q-" : ""));
	if (readstr(repstr,"",0) == 1) {
		aborted(); return(TRUE);
	}
	if (msgbuf[0] == 0) {
		if (repfrom[0] == 0) {
			warn("nothing to replace"); return(TRUE);
		}
		if (!qp) {
			SPRINTF3(repstr, "%sreplace \"%s\" with \"%s\"",
			         (qp ? "Q-" : ""), repfrom, repto);
			if (!query(repstr)) {
				aborted(); return(TRUE);
			}
		}
	} else {
		strcpy(repfrom1,msgbuf);
		SPRINTF2(repstr, "%sreplace \"%s\" with \"",
		         (qp ? "Q-" : ""), repfrom1);
		if (readstr(repstr,"",0) == 1) {
			aborted(); return(TRUE);
		}
		strcpy(repto,msgbuf);
		strcpy(repfrom,repfrom1);
	}
	return(FALSE);
}

void query_replace() {
	character c, sc, sc1;
	int i, si, bi, tolen, dec, bound, ccol0, crow0, prev0;
	int found = FALSE, needredraw = FALSE;

	readonly_check
	if (replace_strings(TRUE)) return;
	if (mark <= prev) {
		warn("the region is empty"); return;
	}
	bound = min(mark - prev + next, BEND);
	tolen = strlen(repto);
	SPRINTF2(repstr, "Q-replace \"%s\" with \"%s\"? (SP or DEL): ",
	         repfrom, repto);
	sc = repfrom[0];
	for (i = next; i < bound; i++) {
		if (buf[i] == sc) {
			for (si=1,bi=i+1; (sc1 = repfrom[si]) != 0; si++,bi++) {
				if (bi >= BEND) break;
				if ((c = buf[bi]) == LS) c = buf[++bi];
				if (sc1 != c) break;
			}
		if (sc1 == 0 && valid_fwd_pos(bi)) {
			if (bi > bound) continue;
			found = TRUE;
			fwdbufcsr(i);

			setcsr();
			refresh_screen();
			warn(repstr);
			c = keyin();
			if (c == SP) addmsg("yes");
			else if (c == DEL || c == Rubout) {
				addmsg("no"); continue;
			} else {
				if (c != CtrlG) unkeyin(c);
				aborted();
				if (needredraw) center();
				return;
			}

			if (check_buf(tolen)) break;
			dec = 0;
			ccol0 = ccol; crow0 = crow; prev0 = prev;
			/* insert to-string */
			for (si = 0; si < tolen; si++) {
				c = repto[si];
				if (kanji(c)) ins2(c,repto[++si]);
				else ins1(c);
			}
			/* delete from-string */
			while (next < bi) {
				c = buf[next++];
				if (c == LS || c == NL) dec++;
			}
			reform(crow-crow0,dec);
			if (crow0 >= 0 && crow <= MAXROW) {
				redraw1(ccol0, crow0, prev0+1);
				needredraw = FALSE;
			} else
				needredraw = TRUE;
			modified();
			i = next-1;
			}
		}
	}
	if (!found) {
		warn1("Failed search \"%s\"", repfrom);
	} else {
		if (needredraw) center();
		addmsg(" Done");
	}
}
void replace() {
	character c, sc, sc1;
	int i = next, si, bi, tolen, bound;
	int dec = 0, ccol0 = ccol, crow0 = crow, prev0 = prev;
	int found = FALSE, needredraw = FALSE;

	readonly_check
	if (replace_strings(FALSE)) return;
	if (mark <= prev) {warn("the region is empty"); return;}
	bound = min(mark - prev + next, BEND);
	tolen = strlen(repto);
	sc = repfrom[0];
	for (; i < bound; i++) {
		if (buf[i] == sc) {
			for (si=1,bi=i+1; (sc1 = repfrom[si]) != 0; si++,bi++) {
				if (bi >= BEND) break;
				if ((c = buf[bi]) == LS) c = buf[++bi];
				if (sc1 != c) break;
			}
			if (sc1 == 0 && valid_fwd_pos(bi) && bi <= bound) {
				found = TRUE;
				while (next < i) {
					c = buf[next++];
					if (kanji(c)) ins2(c,buf[next++]);
					else if (c == LS) dec++;
					else {
						if (c == NL) dec++;
						ins1(c);
					}
				}
				adjust_LS();

				if (check_buf(tolen)) break;
				/* insert to-string */
				for (si = 0; si < tolen; si++) {
					c = repto[si];
					if (kanji(c)) ins2(c,repto[++si]);
					else ins1(c);
				}
				/* delete from-string */
				while (next < bi) {
					c = buf[next++];
					if (c == LS || c == NL) dec++;
				}
				modified();
				i = next-1;
			}
		}
	}
	if (!found) {
		warn1("Failed search \"%s\"", repfrom);
		return;
	}
	reform(crow-crow0, dec);
	redraw1(ccol0, crow0, prev0+1);
	addmsg(" Done");
}
int numericp(c) character c; {
	if (c >= '0' && c <= '9') return(TRUE);
	else return(FALSE);
}
void set_nrepeat() {
	character c;
	int n = 4;
	lastcom_valid = TRUE;
	msg("^U-");
	c = keyin();
	if (c == CtrlU) {
		do {
			n *= 4; addmsg("^U-");
		} while ((c = keyin()) == CtrlU);
	} else if (c >= '0' && c <= '9') {
		n = 0;
		do {
			n = n*10+c-'0'; msgchar(c); msgchar('-');
			c = keyin();
		} while (c >= '0' && c <= '9');
		if (n == 0) n = 1;
		while (c == CtrlU) {
			n *= 4; addmsg("~U-");
			c = keyin();
		}
	}
	if (c == CtrlG) aborted();
	else {
		unkeyin(c); nrepeat = n;
	}
}
void transpose() {
	character c1, c2, c21;
	int inc, dec = 0;
	int ccol0, crow0, prev0;
	if (prev <= BSTART) {
		warn("no previous character"); return;
	}
	if (next >= BEND) {
		warn("no next character"); return;
	}
	if (simple_transpose()) return;

	c2 = buf[next++];
	if (c2 == NL) dec++;
	else {
		if (kanji(c2)) c21 = buf[next++];
		if (buf[next] == LS) {
			next++; dec++;
		}
	}
	reform(0,dec);
	inc = Ginc; dec = Gdec;

	c1 = buf[--next] = buf[prev--];
	if (c1 == NL) {
		findcol(); crow--; inc++; dec++;
	} else if (c1 == LS) {
		c1 = buf[--next] = buf[prev--];
		if (kanji(c1)) buf[--next] = buf[prev--];
		findcol(); crow--; inc++; dec++;
	} else if (kanji(c1)) {
		buf[--next] = buf[prev--];
		ccol -= 2;
	} else if (c1 == TAB) findcol();
	else ccol = bwdcol(ccol,c1);

	ccol0 = ccol; crow0 = crow; prev0 = prev;
	if (kanji(c2)) ins2(c2,c21);
	else ins1(c2);
	reform(crow-crow0,dec);
	redraw1(ccol0, crow0, prev0+1);
	modified();
}
change_kanji_code1(name, modep) char *name; int *modep; {
	character c;
/*
	 123456789 123456789 123456789 123456789 123456789
	Kanji code (SP, DEL or NL): File Output: SJIS");
*/
	throw_MSGcursor(28);
	tyostr(name);
	for (;;) {
		throw_MSGcursor(41);
		switch(*modep) {
		case EUC:	tyostr("EUC "); break;
		case SJIS1:	*modep = SJIS;
		case SJIS:	tyostr("SJIS"); break;
		case JISESC:
		case JISKANJI: *modep = JISASCII;
		case JISASCII:	tyostr("JIS "); break;
		default:	tyoint(*modep);
		}
		c = keyin();
		if (c == CR) return(NL);
		else if (c == SP || c == NL || c == CtrlG) return(c);
		else if (c == DEL || c == Rubout) {
			if ((++(*modep)) > 2) *modep = 0;
		} else ring_bell();
	}
}
void change_kanji_code() {
	character c;
	int tyim = tyimode, tyom = tyomode, finm = finmode, foutm = foutmode;
	int fnamem = fnamemode;
	throw_MSGcursor(0);
	clrtoeol();
	tyostr("Kanji code (SP, DEL or NL): ");
	for (;;) {
		c = change_kanji_code1(" File Input: ", &finm);
		if (c != SP) break;
		c = change_kanji_code1("File Output: ", &foutm);
		if (c != SP) break;
		c = change_kanji_code1("   Keyboard: ", &tyim);
		if (c != SP) break;
		c = change_kanji_code1("    Display: ", &tyom);
		if (c != SP) break;
		c = change_kanji_code1("  File Name: ", &fnamem);
		if (c != SP) break;
	}
	throw_MSGcursor(0); clrtoeol();
	if (c == NL) {
		if (tyomode != tyom) {
			center(); tyomode = tyom;
		} else
			throw_cursor(ccol,crow);
		tyimode = tyim; finmode = finm; foutmode = foutm;
		if (fnamemode != fnamem) {
			fnamemode = fnamem;
			write_fname(cfile, internal_cfile);
		}
	} else aborted();
}

void ESCdispatch() {
	character c;
	addmsg("ESC-"); c = keyin(); clearmsg();
	switch(c) {
	case CtrlG: aborted(); break;
	case DEL: case Rubout: delete_prev_wd(); break;
	case ESC: exchange_mark(); break;
	case SP: set_mark(); break;
	case 'A': case 'a': goto_real_tol(); break;
	case 'B': case 'b': bwd_wd(); break;
	case 'C': case 'c': updown_wd('C'); break;
	case 'D': case 'd': delete_next_wd(); break;
	case 'E': case 'e': goto_real_eol(); break;
	case 'F': case 'f': fwd_wd(); break;
	case 'L': case 'l': updown_wd('L'); break;
	case 'M': case 'm': show_macro(); break;
	case 'N': case 'n': fwd_real_ln(); break;
	case 'P': case 'p': bwd_real_ln(); break;
	case 'Q': case 'q': query_replace(); break;
	case 'R': case 'r': replace(); break;
	case 'S': case 's': search_fwd(); break;
	case 'U': case 'u': updown_wd('U'); break;
	case 'V': case 'v': bwd_pg(); break;
	case '<': goto_top(); break;
	case '>': goto_bottom(); break;
	case '^': append_line(); break;
	case '(': begin_klog(); break;
	case ')': end_klog(); break;
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case '0': begin_replay(c-'0'); break;
	default: unbound(); if (kanji(c)) keyin();
	}
}
int CXdispatch() {
	character c;
	addmsg("^X-"); c = keyin(); clearmsg();
	switch(c) {
	case CtrlC:
		if (modp) {
			if (query("Discard changes?")) return(SIG_EXIT);
			else {
				aborted(); return(SIG_CONT);
			}
		} else return(SIG_EXIT);
	case CtrlG: aborted(); break;
	case CtrlS: case CtrlW: save_file(TRUE); break;
	case CtrlV: visit_file(FALSE); break;
	case CtrlR: visit_file(TRUE); break;
	case CtrlK: change_kanji_code(); break;
	case CtrlL: count_lines(); break;
	case CtrlB: count_bytes(); break;
	case CtrlT: set_tab(); break;
	case 'S': case 's': save_region(); break;
	case 'I': case 'i': insert_file(); break;
#ifdef EMBED
	case 'Y': case 'y': if (save_file(FALSE)) return SIG_LOAD; else break;
#endif
	default: unbound(); if (kanji(c)) keyin();
	}
	return(SIG_CONT);
}
void CZdispatch() {
	void stop_mine();
	character c;
	addmsg("^Z-"); c = keyin(); clearmsg();
	switch(c) {
	case CtrlG: aborted(); break;
	case DEL: case Rubout: delete_prev_Sexpr(); break;
	case CtrlB: bwd_Sexpr(); break;
	case CtrlK: delete_next_Sexpr(); break;
	case CtrlF: fwd_Sexpr(); break;
#ifdef unix
	case CtrlZ: stop_mine(); break;
#endif
	default: unbound(); if (kanji(c)) keyin();
	}
}
int dispatch() {
	character c;
	int sig;
	lastcom = 0;
	lastcom_valid = FALSE;
	nrepeat = 1;
	for (;;) {
		c = keyin();
		if (msgcol != 0 && nrepeat == 1) clearmsg();
		if (kanji(c)) insert2(c, keyin());
		else if (c > SP && c < Rubout)
			switch(c) {
			case ')': case '}': case ']': insert_paren(c); break;
			default: insert(c);
			}
		else
			switch(c) {
			case LS: break;
			case CtrlA: goto_tol(); break;
			case CtrlB: bwd_char(); break;
			case CtrlC: unbound(); break;
			case CtrlD: delete_next_char(); break;
			case CtrlE: goto_eol(); break;
			case CtrlF: fwd_char(); break;
			case CtrlG: warn("nothing to abort"); break;
			case Rubout:
			case CtrlH: delete_prev_char(); break;
			case CtrlI: indent(FALSE); break;
			case CtrlJ: insert(NL); break;
			case CtrlK: kill_ln(); break;
			case CtrlL: redisplay(); break;
			case CtrlM: indent(TRUE); break;
			case CtrlN: fwd_ln(); break;
			case CtrlO: open_line(); break;
			case CtrlP: bwd_ln(); break;
			case CtrlQ: addmsg("^Q-"); insert(keyin()); clearmsg(); break;
			case CtrlR: search_bwd(); break;
			case CtrlS: search_fwd(); break;
			case CtrlT: transpose(); break;
			case CtrlU: set_nrepeat(); break;
			case CtrlV: fwd_pg(); break;
			case CtrlW: kill_region(); break;
			case CtrlX: if(sig = CXdispatch()) return(sig); else break;
			case CtrlY: yank(); break;
			case CtrlZ: CZdispatch(); break;
			case ESC: ESCdispatch(); break;
			default: insert(c);
			}
		if (c != CtrlU) nrepeat = 1;
		if (lastcom_valid) lastcom_valid = FALSE;
		else lastcom = 0;
		refresh_screen();
	}
        return 0;               /* sig = SIG_CONT */
}

#ifndef unix

void read_cs(fp, narg, str)
FILE *fp; int narg; char *str;
{
	character c;
	char *s = str;
	while ((c = getc(fp)) != '"')
		if (c == ';')
			while ((c = getc(fp)) != '\n');
	while ((c = getc(fp)) != '"') {
		if (c == '\\') {
			c = getc(fp);
			if (c >= '0' && c <= '7') {
				int i = (c - '0') << 6;
				i |= (getc(fp) - '0') << 3;
				*(s++) = i | (getc(fp) - '0');
			} else
				*(s++) = c;
		} else
			*(s++) = c;
	}
	*s = 0;
	s = str;
	while ((c = *(s++)) != 0) {
		if (c == '%')
			if ((c = *(s++)) != '%') {
				if (c == 'c') c = *s++;
				if (c <= '0' || c > narg) {
					fprintf(stderr, "Illegal minecap arg spec '%c'.\n", c);
					exit(1);
				}
				if ((c = *(s++)) != '%') {
					if (c != '+' && c != '-') {
						fprintf(stderr, "Illegal minecap arg offset '%c'.\n", c);
						exit(1);
					}
					while (*(s++) != '%');
				}
			}
	}
}

void load_minecap(arg0)
char *arg0;
{
	FILE *fp;
	char *p, *q = tmpstr-1;

    for (p = tmpstr ; *p = *arg0++; ++p)
    	if (*p == '\\')
    		q = p;
	
	strcpy(q+1, "minecap");

	if ((fp = fopen(tmpstr, "r")) == NULL) {
		fputs("No minecap -- MINE cannot work.", stderr);
		exit(1);
	}
	read_cs(fp, '2', CSthrow_cursor);
	read_cs(fp, '0', CSclear_display);
	read_cs(fp, '0', CSclear_to_eol);
	read_cs(fp, '0', CSring_bell);
	read_cs(fp, '0', CSbegin_reverse_video);
	read_cs(fp, '0', CSend_reverse_video);
	read_cs(fp, '1', CSset_scroll_region);
	read_cs(fp, '1', CSinsert_nlines);
	read_cs(fp, '1', CSdelete_nlines);
	read_cs(fp, '0', CSinsert_line);
	read_cs(fp, '0', CSdelete_line);
	fclose(fp);	
	if (CSdelete_nlines[0] == 0) CSinsert_nlines[0] = 0;
	if (CSdelete_line[0] == 0) CSinsert_line[0] = 0;
}

#endif

#ifdef unix
char *ta, *ti, *te;
#endif

#if defined(SUNSOL2) || defined(__CYGWIN32__)
int sigsetmask(mask)            /* from FreeBSD source */
int mask;
{
    int omask, n;

    n = sigprocmask(SIG_SETMASK, (sigset_t *) &mask, (sigset_t *) &omask);
    if (n)
        return (n);
    return (omask);
}
#endif

#ifdef unix
void stop_mine() {
#if !defined(SUNSOL2) || !defined(__i386)
	int omask = sigsetmask(0);
#endif

	if (nonnil(CSset_scroll_region))
	     format1(CSset_scroll_region,LINES-1);
#ifdef LINUX_OLD
	sysV_ioctl(LOAD);
#else
# ifdef SUNSOL2
	system("/bin/stty -raw echo iexten");
# else
	system("/bin/stty -raw echo");
# endif
#endif
	if(msgcol != 0) clearmsg();
	throw_cursor1(0,LINES-1);
	refresh_screen();

	if (te) puts(te);
#ifdef LINUX_OLD
	endwin();
#endif
	kill(0,SIGTSTP);
#ifdef LINUX_OLD
	initscr();
#endif
#if !defined(SUNSOL2) || !defined(__i386)
	sigsetmask(omask);
#endif
	if (ti) puts(ti);
	if (nonnil(CSset_scroll_region))
	     format1(CSset_scroll_region,LINES-1);
	clear();
#ifdef LINUX_OLD
	sysV_ioctl(MINE);
#else
# ifdef SUNSOL2
	system("/bin/stty raw -echo -iexten");
# else
	system("/bin/stty raw -echo");
# endif
#endif
	if (nonnil(CSset_scroll_region))
	     format1(CSset_scroll_region,MAXROW);
	refresh_screen();
	redisplay();
}
#endif

#ifdef KCL
void init_mine1()
#else
main(argc,argv) int argc; char *argv[];
#endif
{
	FILE *fp;
	int i;
	int kanjispec = -1;
#ifdef LINUX_OLD
	initscr();
	sysV_ioctl(SAVE);  /* Save current terminal mode. */
#endif
#ifdef MALLOCED_BUF
        buf = malloc(BUFlen);
#endif
	internal_cfile[0] = 0;
	TABSPACE = 4;
	readonly = FALSE;
#ifdef unix
	fixed_window = FALSE;
#endif
#ifdef pc
	fixed_window = TRUE;
#endif
#ifndef LINUX_OLD
	LINES = 24; COLS = 80;
#endif
#ifndef EMBED
	for (i=1; i<argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			case 'r': readonly = TRUE; break;
			case 't': sscanf(argv[i]+2, "%d", &TABSPACE); break;
			case 'j': kanjispec = JISASCII; break;
			case 's': kanjispec = SJIS; break;
			case 'e': kanjispec = EUC; break;
			case 'd': fixed_window = TRUE;
			          sscanf(argv[i]+2, "%d,%d", &COLS, &LINES); break;
			default: fputs("usage: mine [-rtjsed] [filename]", stderr);
					 exit(1);
			}
		} else
			strcpy(cfile, argv[i]);
	}
#endif
#ifdef unix
	if (!(tty_name = getenv("TERMCAP")))
		tty_name = getenv("TERM");
	if (tty_name && tgetent(termcap_bp, tty_name) == 1) {
#ifndef LINUX_OLD
		if (!fixed_window) {
# ifdef __CYGWIN32__
                    LINES = atoi(getenv("LINES"));
                    COLS = atoi(getenv("COLUMNS"));
# else
                    LINES = tgetnum("li");
                    COLS = tgetnum("co");
# endif
		}
#endif
		ta = tgetstr_area;
		ti = tgetstr("ti", &ta);
		te = tgetstr("te", &ta);
	}
#endif

	set_display_parameters();
	if (MAXROW < 10 || MAXCOL < 10) {
#if (defined(KCL) && !defined(ED_MINE))
		FEerror("The window is too small.", 0);
#else
		fputs("The window is too small.", stderr);
		exit(1);
#endif
	}

	if (kanjispec != -1) {
		tyimode = tyomode = finmode = foutmode = fnamemode = kanjispec;
	} else {
#ifdef unix
		tyimode = tyomode = EUC;
		finmode = foutmode = fnamemode = JISASCII;
#endif
#ifdef pc
		tyimode = tyomode = finmode = foutmode = fnamemode = SJIS;
#endif
	}
	read_fname(internal_cfile,cfile);

#ifdef TUS
	puts("TUTScheme version 1.00.  Oct 7, 1991");
#else
#ifndef KCL
	puts("MINE version 1.00.  Sept 21, 1990");
#endif
#endif
	puts("(C) Copyright Taiichi Yuasa, 1990.  All rights reserved.");

	buf[BSTART] = buf[BEND] = NL;
	prev = mark = BSTART; next = BEND;
	crow = ccol = ccol1 = pcol = 0;
	kbufp = 0;
	kbufleft = TRUE;
	modp = FALSE;
	msgcol = 0;
	schstr[0] = repfrom[0] = repto[0] = 0;
	for (i = 0; i <= 9; i++) {
		klog[KLOGlen-10+i] = 0;
		klogindex[i] = KLOGlen-10+i;
	}
	cklog = klogp = replayp = -1;

	cdir[0] = 0;
#ifdef unix
#ifdef HP
        getcwd(cdir, HYAKU);
#else
	getwd(cdir);
#endif
#endif
#ifdef pc
	getcwd(cdir,HYAKU);
#endif

	tyiptr = finptr = 0;

#ifndef unix
	load_minecap(*argv);
#endif

#ifdef TUS
	sc_main(argc,argv);
#else
#ifdef KCL
	return;
#else
#ifdef unix
	c_edit(te);
#else
	c_edit();
#endif
#endif
#endif
}
#ifdef TUS
char far * far c_edit(fname) char far *fname;
#else
#ifdef KCL
char *c_edit(fname, l) char *fname; int l;
#else
c_edit()
#endif
#endif
{
#ifdef EMBED
	int endval;
	char *p = cfile;
#endif
#ifdef unix
	if (ti) puts(ti);
#endif
	if (nonnil(CSset_scroll_region))
		format1(CSset_scroll_region,MAXROW);
    clear();

#ifdef EMBED
    if (fname && *fname) {
#ifdef TUS
        while (*p++ = *fname++) ;
#endif
#ifdef KCL
        while (l--) *p++ = *fname++;
		*p = '\0';
#endif
        read_fname(internal_cfile, cfile);
        load_file(FALSE);
    } else {
    	redisplay();
    	/* 
    	center(); banner();
    	*/
    }
#else
	if (internal_cfile[0] == 0) {
		redraw(BSTART);
		banner();
	} else load_file(readonly);
#endif

#ifdef unix
#ifdef LINUX_OLD
	sysV_ioctl(MINE);
#else
# ifdef SUNSOL2
	system("/bin/stty raw -echo -iexten");
# else
	system("/bin/stty raw -echo");
# endif
#endif
#endif
#ifdef pc
	{
		union REGS inregs;
		union REGS outregs;
		inregs.x.ax = 0x4400;	/* get */
		inregs.x.bx = 0;	/* console-input */
		intdos(&inregs, &outregs);
		inregs.h.dh = 0;
		inregs.h.dl = outregs.h.dl | 0x20;
		inregs.x.ax = 0x4401;	/* set */
		intdos(&inregs, &outregs);

		inregs.x.ax = 0x4400;	/* get */
		inregs.x.bx = 1;	/* console-output */
		intdos(&inregs, &outregs);
		inregs.h.dh = 0;
		inregs.h.dl = outregs.h.dl | 0x20;
		inregs.x.ax = 0x4401;	/* set */
		intdos(&inregs, &outregs);
	}
#endif

	if (nonnil(CSset_scroll_region))
		format1(CSset_scroll_region,MAXROW);
    refresh_screen();

#ifdef EMBED	/* for xterm by Ushijima */
	redisplay();
#endif

#ifdef EMBED
	endval = 
#endif
	dispatch();

	if (nonnil(CSset_scroll_region))
		format1(CSset_scroll_region,LINES-1);

#ifdef unix
#ifdef LINUX_OLD
	sysV_ioctl(LOAD);
#else
# ifdef SUNSOL2
	system("/bin/stty -raw echo iexten");
# else
	system("/bin/stty -raw echo");
# endif
#endif
#endif
#ifdef pc
	{
		union REGS inregs;
		union REGS outregs;
		inregs.x.ax = 0x4400;	/* get */
		inregs.x.bx = 0;	/* console-input */
		intdos(&inregs, &outregs);
		inregs.h.dh = 0;
		inregs.h.dl = outregs.h.dl & 0xdf;
		inregs.x.ax = 0x4401;	/* set */
		intdos(&inregs, &outregs);

		inregs.x.ax = 0x4400;	/* get */
		inregs.x.bx = 1;	/* console-output */
		intdos(&inregs, &outregs);
		inregs.h.dh = 0;
		inregs.h.dl = outregs.h.dl & 0xdf;
		inregs.x.ax = 0x4401;	/* set */
		intdos(&inregs, &outregs);
	}
#endif
	if(msgcol != 0) clearmsg();
	throw_cursor1(0,LINES-1);
	refresh_screen();
#ifdef unix
	if (te) puts(te);
#endif
#ifdef LINUX_OLD
	endwin();
#endif
#ifdef TUS
	if (endval == SIG_LOAD) return((char far *) cfile);
	else return((char far *) NULL);
#endif
#ifdef KCL
	if (endval == SIG_LOAD) return(cfile);
	else return(NULL);
#endif
}

#ifdef KCL
#include <sys/ioctl.h>

int mine_initialized;

#ifndef ED_MINE
void siLmine() {
	char *endval;
	object *base = vs_base;
	int n = stdin->_cnt;

	check_arg(1);

	if (vs_base[0] != Cnil)
		check_type_string(&vs_base[0]);

	if (!mine_initialized) {
		init_mine1();
		mine_initialized = TRUE;
	}

	while(n--) getchar();

	if (vs_base[0] == Cnil)
		endval = c_edit(NULL,0);
	else
		endval = c_edit(vs_base[0]->st.st_self, vs_base[0]->st.st_fillp);

	if (endval == NULL)
		vs_base[0] = Cnil;
	else
		vs_base[0] = make_simple_string(endval);
}

void init_mine()
{
	mine_initialized = FALSE;
	make_si_function("MINE", siLmine);
}
#endif /* ED_MINE */

#endif

#ifdef LINUX_OLD
void sysV_ioctl(x)  /* By N.Watanabe */
int x;
{
    switch(x){
    case SAVE : 
        ioctl(0, TCGETA, &tin);
        tio = tin;
        break;
    case MINE : /*    system("stty -raw echo"); ... original    */
        tio.c_iflag &= ~ISTRIP;
        tio.c_iflag &= ~IXON;
        tio.c_lflag &= ~ICANON;
        tio.c_lflag &= ~ECHO;
        tio.c_lflag &= ~ISIG;
        /*
        tio.c_cc[VINTR] = -1;
        tio.c_cc[VSUSP] = -1;
        */
         tio.c_cc[VTIME] = 0;
        tio.c_cc[VMIN] = 1;
        ioctl(0, TCSETA, &tio);
        break;
    case LOAD :
        ioctl(0, TCSETAW, &tin);
        break;
    }
}
#endif
