/*
 * msgcunit.c - 
 */

/*
 * $Id: msgcunit.c,v 1.4 2014/10/28 02:22:55 komiya Exp $
 *
 *              MARK and SWEEP gc routine
 *       By T.Komiya & T.Hayasaka (Sep. 2 ,1990)
 *       Original routine was programed by M.Yamamoto
 * 
 *    Update  : T.Komiya      ??
 *                        Oct. 26 '90 (wp)
 */

static char *RCSid="$Id: msgcunit.c,v 1.4 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"

#define MARKP(obj)              (0x1 & (unsigned)(obj))
#define BACKP(obj)              (0x2 & (unsigned)(obj))
#define CLR_GC_MARK(obj)        (obj &= 0xfffffffe)
#define CLR_GC_BACK(obj)        (obj &= 0xfffffffd)
#define TAG_AND_MARK_MASK(obj)  (0xfc000003 & (unsigned)(obj))
#define GETtag2bit(obj)         (0x0c000000 & (unsigned)(obj))
#define SPECIAL_TAG_FUNCTION    0x80000000
#define SPECIAL_FUNTAG_MASK     0xe0000000
#define SPECIAL_TAG_LOSTOBJ_E   0x38000000

unsigned *previous;          /* a2 */
unsigned *current;           /* a3 */
unsigned current_content;    /* d1(d2) */
unsigned *next_area;         /* a0 */
int next_area_content;  /* d3 */
#if defined(THINK_C) || defined(MACHTEN)
char dummy_dummy[8];
unsigned *dummy_source;
extern unsigned *src_tmp;
#elif defined(__CYGWIN32__)
unsigned *dummy_source;
unsigned *src_tmp;
#else
unsigned dummy_source;
#endif

void mark_unit(src_add)
register unsigned    *src_add;
{
    void unit_block1();
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
    unsigned *org_src;
	
    org_src = src_add;
    src_add = src_tmp;
    *src_add = *org_src;
#endif	
    previous = src_add;
    if (0xe0000000 & *src_add && !(0x80000000 & *src_add)) {
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
        *org_src = *src_tmp;
#endif
        return;         /* immediate */
    }
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
    *dummy_source = *previous;
#else
    dummy_source = *previous;
#endif
    *previous = (unsigned)previous;
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
    current = dummy_source;
#else
    current = &dummy_source;
#endif
    *current |= 0x3;

unit_go_ahead :
    current_content = *current;
    if (0xfc000000 & current_content && !(0x80000000 & current_content)) {
        /* the case of immediate(include block1,2) */
        if (unit_change_or_back())
            goto unit_go_ahead;
        else {
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
            *org_src = *src_tmp;
#endif
            return;
        }
    }
    next_area = (unsigned *)ADDRESS(current_content);
    next_area_content = *next_area;
    if (MARKP(next_area_content))         /* Is the next area marked ? */
        /* marked */
        if (unit_change_or_back())
            goto unit_go_ahead;
        else {
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
            *org_src = *src_tmp;
#endif	
            return;
        }
    else {
        /*        if (BACKP(next_area_content)) */  /* Is the next area weak-linked? */
        /*            unit_weak_linked();       */                    /* weak-linked */
        if (current_content & 0x40000000) {      /* unit_not_weak_linked */
            if (current_content & 0x20000000) {
                /* unit_block2 */
                if (current_content & 0x18000000)
                    *next_area |= 0x1;         /* unit_normal_block2 */
                else {
                    /* unit_string */
                    if (*next_area & 0x04000000)
                        *next_area |= 0x1;     /* unit_normal_block2 */
                    else {                     /* unit_indirect_string */
                        *current = TAG_AND_MARK_MASK(current_content)
                         | SSERDDA(ADDRESS(*(next_area + 1)));
                        next_area = (unsigned *)ADDRESS(*(next_area + 1));
                        *next_area |= 0x1;
                    }
                }
                if (unit_change_or_back())
                    goto unit_go_ahead;
                else {
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
                    *org_src = *src_tmp;
#endif	
                    return;
                }
            } else {                /* NOTE!! this block depend on tagsystem */
                /* unit_block1d */
                if (GETtag(current_content) == LTAG_BIGNUM) {
                    /* unit_bignum */
                    *next_area |= 0x3;
                    *(next_area + 2) |= 0x1;

                    /* rotate */
                    *current = TAG_AND_MARK_MASK(current_content) | SSERDDA(previous);
                    previous = current;
                    next_area++;
                    current = next_area;
                    goto unit_go_ahead;
                } else if (GETtag(current_content) == LTAG_LOCK) {
                    /* lock object */
                    *next_area |= 0x3;
                    *(next_area + 2) |= 0x1;

                    /* rotate */
                    *current = TAG_AND_MARK_MASK(current_content) | SSERDDA(previous);
                    previous = current;
                    next_area++;
                    current = next_area;
                    goto unit_go_ahead;
                } else {
                    /* unit_block1 */
                    unit_block1();
                    goto unit_go_ahead;
                }
            }
        } else if (GETtag(current_content)) {      /* unit_cons_or_block1 */
            /* unit_block1 */
            unit_block1();
            goto unit_go_ahead;
        } else {                        /* unit_cons */
            *next_area |= 0x3;
            *(next_area + 1) |= 0x1;

            /* rotate */
            *current = TAG_AND_MARK_MASK(current_content) | SSERDDA(previous);
            previous = current;
            current = next_area;
            current++;
            goto unit_go_ahead;
        }
    }
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
    *org_src = *src_tmp;
#endif	
}


void unit_block1()
{
    switch (GETtag(current_content)) {
    case LTAG_WEAKPAIR :
        /*  unit_weakpair();
    return;
*/
    case LTAG_PORT :
        unit_port();
        return;
    case LTAG_FUTURE :
        unit_future();
        return;
    }
    if ((GETtag(current_content) & SPECIAL_FUNTAG_MASK) == SPECIAL_TAG_FUNCTION)
        if (unit_function())
            return;
    /* unit_normal_block1_object */
    *next_area |= 0x3;
/*  next_area_content = ADDRESS(next_area_content) >> 2; */    /* = "num" */
    next_area_content = IMM(next_area_content) >> 2;       /* = "num" */
    *(next_area + next_area_content) |= 0x1;
    *current = TAG_AND_MARK_MASK(current_content) | SSERDDA(previous);
    previous = current;
    current = next_area + next_area_content;
    return;
}


unit_change_or_back()
{
    register unsigned    *save_current,      /* a1 */
    curnt;              /* d5 */
    /*  static int bk, ch; */

change_or_back :

    if (BACKP(current_content)) {
        /*      bk++; */
        /* unit_back */
        CLR_GC_BACK(*current);
        save_current = current;
        curnt = (unsigned)save_current;
        current = previous;
#ifdef SPECIAL_CONS_TAG
        curnt |= TAG_AND_MARK_MASK(*previous);
#else
        curnt = SSERDDA(curnt) | TAG_AND_MARK_MASK(*previous);
#endif
        previous = (unsigned *)ADDRESS(*previous);
        if (MARKP(current_content)) {
            current_content = *current = curnt;
            if (previous != current)
                goto change_or_back;
            else {                  /* unit_end */
#if defined(THINK_C) || defined(MACHTEN) || defined(__CYGWIN32__)
                *previous = *dummy_source;
#else
                *previous = dummy_source;
#endif
                CLR_GC_MARK(*previous);
                return(0);          /* don't goto unit_go_ahead */
            }
        } else {        /* unit_back_weakpair */
            system_error(0);
            /*            
            curnt -= 4;
            current_content = *current = curnt;
            *save_current |= 0x1;
            if (previous != current)
                goto change_or_back;
*/
            /*            else */ /* unit_end */
            /*                return(0); */          /* don't goto unit_go_ahead */
        }
    } else {        /* unit_change */
        /*      ch++; */
        current--;
        return(1);  /* goto unit_go_ahead */
    }
}


/*
unit_weak_linked()
{
    register int new_obj, *start_add;

    new_obj = current_content;
    start_add = (unsigned *)ADDRESS(next_area_content);
    next_area_content
        = GETtag(next_area_content) | ms_reactive_lostobj(new_obj, start_add);
    *next_area = next_area_content;
}
*/

unit_port()
{
    *next_area |= 0x3;
    *(next_area + 2) |= 0x1;

    /* rotate */
    *current = SSERDDA(previous) | TAG_AND_MARK_MASK(current_content);
    previous = current;
    current = next_area + 1;
}

unit_future()
{
    *next_area |= 0x3;
    *(next_area + 5) |= 0x1;

    /* rotate */
    *current = SSERDDA(previous) | TAG_AND_MARK_MASK(current_content);
    previous = current;
    current = next_area + 4;
}

#ifdef CALLIOC
extern unsigned *marked_stack_list;
#endif

unit_function()
{
    switch (GETtag2bit(current_content)) {
    case GETtag2bit(LTAG_BCLOSURE) :
    case GETtag2bit(LTAG_SFUNCTION) :
    case GETtag2bit(LTAG_NCLOSURE) : /* NCLOSURE and MNCLOSURE */
        /* byte code closure (unit_byte_code) */
        *next_area |= 0x3;
        *(next_area + 6) |= 0x1;

        /* rotate */
        *current = SSERDDA(previous) | TAG_AND_MARK_MASK(current_content);
        previous = current;
        current = next_area + 5;
        return(1);
    case GETtag2bit(LTAG_CONTINUATION) :
#ifdef CALLIOC
        if (GETtag(current_content) == LTAG_IOC) {
            unsigned *bottom;
#ifdef SPECIAL_CONS_TAG
            bottom = (unsigned *)fix_to_int(((ENTITYP)ADDRESS(current_content))->ioc.bottom.obj);
#else
            bottom = (unsigned *)ADDRESS(fix_to_int(((ENTITYP)ADDRESS(current_content))->ioc.bottom.obj));
#endif
            if (!bottom[2] && bottom[5]) {
                bottom[2] = (unsigned)marked_stack_list;
                marked_stack_list = bottom;
            }
        }
#endif
	return(0);          /* continuation */
    }
}


#ifdef CALLIOC
mark_init_stack(bottom)
unsigned *bottom;
{
    unsigned    *base_p, *stacktop;

    if (bottom == (unsigned *)stack_bottom) {
        unsigned    *top, *top1;

        base_p = base_val();
        stacktop = top_val();
        top = (unsigned *)old_top;
        top1 = (unsigned *)old_top1;
        
        while (top != top1) {
            while (top1 != stacktop) {
                mark_unit(stacktop);
                stacktop++;
            }
            top1 = (unsigned *) * top1;
            stacktop++;
        }
    } else {
        base_p = stacktop = (unsigned *)bottom[1];
        mark_unit(bottom);
    }

    /* gc initialization for system stack */
    while (base_p) {
        while (base_p != stacktop) {
            mark_unit(stacktop);
            stacktop++;
        }
        base_p = (unsigned *) * stacktop;
        stacktop += 4;
    }
}
#else
mark_init_stack(base_p, stacktop)
register unsigned    *base_p, *stacktop;
{
    register unsigned    *top, *top1;

    top = (unsigned *)old_top;
    top1 = (unsigned *)old_top1;

    while (top != top1) {
        while (top1 != stacktop) {
            mark_unit(stacktop);
            stacktop++;
        }
        top1 = (unsigned *) * top1;
        stacktop++;
    }

    /* gc initialization for system stack */
    while (base_p) {
        while (base_p != stacktop) {
            mark_unit(stacktop);
            stacktop++;
        }
        base_p = (unsigned *) * stacktop;
        stacktop += 4;
    }
}
#endif

wp_msgc_start()
{
    unsigned **wpl, **wp_p;
    unsigned int tmp;

    /*    wpl = &get_sval(wplist.obj); */
    /*    if ((wpl = get_sval(wplist.obj)) != LTAG_NIL)
    set_sval(wplist.obj, ADDRESS(wpl));
*/
    wpl = (unsigned **)&(((ENTITYP)ADDRESS(wplist.obj))->sym.sym_val.obj);
    /*    while (GETtag(*wpl) != LTAG_NIL) { */
    while (*wpl != (unsigned *)LTAG_NIL) {
        wp_p = (unsigned **)ADDRESS(*wpl);
        if (MARKP(*wp_p)) {
            tmp = (unsigned)*(wp_p + 3);
            if (tmp & 0xe0000000 && !(tmp & 0x80000000))
                wpl = wp_p + 4;
            else if (MARKP(*(unsigned *)ADDRESS(tmp)))
                wpl = wp_p + 4;
            else {
                *(wp_p + 3) = (unsigned *)(LTAG_LOSTOBJ1 | SSERDDA(ADDRESS(tmp)));
                *wpl = *(wp_p + 4);
                *(wp_p + 4) = (unsigned *)LTAG_NILVECTOR;
            }
        } else
            *wpl = *(wp_p + 4);
    }
    if ((wpl = (unsigned **)get_sval(wplist.obj)) != (unsigned **)LTAG_NIL)
        set_sval(wplist.obj, ADDRESS(wpl) | LTAG_NIL);
}


/*
        Sweeper
*/

sweeper()
{
    register unsigned    *base, *end, *vacant_start;
    register int    vacant_continue, content, i;
    int total_vacant_area;

    wp_msgc_start();
    close_garbage_port();
    free_garbage_ncc();

    base = entity_base;
    end = entity_end;

    for (i = 0; i < BLOCK_TABLE_SIZE; i++)
        *(block_table + i) = 0;
    big_block = 0;
    total_vacant_area = 0;
    vacant_continue = FALSE;

    /* sweeper loop */

    while (base < end) {
        content = *base;
        if (MARKP(content)) {        /* marked */
            if (vacant_continue) {
                recovery_garbage(base, vacant_start, &total_vacant_area);
                vacant_continue = FALSE;
            }
            if (GETtag(content) == LTAG_BLOCK1) {
                ac_block1++;
                base += (IMM(content) >> 2) + 1;  /* block */
            } else if (GETtag(content) == LTAG_BLOCK2) {
                ac_block2++;
                base += (IMM(content) >> 2) + 1;  /* block */
            } else {
                ac_cons++;   /* cons */
                base += 2;
            }
        }                          /* end of marked */ else {                     /* non marked */
            if (!vacant_continue) {
                vacant_start = base;
                vacant_continue = TRUE;
            }
            if (GETtag(content) == LTAG_BLOCK1 || 
                GETtag(content) == LTAG_BLOCK2)
                /*      if(BACKP(content))
            base += (get_esc_word((unsigned *)ADDRESS(content)) >> 2) + 1;
        else
*/
                base += (IMM(content) >> 2) + 1;
            else
                base += 2;
        }                          /* end of non marked */
    }                              /* end of while */
    if (vacant_continue)
        recovery_garbage(base, vacant_start, &total_vacant_area);
#ifdef X
    draw_free_list();
#endif      
    return(total_vacant_area);
}


recovery_garbage(base, vacant_start, total_vacant_area)
register unsigned    *base, *vacant_start, *total_vacant_area;
{
    register int    vacant_area, quarter_vacant_area, header_of_vacancy;
    register unsigned    *table_addr;

    vacant_area = (unsigned)base - (unsigned)vacant_start;
    *total_vacant_area += vacant_area;
    header_of_vacancy = LTAG_BLOCK2 | (vacant_area - 4);
    quarter_vacant_area = vacant_area >> 2;
    if (quarter_vacant_area > 1) {
        if (quarter_vacant_area < BLOCK_TABLE_SIZE)
            table_addr = (unsigned *)(block_table + quarter_vacant_area);
        else
            table_addr = (unsigned *) & big_block;
        *vacant_start = header_of_vacancy;
        *(vacant_start + 1) = *table_addr;
        *table_addr = (unsigned)vacant_start;
    } else {
        if (quarter_vacant_area  == 1)
            *vacant_start = LTAG_BLOCK2;
        else
            system_error(1);
    }
}


system_error(n)
int n;
{
    printf("system error in the MARK & SWEEP GC(%d)\n", n);
}


clear_gc_bit()
{
    register unsigned    *base, *end;
    register int    content;

    base = entity_base;
    end = entity_end;

    while (base < end) {
        content = *base;
        switch (GETtag(content)) {
        case LTAG_BLOCK1 :
            if (!MARKP(content))
                base += ((IMM(content) >> 2) + 1);
            else {
                /*              ac_block1++; */
                CLR_GC_MARK(*base);
                switch (IMM(content)) {
                case 12 :          /* block1 type 2 big num */
                    if (GETtag(*(base + 2)) ==  LTAG_BLOCK2)
                        base += ((IMM(content) >> 2) + 1);
                    else {
                        base += IMM(content) >> 2;
                        CLR_GC_MARK(*base);
                        base++;
                    }
                    break;
                case 24 :          /* block1 type 2 lock object */
		                   /* block1 type 2 port */
                    if (GETtag(*(base + 2)) ==  LTAG_BLOCK2)
                        base += ((IMM(content) >> 2) + 1);
                    else {
                        base += IMM(content) >> 2;
                        CLR_GC_MARK(*base);
                        base++;
                    }
                    break;
                case 28 :           /* block1 type 2 native code */
                    if (GETtag(*(base + 4)) == LTAG_BLOCK2)
                        base += ((IMM(content) >> 2) + 1);
                    else {
                        base += IMM(content) >> 2;
                        CLR_GC_MARK(*base);
                        base++;
                    }
                    break;
                case 32 :           /* block1 type 2 byte code */
                    if (GETtag(*(base + 6)) == LTAG_BLOCK2)
                        base += ((IMM(content) >> 2) + 1);
                    else {
                        base += IMM(content) >> 2;
                        CLR_GC_MARK(*base);
                        base++;
                    }
                    break;
                case 44 :           /* future */
                    if (GETtag(*(base + 5)) == LTAG_BLOCK2)
                        base += ((IMM(content) >> 2) + 1);
                    else {
                        base += IMM(content) >> 2;
                        CLR_GC_MARK(*base);
                        base++;
                    }
                    break;
                default :                     /* block1 type 1 */
                    base += IMM(content) >> 2;
                    CLR_GC_MARK(*base);
                    base++;
                }
            }
            break;
        case LTAG_BLOCK2 :
            /*          ac_block2++; */
            CLR_GC_MARK(*base);
            base += ((IMM(content) >> 2) + 1);
            break;
        default :                        /* cons */
            /*          ac_cons++; */
            CLR_GC_MARK(*base);
            CLR_GC_MARK(*(base + 1));
            base += 2;
        }
    }
}
