;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;                                                                   ;;;;;
;;;;;          Scheme with PVM                by A243                   ;;;;;
;;;;;                                                                   ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(load "/mnta/tus-macro/Source/lsp/s_pvm3_err.lsp")
(load "/mnta/tus-macro/Source/lsp/s_pvm3_han.lsp")

(macro xpvm 
  (lambda (form) (pvm-mytid) (xpvm-start) (eval (cadr form)) (pvm-exit)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;; PVM Functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (pvm-addhosts . hosts)
  (let* ((nhost (length hosts))
         (infos (make-list nhost #f))
         (info (p-addhosts hosts nhost infos)))
    (if (= nhost info)
      #t
      (change-num-pvmsym infos))))
;      (begin (for-each (lambda(x)
;                     (if (negative? x) (pvm-error x 'pvm-addhosts))) infos)
;        (change-num-pvmsym infos)))))

(define (pvm-barrier group count)
  (let ((info (p-barrier group count)))
    (if (negative? info)
      (pvm-error info 'pvm-barrier)
      #t)))

(define (pvm-bcast group msgtag)
  (let ((info (p-bcast group msgtag)))
    (if (negative? info)
      (pvm-error info 'pvm-bcast)
      #t)))

(define (pvm-bufinfo bufid)
  (let* ((bytes '(#f)) (msgtag '(#f)) (tid '(#f))
         (info (p-bufinfo bufid bytes msgtag tid)))
    (if (negative? info)
      (pvm-error info 'pvm-bufinfo)
      (list (car tid) (car msgtag) (car bytes)))))

(define pvm-catchout-on '())
(define pvm-catchout-off '())

(let ((port #f))
  (set! pvm-catchout-on
    (lambda file
      (if port
          (error "already catchout on.")
          (begin
            (set! port
                  (if (null? file)
                      (current-output-port)
                      (open-output-file (car file))))
            (p-catchout port)
            (if (null? file)
              #t
              (car file))))))
  (set! pvm-catchout-off
    (lambda ()
      (if (not port)
          #f
          (begin (set! port #f) (p-catchout 0) #t)))))

(define (pvm-config)
  (let ((num '(#f)))
    (p-config num '() '())
    (let* ((hostp (do ((x (car num) (- x 1))
                       (ret '() (cons (make-list 4 #f) ret)))
                    ((= x 0) ret)))
           (arc '(#f))
           (info (p-config () arc hostp)))
      (if (negative? info)
        (pvm-error info 'pvm-config)
        (append num arc hostp)))))

(define (pvm-delete name index)
  (let ((cc (p-delete name index)))
    (if (negative? cc)
      (pvm-error cc 'pvm-delete)
      #t)))

(define (pvm-delhosts . hosts)
  (let* ((nhost (length hosts))
         (infos (make-list nhost #f))
         (info (p-delhosts hosts nhost infos)))
    (if (= nhost info)
      #t
      (change-num-pvmsym infos))))
;      (begin (for-each (lambda(x)
;                   (if (negative? x) (pvm-error x 'pvm-delhosts))) infos)
;        (change-num-pvmsym infos)))))

(define (pvm-exit)
  (let ((info (p-exit)))
    (if (negative? info)
      (pvm-error info 'pvm-exit)
      #t)))

(define (pvm-freebuf bufid)
  (let ((info (p-freebuf bufid)))
    (if (negative? info)
      (pvm-error info 'pvm-freebuf)
      #t)))

(define (pvm-getinst group tid)
  (let ((inum (p-getinst group tid)))
    (if (negative? inum)
      (pvm-error inum 'pvm-getinst)
      inum)))

(define (pvm-getopt what)
  (let ((val (p-getopt what)))
    (if (negative? val)
      (pvm-error val 'pvm-getopt)
      (case what
        ((1) (case val
               ((1) 'PvmDontRoute)
               ((2) 'PvmAllowDirect)
               ((3) 'PvmRouteDirect)
               (else (pvm-error val 'pvm-getopt))))
        ((2 3 9) (case val
                   ((0) 'off)
                   ((1) 'on)))
        (else val)))))

(define (pvm-getrbuf)
  (let ((bufid (p-getrbuf)))
    (if (negative? bufid)
      (pvm-error bufid 'pvm-getrbuf)
      (if (zero? bufid)
        #f
        bufid))))

(define (pvm-getsbuf)
  (let ((bufid (p-getsbuf)))
    (if (negative? bufid)
      (pvm-error bufid 'pvm-getsbuf)
      (if (zero? bufid)
        #f
        bufid))))

(define (pvm-gettid group inum)
  (let ((tid (p-gettid group inum)))
    (if (negative? tid)
      (pvm-error tid 'pvm-gettid)
      tid)))

(define (pvm-gsize group)
  (let ((size (p-gsize group)))
    (if (negative? size)
      (pvm-error size 'pvm-gsize)
      size)))

(define (pvm-halt)
  (let ((info (p-halt)))
    (if (negative? info)
      (pvm-error info 'pvm-halt)
      #t)))

(define (pvm-hostsync host)
  (let* ((clk (list #f #f))
         (delta (list #f #f))
         (info (p-hostsync host clk delta)))
    (if (negative? info)
      (pvm-error info 'pvm-hostsync)
      (append clk delta))))

(define (pvm-initsend . encoding)
  (let ((bufid (p-initsend (if (null? encoding)
                             PvmDataDefault
                             (car encoding)))))
    (if (negative? bufid)
      (pvm-error bufid 'pvm-initsend)
      bufid)))

(define (pvm-insert name index data)
  (let ((cc (p-insert name index data)))
    (if (negative? cc)
      (pvm-error cc 'pvm-insert)
      #t)))

(define (pvm-joingroup group)
  (let ((inum (p-joingroup group)))
    (if (negative? inum)
      (pvm-error inum 'pvm-joingroup)
      inum)))

(define (pvm-kill tid)
  (let ((info (p-kill tid)))
    (if (negative? info)
      (pvm-error info pvm-kill)
      #t)))

(define (pvm-lookup name index)
  (let* ((data '(#t))
         (cc (p-lookup name index data)))
    (if (negative? cc)
      (pvm-error cc 'pvm-lookup)
      (car data))))

(define (pvm-lvgroup group)
  (let ((info (p-lvgroup group)))
    (if (negative? info)
      (pvm-error info 'pvm-lvgroup)
      #t)))

(define (pvm-mcast msgtag . tids)
  (let* ((tids (if (pair? (car tids)) (car tids) tids))
         (ntask (length tids))
         (info (p-mcast tids ntask msgtag)))
    (if (negative? info)
      (pvm-error info 'pvm-mcast)
      #t)))

(define (pvm-mkbuf . encoding)
  (let ((bufid (p-mkbuf (if (null? encoding)
                          PvmDataDefault
                          (car encoding)))))
    (if (negative? bufid)
      (pvm-error bufid 'pvm-mkbuf)
      bufid)))

(define (pvm-mstat host)
  (let ((mstat (p-mstat host)))
    (case mstat
      ((0) 'PvmOk)
      ((-6) 'PvmNoHost)
      ((-22) 'PvmHostFail)
      (else (pvm-error mstat 'pvm-mstat)))))
                          
(define (pvm-mytid)
  (let ((tid (p-mytid)))
    (if (negative? tid)
      (pvm-error tid 'pvm-mytid)
      tid)))

(define (pvm-notify what msgtag . tids)
  (let* ((tids (if (pair? (car tids)) 
                 (car tids) 
                 tids))
         (ntask (if (= what PvmHostAdd)
                  (car tids)
                  (length tids)))
         (info (p-notify what msgtag ntask tids)))
    (if (negative? info)
      (pvm-error info 'pvm-notify)
      #t)))

(define (pvm-nrecv tid msgtag)
  (if (not tid) (set! tid -1))
  (if (not msgtag) (set! msgtag -1))
  (let ((bufid (p-nrecv tid msgtag)))
    (if (negative? bufid)
      (pvm-error bufid 'pvm-nrecv)
      (if (zero? bufid)
        #f
        bufid))))

(define (pvm-pk sp)
  (let ((info (pvm-pack sp)))
    (if (negative? info)
      (pvm-error info 'pvm-pk)
      #t)))

(define (pvm-pkstr sp)
  (let ((info (p-pkstr sp)))
    (if (negative? info)
      (pvm-error info 'pvm-pkstr)
      #t)))

(define (pvm-pack sp)
  (let ((port (open-output-string)))
    (write-pvm sp port)
    (p-pkstr (get-output-string port))))

(define (pvm-pkfloat ip)
  (let ((info (if (pair? ip) 
                (p-pkfloat ip (length ip) 1)
                (p-pkfloat (list ip) 1 1))))
    (if (negative? info)
      (pvm-error info 'pvm-pkfloat)
      #t)))
        
(define (pvm-pkfloat# ip . arg)
  (let* ((nitem (if (null? arg) 
                  (vector-length ip)
                  (1+ (- (cadr arg) (car arg)))))
         (info (p-pkfloat2 ip nitem 1 (if (null? arg) 
                                        0
                                        (car arg)))))
    (if (negative? info)
      (pvm-error info 'pvm-pkfloat#)
      #t)))
        
(define (pvm-pkfix ip)
  (let ((info (if (pair? ip) 
                (p-pkfix ip (length ip) 1)
                (p-pkfix (list ip) 1 1))))
    (if (negative? info)
      (pvm-error info 'pvm-pkfix)
      #t)))
        
(define (pvm-pkfix# ip . arg)
  (let* ((nitem (if (null? arg) 
                  (vector-length ip)
                  (1+ (- (cadr arg) (car arg)))))
         (info (p-pkfix2 ip nitem 1 (if (null? arg) 
                                      0
                                      (car arg)))))
    (if (negative? info)
      (pvm-error info 'pvm-pkfix#)
      #t)))
        
(define (pvm-parent)
  (let ((tid (p-parent)))
    (if (negative? tid)
      (if (= tid -23)
        'PvmNoParent
        (pvm-error tid 'pvm-parent))
      tid)))

(define (pvm-perror msg)
  (let ((info (p-perror msg)))
    (if (negative? info)
      (pvm-error info 'pvm-perror)
      #t)))

(define (pvm-precvfix tid msgtag . arg)
  (let* ((buf '())
         (atid '(#f))
         (atag '(#f))
         (alen '(#f))
         (info 0)
         (len (if (null? arg)
                1
                (car arg)))
         (buf (make-list len)))
    (if (not tid) (set! tid -1))
    (if (not msgtag) (set! msgtag -1))
    (set! info (p-precv tid msgtag buf len PvmFix atid atag alen))
    (if (negative? info)
      (pvm-error info 'pvm-precvfix)
      (if (null? arg)
        (car buf)
        buf))))

(define (pvm-precvfloat tid msgtag . arg)
  (let* ((buf '())
         (atid '(#f))
         (atag '(#f))
         (alen '(#f))
         (info 0)
         (len (if (null? arg)
                1
                (car arg)))
         (buf (make-list len)))
    (if (not tid) (set! tid -1))
    (if (not msgtag) (set! msgtag -1))
    (set! info (p-precv tid msgtag buf len PvmFloat atid atag alen))
    (if (negative? info)
      (pvm-error info 'pvm-precvfloat)
      (if (null? arg)
        (car buf)
        buf))))

(define (pvm-precvfix# tid msgtag len)
  (let* ((buf '())
         (atid '(#f))
         (atag '(#f))
         (alen '(#f))
         (info 0)
         (buf (make-vector len)))
    (if (not tid) (set! tid -1))
    (if (not msgtag) (set! msgtag -1))
    (set! info (p-precv tid msgtag buf len PvmFix atid atag alen))
    (if (negative? info)
      (pvm-error info 'pvm-precvfix#)
      (if (null? arg)
        (vector-ref  buf 0)
        buf))))

(define (pvm-precvfloat# tid msgtag len)
  (let* ((buf '())
         (atid '(#f))
         (atag '(#f))
         (alen '(#f))
         (info 0)
         (buf (make-vector len)))
    (if (not tid) (set! tid -1))
    (if (not msgtag) (set! msgtag -1))
    (set! info (p-precv tid msgtag buf len PvmFloat atid atag alen))
    (if (negative? info)
      (pvm-error info 'pvm-precvfloat#)
      buf)))

(define (pvm-probe tid msgtag)
  (let ((bufid (p-probe tid msgtag)))
    (if (negative? bufid)
      (pvm-error bufid 'pvm-probe)
      (if (zero? bufid)
        #f
        bufid))))

(define (pvm-psendfix tid msgtag buf)
  (let ((info (if (pair? buf)
                (p-psend tid msgtag buf (length buf) PvmFix)
                (p-psend tid msgtag (list buf) 1 PvmFix))))
    (if (negative? info)
      (pvm-error info 'pvm-psendfix)
      #t)))

(define (pvm-psendfix# tid msgtag buf . arg)
  (let* ((nitem (if (null? arg)
                  (vector-length buf)
                  (1+ (- (cadr arg) (car arg)))))
         (info (p-psend2 tid msgtag buf nitem PvmFix (if (null? arg)
                                                       0
                                                       (car arg)))))
    (if (negative? info)
      (pvm-error info 'pvm-psendfix#)
      #t)))

(define (pvm-psendfloat tid msgtag buf)
  (let ((info (if (pair? buf)
                (p-psend tid msgtag buf (length buf) PvmFloat)
                (p-psend tid msgtag (list buf) 1 PvmFloat))))
    (if (negative? info)
      (pvm-error info 'pvm-psendflaot)
      #t)))

(define (pvm-psendfloat# tid msgtag buf . arg)
  (let* ((nitem (if (null? arg)
                  (vector-length buf)
                  (1+ (- (cadr arg) (car arg)))))
         (info (p-psend2 tid msgtag buf nitem PvmFloat (if (null? arg)
                                                         0
                                                         (car arg)))))
    (if (negative? info)
      (pvm-error info 'pvm-psendfloat#)
      #t)))

(define (pvm-pstat tid)
  (let ((status (p-pstat tid)))
    (case status
      ((0) 'PvmOk)
      ((-31) 'PvmNoTask)
      ((-2) 'PvmBadParam)
      (else (pvm-error status 'pvm-pstat)))))
                     
(define (pvm-recv tid msgtag)
  (if (not tid) (set! tid -1))
  (if (not msgtag) (set! msgtag -1))
  (let ((bufid (p-recv tid msgtag)))
    (if (negative? bufid)
      (pvm-error bufid 'pvm-recv)
      bufid)))

(define (pvm-recvf new-function)
  (let ((old-function (p-recvf new-function)))
    (if (number? old-function)
      #f
      old-function)))

(define (pvm-reducefix func data msgtag group root)
  (let* ((flag (number? data))
         (data (if (number? data)
                 (list data)
                 data))
         (count (if (pair? data)
                  (length data)
                  (if (vector? data)
                    (vector-length data))))
         (info (p-reduce func data count PvmFix msgtag group root)))
    (if (negative? info)
      (pvm-error info 'pvm-reducefix)
      (if (= (pvm-getinst group (mytid)) root)
        (if flag (car data) data)
        #t))))

(define (pvm-reducefloat func data msgtag group root)
  (let* ((flag (number? data))
         (data (if (number? data)
                 (list data)
                 data))
         (count (if (pair? data)
                  (length data)
                  (if (vector? data)
                    (vector-length data))))
         (info (p-reduce func data count PvmFloat msgtag group root)))
    (if (negative? info)
      (pvm-error info 'pvm-reducefloat)
      (if (= (pvm-getinst group (mytid)) root)
        (if flag (car data) data)
      #t))))

(define (pvm-reg-hoster)
  (let ((info (p-reg-hoster)))
    (if (negative? info)
      (pvm-error info 'pvm-reg-hoster)
      #t)))

(define (pvm-reg-tasker)
  (let ((info (p-reg-tasker)))
    (if (negative? info)
      (pvm-error info 'pvm-reg-tasker)
      #t)))

(define (pvm-send tid msgtag)
  (let ((info (p-send tid msgtag)))
    (if (negative? info)
      (pvm-error info 'pvm-send)
      #t)))

(define (pvm-sendsig tid signum)
  (let ((info (p-sendsig tid signum)))
    (if (negative? info)
      (pvm-error info 'pvm-sendsig)
      #t)))

(define (pvm-setopt what val)
  (let ((oldval (p-setopt what val)))
    (if (negative? oldval)
      (pvm-error oldval 'pvm-setopt)
      (case what
        ((1) (case oldval
               ((1) 'PvmDontRoute)
               ((2) 'PvmAllowDirect)
               ((3) 'PvmRouteDirect)
               (else (pvm-error oldval 'pvm-setopt))))
        ((2 3 9) (case oldval
                   ((0) 'off)
                   ((1) 'on)))
        (else oldval)))))

(define (pvm-setrbuf bufid)
  (let ((oldbuf (p-setrbuf (if bufid 0 bufid))))
    (if (negative? oldbuf)
      (pvm-error oldbuf 'pvm-setrbuf)
      (if (zero? oldbuf)
        #f
        oldbuf))))

(define (pvm-setsbuf bufid)
  (let ((oldbuf (p-setsbuf (if bufid 0 bufid))))
    (if (negative? oldbuf)
      (pvm-error oldbuf 'pvm-setsbuf)
      (if (zero? oldbuf)
        #f
        oldbuf))))

(define (pvm-spawn task argv flag where ntask)
  (let* ((tids (make-list ntask #f))
         (numt (p-spawn task argv flag where ntask tids)))
    (if (= ntask numt)
      tids
      (do ((child tids (cdr child))
           (error-num 0))
          ((or (negative? error-num) (null? child))
           (pvm-error error-num 'pvm-spawn))
        (if (negative? (car child))
          (set! error-num (car child)))))))

(define (pvm-spawntus func flag where ntask)
  (let* ((tids (make-list ntask #f))
         (task "pvmtus")
         (argv (list (string-append "-e(begin (pvm-recv -1 -1) (eval (pvm-upk)) (pvm-exit))")))
         (numt (p-spawn task argv flag where ntask tids)))
    (if (= ntask numt)
     (begin (pvm-initsend 0)
            (pvm-pk func)
            (pvm-mcast 1 tids)
            tids)
    (do ((child tids (cdr child))
         (error-num 0))
        ((or (negative? error-num) (null? child))
         (pvm-error error-num 'pvm-spawntus))
      (if (negative? (car child))
        (set! error-num (car child)))))))
 
(define (pvm-tasks . where)
  (let* ((ntask '(#f))
         (where (if (null? where)
                  0 
                  (car where)))
         (info (p-tasks where ntask '())))
    (if (negative? info)
      (pvm-error info 'pvm-tasks)
      (let* ((taskp (do ((x (car ntask) (- x 1))
                         (ret '() (cons (make-list 6 #f) ret)))
                      ((= x 0) ret)))
             (info (p-tasks where ntask taskp)))
    (if (negative? info)
          (pvm-error info 'pvm-tasks)
          (append ntask taskp))))))

(define (pvm-tidtohost tid)
  (let ((dtid (p-tidtohost tid)))
    (if (negative? dtid)
      (pvm-error dtid 'pvm-tidtohost)
      dtid)))

(define (pvm-trecv tid msgtag sec usec)
  (if (not tid) (set! tid -1))
  (if (not msgtag) (set! msgtag -1))
  (let ((bufid (p-trecv tid msgtag (list sec usec))))
    (if (negative? bufid)
      (pvm-error bufid 'pvm-trecv)
      bufid)))

(define (pvm-upk)
  (read (open-input-string (p-upkstr2))))

(define (pvm-upkstr)
  (p-upkstr2))

(define (pvm-upkfloat . arg)
  (let* ((nitem (if (null? arg)
                  1
                  (car arg)))
         (fp (make-list nitem))
         (info (p-upkfloat fp nitem 1)))
    (if (negative? info)
      (pvm-error info 'pvm-upkfloat)
      (if (null? arg)
        (car fp)
        fp))))

(define (pvm-upkfloat# nitem)
  (let ((fp (make-vector nitem))
        (info (p-upkfloat fp nitem 1)))
    (if (negative? info)
      (pvm-error info 'pvm-upkfloat#)
      fp)))

(define (pvm-upkfix . arg)
  (let* ((nitem (if (null? arg)
                  1
                  (car arg)))
         (ip (make-list nitem))
         (info (p-upkfix ip nitem 1)))
    (if (negative? info)
      (pvm-error info 'pvm-upkfix)
      (if (null? arg)
        (car ip)
        ip))))
    
(define (pvm-upkfix# nitem)
  (let ((ip (make-vector nitem))
        (info (p-upkfix ip nitem 1)))
    (if (negative? info)
      (pvm-error info 'pvm-upkfix#)
      ip)))
    
;;;;;;;;;;;;;;;;;;;;;;;;;; PVM Error Functions ;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (pvm-error info . arg)
  (if (null? arg)
    (error (pvm-error-symbol info))
    (error (make-condition (pvm-error-symbol info) 'function (car arg)))))

(define (change-num-pvmsym data)
  (let ((ret '()))
    (for-each (lambda(num)
                (set! ret (cons (let ((sym (pvm-error-symbol num)))
                                  (if (symbol? sym) 
                                    sym
                                    'PvmOk))
                            ret))) data)
    (reverse ret)))

(define (pvm-error-symbol info)
  (case info
    ((0)   'PvmOk)
    ((-2)  'PvmBadParam)
    ((-3)  'PvmMismatch)
    ((-5)  'PvmNoData)
    ((-6)  'PvmNoHost)
    ((-7)  'PvmNoFile)
    ((-10) 'PvmNoMem)
    ((-12) 'PvmBadMsg)
    ((-14) 'PvmSysErr)
    ((-15) 'PvmNoBuf)
    ((-16) 'PvmNoSuchBuf)
    ((-17) 'PvmNullGroup)
    ((-18) 'PvmDupGroup)
    ((-19) 'PvmNoGroup)
    ((-20) 'PvmNotInGroup)
    ((-21) 'PvmNoInst)
    ((-22) 'PvmHostFail)
    ((-23) 'PvmNoParent)
    ((-24) 'PvmNotImpl)
    ((-25) 'PvmDSysErr)
    ((-26) 'PvmBadVersion)
    ((-27) 'PvmOutOfRes)
    ((-28) 'PvmDupHost)
    ((-29) 'PvmCantStart)
    ((-30) 'PvmAlready)
    ((-31) 'PvmNoTask)
    ((-32) 'PvmNoEntry)
    ((-33) 'PvmDupEntry)
    (else info)))

;;;;;;;;;;;;;;;;;;;;;;;;;; Other Functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (rfix targetnum len fil)
  (let ((target (if (number? targetnum) 
                   (number->string targetnum) targetnum)))
    (if (> (string-length target) len) target
      (string-append
        (make-string (- len (string-length target)) fil)
        target))))    

(define (make-list . arg)
  (let ((num (car arg))
        (arg (if (null? (cadr arg))
               #f
               (cadr arg))))
    (do ((x num (- x 1))
         (ret '() (cons arg ret)))
        ((= x 0) ret))))

;;;;;;;;;;;;;;;;;;;;;;;;; PVM Console Functions ;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (pvm-psl . arg)
  (if (null? arg)
    (pvm-ps 0 #t)
    (pvm-ps (car arg) #t)))

(define (pvm-ps . arg)
  (let ((ntask '(#f)) 
        (where (if (null? arg) 0 (car arg)))
        (num '(#f))
        (arg (if (eq? (cadr arg) #t) 0 1)))
    (p-config num '() '())
    (if (>= (p-tasks where ntask '()) 0)
      (let ((taskp (do ((x (car ntask) (- x 1))
                        (ret '() (cons (make-list 6 #f) ret)))
                     ((= x 0) ret)))
            (hostp (do ((x (car num) (- x 1))
                        (ret '() (cons (make-list 4 #f) ret)))
                     ((= x 0) ret)))
            (arc '(#f)))
        (p-config () arc hostp)
        (p-tasks where ntask taskp)
        (display "                     HOST      TID")
        (if (zero? arg) 
          (display "     PTID    PID"))
        (display "   FLAG 0x COMMAND") (newline)
        (for-each (lambda (task)
           (display "                 ")		  
           (for-each (lambda (host)
               (if (= (caddr task) (car host))
                 (display (rfix (cadr host) 8 #\ )))) hostp)       
           (display (rfix (number->string (car task) 16) 9 #\ ))
           (if (zero? arg)
             (begin 
               (if (zero? (cadr task))
                 (display "        -")
                 (display (rfix (number->string (cadr task) 16) 9 #\ )))
               (display (rfix (number->string (cadr (cddddr task))) 7 #\ ))))
           (display " ")
           (let ((stat (string-append (number->string (cadddr task) 16) "/"))
                 (sc (cadddr task))
                 (code '((1024 . "T")
                         ( 512 . "H")
                         (   4 . "c")
                         (   2 . "f"))))
             (for-each (lambda (data)
                 (if (not (zero? (logand (car data) sc)))
                   (begin (set! stat (string-append stat (cdr data)))
                          (set! sc (- sc (car data)))
                          (if (not (zero? sc))
                            (set! stat (string-append stat  ","))))))
               code)
             (if (< (string-length stat) 9)
               (display (make-string (- 9 (string-length stat)))))
             (display stat))
           (display " ")
           (if (zero? (string-length (car (cddddr task))))
             (display "-");
             (display (car (cddddr task))))
           (newline)) taskp)))))
  
(define (pvm-conf)
  (let ((num '(#f)))
    (p-config num '() '())
    (let ((hostp (do ((x (car num) (- x 1))
                      (ret '() (cons (make-list 4 #f) ret)))
                   ((= x 0) ret)))
          (arc '(#f)))
          (p-config () arc hostp)
      (format #t "~a hosts, ~a data format~%" (car num) (car arc))
      (format #t "                     HOST     DTID     ARCH   SPEED~%")
      (for-each (lambda (host)
		  (format #t "                 ~a" (rfix (cadr host) 8 #\ ))
		  (format #t "~a" (rfix (number->string (car host) 16) 9 #\ ))
		  (format #t "~a" (rfix (caddr host) 9 #\ )) 
		  (format #t "    ~a~%" (cadddr host))) hostp)
      #t)))

(define (pvm-add host)
  (let ((host (cond ((string? host) host)
                    ((symbol? host) (symbol->string host))
                    (else #f)))
	(dtid '(#f)))
    (if host 
      (let ((flag (if (positive? (p-addhosts (list host) 1 dtid)) 1 0)))
        (format #t "~a successful~%" flag) 
        (format #t "                    HOST     DTID~%") 
        (format #t "                ~a" (rfix host 8 #\ ))
        (if (zero? flag)
          (if (eq? (pvm-mstat host) 'PvmOk)
            (format #t " Duplicate host~%")
            (format #t " Can't start pvmd~%"))
          (format #t "~a~%" (rfix (number->string (car dtid) 16) 9 #\ ))))
      (format #t "0 successful~%"))
    #t))

(define (pvm-del host)
  (let ((host (cond ((string? host) host)
                    ((symbol? host) (symbol->string host))
                    (else #f))))
    (if host 
      (let ((flag (if (positive? (p-delhosts (list host) 1 '())) 1 0)))
        (format #t "~a successful~%" flag) 
        (format #t "                    HOST  STATUS~%") 
        (format #t "                ~a" (rfix host 8 #\ ))
        (if (zero? flag)
          (format #t "  No such host~%")
          (format #t "  deleted~%")))
      (format #t "0 successful~%"))
    #t))

(define (kill-pvm-child-process)
  (let ((all-process (cdr (pvm-tasks 0)))
        (kill-process '()))
    (for-each (lambda (child)
                (if (equal? "pvmtus" (car (cddddr child)))
                  (set! kill-process (cons (car child) kill-process))))
      all-process)
    (for-each (lambda (child-id)
                (pvm-sendsig child-id SIGKILL)) kill-process)))
                
;(let ((pvm-bye bye))
; (set! bye (lambda() (pvm-exit) (pvm-bye))))

;;;;;;;;;;;;;;;;;;;;;; PVM On-demand Functions ;;;;;;;;;;;;;;;;;;;;;;;;;

(define *pvm-on-demand-tag* 33554430)
(define *pvm-on-demand* #f)
(define *pvm-on-demand-print* #f)

(define (pvm-get-unbound-obj obj)
  (let ((p-tid (pvm-parent)))
    (if (not (eq? p-tid 'PvmNoParent))
      (let ((bufid (pvm-mkbuf 0)))
        (pvm-setrbuf bufid)
        (pvm-joingroup2 "pvm-unbound-obj")
        (pvm-sendsig p-tid SIGUSR1)
        (pvm-recv p-tid *pvm-on-demand-tag*)
        (pvm-upkfix)
        (pvm-initsend 0)
        (pvm-pk obj)
        (pvm-send p-tid *pvm-on-demand-tag*)
        (pvm-recv p-tid *pvm-on-demand-tag*)
        (let ((x (pvm-upk)))
          (if (car x)
            (if (eq? (cadr x) 'unbound)
              (interrupt-pvm-parent (string-append (symbol->string obj) 
                                      " is an unbound symbol") '())
              (bctran-exec-form (child-define (car x) (cdr x)) '()))
            (if (function? (cddr x))
              (put-closure-env (cddr x) (cadr x))))
          (pvm-lvgroup2 "pvm-unbound-obj")
          (pvm-initsend 0)
          (pvm-pkfix (pvm-mytid))
          (pvm-send p-tid *pvm-on-demand-tag*)
          (pvm-freebuf bufid)
          (cddr x))))))

(define (pvm-put-unbound-obj)
  (if (negative? (pvm-getinst2 "pvm-unbound-obj" (pvm-mytid)))
    (pvm-joingroup2 "pvm-unbound-obj"))
  (let* ((child-tids (serch-child-process (cdr (pvm-tasks 0)) (pvm-mytid)))
         (group-tids (serch-group-process child-tids)))
    (if (not (null? group-tids))
      (send-unbound-to-child group-tids)))
  (if (negative? (pvm-getinst2 "pvm-unbound-obj" (pvm-mytid)))
    (pvm-lvgroup2 "pvm-unbound-obj")))

(define (send-unbound-to-child group-tids)
  (let ((child (car group-tids))
        (bufid (pvm-mkbuf 0)))
    (pvm-setrbuf bufid)
    (pvm-initsend 0)
    (pvm-pkfix (pvm-mytid))
    (pvm-send child *pvm-on-demand-tag*)
    (pvm-recv child *pvm-on-demand-tag*)
    (let* ((x (pvm-upk))
           (value (if (unbound? x)
                    (if (eq? (pvm-parent) 'PvmNoParent)
                      'unbound
                      (pvm-get-unbound-obj x))
                    (get-symbol-value x)))
           (ret (if (function? value)
                  (if (sysfun? x)
                    (cons x (cons 'get-gval 
                              (string->symbol (get-fun-name value))))
                    (cons #f (cons (get-closure-env value)
                               (change-pvm-l-t-e-sym value))))
                  (if (and (unbound? x) (eq? (pvm-parent) 'PvmNoParent))
                    (cons x (cons 'unbound value))
                    (cons x (cons 'get-const value))))))
      (pvm-initsend 0)
      (pvm-pk ret)
      (pvm-send child *pvm-on-demand-tag*)
      (if *pvm-on-demand-print*
        (format #t "++++++++++ send +++++++++++ : ~a~%" child))
      (pvm-recv child *pvm-on-demand-tag*)
      (pvm-upkfix)
      (pvm-freebuf bufid)))
  (let* ((tmp-child-tids (serch-child-process (cdr (pvm-tasks 0)) 
                           (pvm-mytid)))
         (tmp-group-tids (serch-group-process tmp-child-tids)))
    (format #t "===== tmpgroup-tids : ~a~%~%" tmp-group-tids)    
    (if (not (null? tmp-group-tids))
      (send-unbound-to-child tmp-group-tids))))
  
(define (serch-child-process data num)
  (let ((ret '()))
    (for-each (lambda (child)
                (if (and (cadr child) (= (cadr child) num)
                         (equal? (car (cddddr child)) "pvmtus"))
                  (set! ret (cons (car child) ret)))) data)
    ret))

(define (serch-group-process childs)
  (let ((ret '()) (tmp '()))
    (for-each (lambda (child)
                (let ((num (pvm-getinst2 "pvm-unbound-obj" child)))
                  (if (>= num 0)
                    (set! tmp (cons num tmp))))) childs)
    (for-each (lambda (child)
                (let ((num (pvm-gettid "pvm-unbound-obj" child)))
                  (set! ret (cons num ret)))) tmp)
    ret))