#include "scheme.h"
#ifdef SHE
#include <stdlib.h>
#include <unistd.h>

#undef DEBUG_MUTEX
#undef DEBUG_STEAL

void free_stolen_frame(void);

/*
 * iretaddrx for Sequential Evaluation mode of future/stealable
 */

int iretaddrx_she_seq_fut()
{
    ENTITYP fut;
    sst_frame_t *sstfr;
#if 1                        /* debug */
    if (GETtag(sp_reg[-5]) != LTAG_FUTURE)
        fprintf(stderr, "Error in iretaddrx_she_seq_fut()\n"
                "sp_reg[-5] = %x\n", sp_reg[-5]);
#endif
    fut = (ENTITYP)ADDRESS(sp_reg[-5]);
    sstfr = fut->future.sstfr;
    if (sstfr != NULL) {        /* not killed */
        mutex_lock(sstfr->frame_lock);
        fut->future.value.obj = reg_reg;
        fut->future.sstfr = NULL;
        set_sstfr_status(sstfr, DETERMINED); /* ??? */
        condition_broadcast(sstfr->cond_obj); 
        mutex_unlock(sstfr->frame_lock);
        free_sst_frame(sstfr);
    }
    reg_reg = sp_reg[-6];       /* restore reg_reg */
    bcpc_reg = SCPOP();
    bcbody_reg = (unsigned *)ADDRESS(SCPOP());
}

int iretaddrx_she_seq_st()
{
    sst_frame_t *sstfr;
#if 1                        /* debug */
    if (GETtag(sp_reg[-5]) != LTAG_STEALABLE)
        fprintf(stderr, "Error in iretaddrx_she_seq_st()\n");
#endif
    sstfr = (sst_frame_t *)ADDRESS(sp_reg[-5]);
    mutex_lock(sstfr->frame_lock);
    sstfr->result.obj = reg_reg;
    reg_reg = sp_reg[-6];       /* restore reg_reg */
    set_sstfr_status(sstfr, DETERMINED);
    condition_broadcast(sstfr->cond_obj);
    mutex_unlock(sstfr->frame_lock);
    bcpc_reg = SCPOP();
    bcbody_reg = (unsigned *)ADDRESS(SCPOP());
}

/*
 * SST frame allocator
 */

sst_frame_t *real_alloc_sst_frame(int which_par_constr)
{
    sst_frame_t *sstfr;

    if ((sstfr = (sst_frame_t *)malloc(sizeof(sst_frame_t))) == NULL) {
        fprintf(stderr, "Cannot allocate SST frame.\n");
        exit(1);
    }
    sstfr->frame_lock = mutex_alloc();
    sstfr->cond_obj = condition_alloc();
    sstfr->func_obj.obj = LTAG_NIL;
    sstfr->result.obj = LTAG_NIL;
#ifdef SHE_ENV
    sstfr->env_array = NULL;
#endif
    sstfr->env_list.obj = LTAG_NIL;
    sstfr->link = NULL;
    return sstfr;
}

void init_sst_frame_free_list()
{
    int i;
    sst_frame_t *tmp;
    
    tmp = real_alloc_sst_frame(FUTURE_CONST); /* dummy sst-frame */
    tmp->link = tmp;
    thr_data[PT_2_HEAD] = (unsigned int)(tmp);
    thr_data[PT_2_TAIL] = (unsigned int)(tmp);

    for(i = 0;i < 300; i++){
        tmp = real_alloc_sst_frame(FUTURE_CONST);
        tmp->link = (sst_frame_t *)(thr_data[PT_2_HEAD]);
        thr_data[PT_2_HEAD] = (unsigned int)(tmp);
    }
}

#ifdef SHE_ENV
sst_frame_t *alloc_sst_frame(int which_par_constr, unsigned int func,
                             unsigned int pc1, unsigned int pc2,
                             unsigned int init_val, int nargs, int mode,
                             unsigned int env)
#else
sst_frame_t *alloc_sst_frame(int which_par_constr, unsigned int func,
                             unsigned int pc1, unsigned int pc2,
                             unsigned int init_val, unsigned int env)
#endif
{
    sst_frame_t *sstfr;

    if (thr_data[PT_2_HEAD] == thr_data[PT_2_TAIL])
        sstfr = real_alloc_sst_frame(which_par_constr);
    else {
        sstfr = (sst_frame_t *)(thr_data[PT_2_HEAD]);
        thr_data[PT_2_HEAD] = (unsigned int)(sstfr->link);
        sstfr->link = NULL;
    };

    init_sstfr_status(sstfr, which_par_constr);
    sstfr->func_obj.obj = func;
    sstfr->pc1 = pc1;
    sstfr->pc2 = pc2;
    sstfr->result.obj = init_val;
#ifdef SHE_ENV
    sstfr->narg = nargs;
    sstfr->env_array
        = (mode == COND) ? thr_data[PT_TPEV_CO] : thr_data[PT_TPEV_NC];
#endif
    sstfr->env_list.obj = env;
    return sstfr;
}

void free_sst_frame(sst_frame_t *sstfr)
{
    if (sstfr->link == NULL) {
        ((sst_frame_t *)(thr_data[PT_2_TAIL]))->link = sstfr;
        thr_data[PT_2_TAIL] = (unsigned int)(sstfr);
        sstfr->link = sstfr;
    }
}

/*
 * Promote and Kill
 */

void cut_connections_with(child_list_t *ch, unsigned int *parent)
{
    for (; ch != NULL; ch = ch->next) {
#ifdef DEBUG_MUTEX
        printf("%d) lock in cut_connections_with()...\n", cthread_self());
#endif
        mutex_lock((mutex_t)ch->child[LOCK_STACK]);
        if ((unsigned int *)(ch->child[PT_2_PAREN]) == parent)
            ch->child[PT_2_PAREN] = NULL;
#ifdef DEBUG_MUTEX
        printf("%d) unlock in cut_connections_with()\n", cthread_self());
#endif
        mutex_unlock((mutex_t)ch->child[LOCK_STACK]);
    }
    free_child_list((child_list_t *)parent[CHILD_LIST]);
    parent[CHILD_LIST] = NULL;
}

void promote_myself()
{
    switch (thr_data[THR_STATUS]) {
    case SUBSIDIARY_S:
#ifdef DEBUG_MUTEX
        printf("%d) lock in promote_myself()...\n", cthread_self());
#endif
        mutex_lock((mutex_t)thr_data[LOCK_STACK]);
        thr_data[THR_STATUS] = SUBSIDIARY_M;
        cut_connections_with((child_list_t *)thr_data[CHILD_LIST], thr_data);
#ifdef DEBUG_MUTEX
        printf("%d) unlock in promote_myself()\n", cthread_self());
#endif
        mutex_unlock((mutex_t)thr_data[LOCK_STACK]);
        set_reg(LTAG_T);
        break;
    case PRIMARY:
    case SUBSIDIARY_M:
        set_reg(LTAG_FALSE);
        break;
    default:
        bokkan("Fatal error 1 (Spromote())");
    }
}

int Spromote()
{
    SOBJ obj;
    ENTITYP fut;
    sst_frame_t *sstf;

    switch (getargnum()) {
    case 0:
        promote_myself();
        break;
    case 1:
        obj.obj = getarg(1);
        if ((GETtag(obj.obj) != LTAG_FUTURE)) {
            set_reg(LTAG_FALSE);
            return 0;
        }
        fut = (ENTITYP)ADDRESS(obj.obj);
        sstf = fut->future.sstfr;
        if (sstf == NULL)       /* determined */
            set_reg(LTAG_FALSE);
        else if (sstf == (sst_frame_t *)thr_data[PT_2_FRAME])
            promote_myself();
        else if (!promoted_sstfr_p(sstf)) {
            mutex_lock(sstf->frame_lock);
            if (fut->future.sstfr == NULL) {
                set_reg(LTAG_FALSE);
            } else if (!promoted_sstfr_p(sstf)) { /* yet not be promoted */
                set_promote_sstfr(sstf);
                set_reg(LTAG_T);
            }
            mutex_unlock(sstf->frame_lock);
        }   
        break;
    default:
        err_illegal_argument("Too many arguments.");
        break;
    }
    return 0;
}

int Skill()
{
    SOBJ obj;
    ENTITYP fut;
    sst_frame_t *sstf;

    switch (getargnum()) {
    case 1:
	thr_data[KILLED_RES] = getarg(1);
        kill_my_descendant_with_value(getarg(1));
        set_reg(LTAG_T);
        break;
    case 2:
        obj.obj = getarg(2);
        if ((GETtag(obj.obj) != LTAG_FUTURE)) {
            set_reg(LTAG_FALSE);
            return 0;
        }
        fut = (ENTITYP)ADDRESS(obj.obj);
        sstf = fut->future.sstfr;
        if (sstf == NULL)       /* determined */
            set_reg(LTAG_FALSE);
        else if (sstf == (sst_frame_t *)thr_data[PT_2_FRAME]) { /* It's myself */
	    thr_data[KILLED_RES] = getarg(1);
            kill_my_descendant_with_value(getarg(1));
            set_reg(LTAG_T);
        } else {
            mutex_lock(sstf->frame_lock);
            if (fut->future.sstfr == NULL) { /* determined */
                set_reg(LTAG_FALSE);
            } else {
                set_kill_sstfr(sstf);
		sstf->result.obj = getarg(1);
		fut->future.value.obj = getarg(1);
		fut->future.sstfr = NULL;
                set_sstfr_status(sstf, DETERMINED);
		condition_broadcast(sstf->cond_obj);
                set_reg(LTAG_T);
            }
            mutex_unlock(sstf->frame_lock);
        }   
        break;
    default:
        err_illegal_argument("Too many arguments.");
        break;
    }
    return 0;
}

void kill_descendant_of(unsigned int *parent)
{
    child_list_t *ch;

    for (ch = (child_list_t *)parent[CHILD_LIST]; ch != NULL; ch = ch->next) {
#ifdef DEBUG_MUTEX
        printf("%d) lock in kill_descendant_of()...\n", cthread_self());
#endif
        mutex_lock((mutex_t)ch->child[LOCK_STACK]);
        if ((unsigned int *)ch->child[PT_2_PAREN] == parent) {
            ch->child[THR_KILLED] = KILL_BY_PAR;
            kill_descendant_of(ch->child);
        }
#ifdef DEBUG_MUTEX
        printf("%d) unlock in kill_descendant_of()\n", cthread_self());
#endif
        mutex_unlock((mutex_t)ch->child[LOCK_STACK]);
    }
}

void kill_my_descendant(void)
{
    child_list_t *ch;

#ifdef DEBUG_MUTEX
    printf("%d) lock in kill_my_descendant()...\n", cthread_self());
#endif
    mutex_lock((mutex_t)thr_data[LOCK_STACK]);
    thr_data[THR_KILLED] = KILL_BY_PAR;
    kill_descendant_of(thr_data);
#ifdef DEBUG_MUTEX
    printf("%d) unlock in kill_my_descendant()\n", cthread_self());
#endif
    mutex_unlock((mutex_t)thr_data[LOCK_STACK]);
}


void kill_descendant_of_aux(unsigned int *parent,unsigned int result)
{
    child_list_t *ch;

    for (ch = (child_list_t *)parent[CHILD_LIST]; ch != NULL; ch = ch->next) {
#ifdef DEBUG_MUTEX
        printf("%d) lock in kill_descendant_of()...\n", cthread_self());
#endif
        mutex_lock((mutex_t)ch->child[LOCK_STACK]);
        if ((unsigned int *)ch->child[PT_2_PAREN] == parent) {
            ch->child[THR_KILLED] = KILL_BY_PAR;
	    ch->child[KILLED_RES] = result;
            kill_descendant_of_aux(ch->child,result);
        }
#ifdef DEBUG_MUTEX
        printf("%d) unlock in kill_descendant_of()\n", cthread_self());
#endif
        mutex_unlock((mutex_t)ch->child[LOCK_STACK]);
    }
}

void kill_my_descendant_with_value(unsigned int result)
{
    child_list_t *ch;

#ifdef DEBUG_MUTEX
    printf("%d) lock in kill_my_descendant()...\n", cthread_self());
#endif
    mutex_lock((mutex_t)thr_data[LOCK_STACK]);
    thr_data[THR_KILLED] = KILL_BY_PAR;
    kill_descendant_of_aux(thr_data,result);
#ifdef DEBUG_MUTEX
    printf("%d) unlock in kill_my_descendant()\n", cthread_self());
#endif
    mutex_unlock((mutex_t)thr_data[LOCK_STACK]);
}

/*
 * byte codes
 */

FUTURE *mk_plchld()
{
    FUTURE *fut;

    fut = (FUTURE *)memget(sizeof(FUTURE));
    fut->head = LTAG_BLOCK1 | (unsigned int)16;
    fut->value.obj = unbound_obj.obj;
    fut->block.obj = LTAG_BLOCK2 | 4;
    return(fut);
}

int make_future()
{
    FUTURE *fut;
    SOBJ fobj;

    fut = mk_plchld();
    fobj.obj = LTAG_FUTURE | SSERDDA(fut);
#ifdef SHE_ENV
    fut->sstfr
        = alloc_sst_frame(FUTURE_CONST, getarg(0), GETbcarg(1), 0, fobj.obj, GETbcarg(2), NO_COND, reg_reg);
#else
    fut->sstfr
        = alloc_sst_frame(FUTURE_CONST, getarg(0), GETbcarg(1), 0, fobj.obj, reg_reg);
#endif
    push_sst_frame(fut->sstfr, MANDATORY_SST);
    reg_reg = fobj.obj;
#ifdef SHE_ENV
    PCUP(3);
#else
    PCUP(2);
#endif
}

#ifdef SHE_ENV
int make_nil_future()
{
    FUTURE *fut;
    SOBJ fobj;

    fut = mk_plchld();
    fobj.obj = LTAG_FUTURE | SSERDDA(fut);
    fut->sstfr
        = alloc_sst_frame(FUTURE_CONST, getarg(0), GETbcarg(1), 0, fobj.obj, GETbcarg(2), NO_COND, LTAG_NIL);
    push_sst_frame(fut->sstfr, MANDATORY_SST);
    reg_reg = fobj.obj;
    PCUP(3);
}
#endif

int make_future_cond()
{
    FUTURE *fut;
    SOBJ fobj;
    sst_frame_t *sstf;

    fut = mk_plchld();
    fobj.obj = LTAG_FUTURE | SSERDDA(fut);
#ifdef SHE_ENV
    fut->sstfr
        = alloc_sst_frame(FUTURE_CONST, getarg(0), GETbcarg(1), 0, fobj.obj, GETbcarg(2), COND, reg_reg);
#else
    fut->sstfr
        = alloc_sst_frame(FUTURE_CONST, getarg(0), GETbcarg(1), 0, fobj.obj, reg_reg);
#endif
    reg_reg = fobj.obj;
#ifdef SHE_ENV
    PCUP(3);
#else
    PCUP(2);
#endif
}

#ifdef SHE_ENV
int make_nil_future_cond()
{
    FUTURE *fut;
    SOBJ fobj;
    sst_frame_t *sstf;

    fut = mk_plchld();
    fobj.obj = LTAG_FUTURE | SSERDDA(fut);
    fut->sstfr
        = alloc_sst_frame(FUTURE_CONST, getarg(0), GETbcarg(1), 0, fobj.obj, GETbcarg(2), COND, LTAG_NIL);
    reg_reg = fobj.obj;
    PCUP(3);
}
#endif

int make_future_cond1()
{
    sst_frame_t *sstf;

    sstf = ((ENTITYP)(ADDRESS(reg_reg)))->future.sstfr;
    sstf->pc2 = GETbcarg(1);
    push_sst_frame(sstf, SPECULATIVE_SST);
    PCUP(2);
}

int make_stealable()
{
    sst_frame_t *sstf;

#ifdef SHE_ENV
    sstf = alloc_sst_frame(STEALABLE_CONST, getarg(0), GETbcarg(1), 0,
                           LTAG_NIL, GETbcarg(2), NO_COND, reg_reg);
#else
    sstf = alloc_sst_frame(STEALABLE_CONST, getarg(0), GETbcarg(1), 0,
                           LTAG_NIL, reg_reg);
#endif
    push_sst_frame(sstf, MANDATORY_SST);
    reg_reg = LTAG_STEALABLE | SSERDDA(sstf);
#ifdef SHE_ENV
    PCUP(3);
#else
    PCUP(2);
#endif
}

#ifdef SHE_ENV
int make_nil_stealable()
{
    sst_frame_t *sstf;

    sstf = alloc_sst_frame(STEALABLE_CONST, getarg(0), GETbcarg(1), 0,
                           LTAG_NIL, GETbcarg(2), NO_COND, LTAG_NIL);
    push_sst_frame(sstf, MANDATORY_SST);
    reg_reg = LTAG_STEALABLE | SSERDDA(sstf);
    PCUP(3);
}
#endif

int make_stealable_cond()
{
    sst_frame_t *sstf;

#ifdef SHE_ENV
    sstf = alloc_sst_frame(STEALABLE_CONST, getarg(0), GETbcarg(1), 0,
                           LTAG_NIL, GETbcarg(2), COND, reg_reg);
#else
    sstf = alloc_sst_frame(STEALABLE_CONST, getarg(0), GETbcarg(1), 0,
                           LTAG_NIL, reg_reg);
#endif
    reg_reg = LTAG_STEALABLE | SSERDDA(sstf);
#ifdef SHE_ENV
    PCUP(3);
#else
    PCUP(2);
#endif
}

#ifdef SHE_ENV
int make_nil_stealable_cond()
{
    sst_frame_t *sstf;

    sstf = alloc_sst_frame(STEALABLE_CONST, getarg(0), GETbcarg(1), 0,
                           LTAG_NIL, GETbcarg(2), COND, LTAG_NIL);
    reg_reg = LTAG_STEALABLE | SSERDDA(sstf);
    PCUP(3);
}
#endif

int make_stealable_cond1()
{
    sst_frame_t *sstf;

    sstf = (sst_frame_t *)ADDRESS(reg_reg);
    sstf->pc2 = GETbcarg(1);
    push_sst_frame(sstf, SPECULATIVE_SST);
    PCUP(2);
}

#ifdef SHE_ENV

int push_env_lval0()
{
    push_env(FETCH0(GETbcarg(1)), NO_COND);
    PCUP(2);
}

int push_env_lval()
{
    push_env(fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3)), NO_COND);
    PCUP(4);
}

int push_env_lval0_c()
{
    push_env(FETCH0(GETbcarg(1)), COND);
    PCUP(2);
}

int push_env_lval_c()
{
    push_env(fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3)), COND);
    PCUP(4);
}

#endif


/* scheduler from here */ 

/* variables */
int     interval = PRI_BC_INTERVAL;
int     processors;
int     sprocessors;
int     no_of_pus;

struct timespec t;  /* timer for sleep */

thread_info_t    count;            /* spawned thread counter */
thread_info_t    pri_act,sub_act;  /* activity controller */
thread_info_t    suspend_act;      /* activity controller */

thread_data_t ** pthrs; /* primary thread list */
thread_data_t ** sthrs; /* subsidiary thread list */


/* 
 *   return address for  condition execution 
*/

int iretaddrx_con_she_seq_fut(void)
{
    /* add code for checking result */

    if (reg_reg == LTAG_FALSE) {
        thr_data[THR_STATUS] = SUBSIDIARY_S;
    } else if (thr_data[THR_STATUS] == SUBSIDIARY_S){
               thr_data[THR_STATUS] = SUBSIDIARY_M;
	       cut_connections_with((child_list_t *)thr_data[CHILD_LIST], thr_data);
	       thr_data[CHILD_LIST] = NULL; 
	       mutex_lock(pri_act.lock); 
	       pri_act.active++;
	       mutex_unlock(pri_act.lock);

	       if (thr_data[THR_STATE] == RUN) {
	           mutex_lock(pri_act.lock);
		   sub_act.active--;
		   mutex_unlock(pri_act.lock);
               }

    }

    thr_data[COND_EXEC]= NO;
    reg_reg = sp_reg[-5];       /* restore reg_reg */
    bcpc_reg = SCPOP();
    bcbody_reg = (unsigned *)ADDRESS(SCPOP());
}

int iretaddrx_con_she_seq_st(void)
{
    /* add code for checking result */

    if (reg_reg == LTAG_FALSE) {
        thr_data[THR_STATUS] = SUBSIDIARY_S;
    } else if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
               thr_data[THR_STATUS] = SUBSIDIARY_M;
	       cut_connections_with((child_list_t *)thr_data[CHILD_LIST], thr_data);
	       thr_data[CHILD_LIST] = NULL;
	       mutex_lock(pri_act.lock); 
	       pri_act.active++;
	       mutex_unlock(pri_act.lock);

	       if (thr_data[THR_STATE] == RUN) {
	           mutex_lock(pri_act.lock);
		   sub_act.active--;
		   mutex_unlock(pri_act.lock);
               }
    }

    thr_data[COND_EXEC] = NO;
    reg_reg = sp_reg[-5];       /* restore reg_reg */
    bcpc_reg = SCPOP();
    bcbody_reg = (unsigned *)ADDRESS(SCPOP());
}

#define MILISEC 1000000;       /* milisec */

void init_she_scheduler (int thr_num)
{
    int i, j;
    char *pp;
    char *env;

    if (thr_num == 0) {

        count.active = 0;  /* task counter */
        count.lock = mutex_alloc();
        count.wait_cond = condition_alloc();

        pri_act.active = 1; /* primary activity counter controller */
        pri_act.lock = mutex_alloc();
        pri_act.lock = condition_alloc();

        sub_act.active = 0; /* subsidiary activity counter controller */
        sub_act.lock = mutex_alloc();
        sub_act.lock = condition_alloc();

        suspend_act.active = 0; /* suspended task counter */
        suspend_act.lock = mutex_alloc();
        suspend_act.lock = condition_alloc();

        if ((env = getenv ( "PRI_BC_INTERVAL" )) != NULL) {
            interval = atoi(env);
        } 

        t.tv_sec = 0;
        t.tv_nsec = 100 * MILISEC; /* 100 milisec */


        if ((env = getenv ( "SLEEP" )) != NULL) {
            t.tv_nsec = atoi(env) * MILISEC;
        } 

        if ((pp = getenv("PROS")) != NULL) {
            processors = atoi(pp);
        } else {

#ifdef SUNSOL2
            /*
             * get current processor numbers
             */
            processors = (int)sysconf(_SC_NPROCESSORS_ONLN); 
#else
            processors = 2;
#endif
        }
	
	no_of_pus = sprocessors = processors;

        if ((pp = getenv ("SPROS")) != NULL) sprocessors = atoi(pp);
        if ((env = getenv ("NO_OF_PUS")) != NULL) no_of_pus = atoi(env);

#ifdef SUNSOL2
        /*
         * setting concurrency level to the number of processors
         */
        thr_setconcurrency(processors+sprocessors);
#endif
    
        /*
         * allocating primary threads' data (the number of processors)
         */
        if ((pthrs = (thread_data_t **)malloc(processors * sizeof(thread_data_t *))) == NULL)
            bokkan("memory allocation error at init_she_scheduler(): "
                   "Can't allocate memory.");
        /*
         * spawning Helper threads
         */
        printf("\n");

        for (i = 1; i < processors; i++) {
            j = (int)cthread_fork(init_she_scheduler, i);
            printf("preparing PROC[%d]: #%x\n", i, (unsigned int) j );
        }
        printf("1.5:thr_data(0) = %x\n", thr_data);

        /*
         * allocating subsidiary threads' data (the number of processors)
         */
        if ((sthrs = (thread_data_t **)malloc(sprocessors * sizeof(thread_data_t *))) == NULL)
            bokkan("memory allocation error at init_she_scheduler(): "
                   "Can't allocate memory.");

        for (i = 0; i < sprocessors; i++) {
            j = (int)cthread_fork(init_she_subthread, i);
            printf("preparing SPROC[%d]: #%x\n", i, (unsigned int) j );
        }

    }  /* if(thr_num == 0) */

    /*
     * allocate and register thread's data
     */

    shestack_alloc(pthrs,thr_num); 

    printf("3:thr_data(%d) = %x\n", thr_num, thr_data);

    if ((pthrs[thr_num] = (thread_data_t *)malloc(sizeof(thread_data_t))) == NULL)
        bokkan("memory allocation error at init_she_scheduler(): "
               "Can't allocate memory.");

    pthrs[thr_num]->mutex = (mutex_t)mutex_alloc();
    pthrs[thr_num]->thr   = cthread_self();           /* pros' threads list */
    pthrs[thr_num]->stack = (unsigned int **)thr_data; /* pros' stack list */
    
    print_stat(pthrs, thr_num);

    mutex_lock(count.lock);
    count.active++;
    mutex_unlock(count.lock);

    mutex_lock(count.lock);
    if(count.active < (processors+sprocessors)) {
        condition_wait(count.wait_cond,count.lock);
        mutex_unlock(count.lock);
    } else {
        condition_broadcast(count.wait_cond);
        mutex_unlock(count.lock);
    }

    if(thr_num != 0)
        start_bci_pri(thr_num);

    return;
}

void init_she_subthread(int thr_num)
{

    shestack_alloc(sthrs,thr_num); 

    if ((sthrs[thr_num] = (thread_data_t *)malloc(sizeof(thread_data_t))) == NULL)
        bokkan("memory allocation error at init_she_subscheduler(): "
               "Can't allocate memory.");

    sthrs[thr_num]->mutex = (mutex_t)mutex_alloc();
    sthrs[thr_num]->thr   = cthread_self();             /* pros' threads list */
    sthrs[thr_num]->stack = (unsigned int **)thr_data;  /* pros' stack list */


    print_stat(sthrs, thr_num);

    mutex_lock(count.lock);
    count.active++;
    mutex_unlock(count.lock);

    mutex_lock(count.lock);
    if(count.active < (processors+sprocessors)) {
        condition_wait(count.wait_cond,count.lock);
        mutex_unlock(count.lock);
    } else {
        condition_broadcast(count.wait_cond);
        mutex_unlock(count.lock);
    }

    start_bci_sub(thr_num);
}

/* this function need some correction */

void shestack_alloc(thread_data_t **thrs,int  thr_num)
{
    sst_frame_t ** tmp_stack;
    unsigned int * envst;

#ifdef CTHREAD
    REGS *regsp;
#endif

    if((thr_num != 0)||(thrs == sthrs)){
#ifdef USE_REG
        main_thread_regs = (REGS *)malloc(sizeof(REGS));
#else /* not USE_REG */
#ifdef CTHREAD
        regsp = (REGS *)malloc(sizeof(REGS));
        cthread_set_data(cthread_self(), (any_t)regsp);
#endif
#endif /* USE_REG */
        if ((stack_top = (char *)malloc(STACK_SIZE)) == NULL)
            bokkan("memory allocation error at shestack_alloc(): "
                   "Can't allocate memory.");
        stack_bottom = stack_top + STACK_SIZE - 28 - BOTTOM_SZ; /* ??? */
        stack_top = stack_top + GETA_OF_STACK;
#ifdef CALLIOC
        *(unsigned *)(stack_bottom + 4) = (unsigned)(stack_bottom - 12);
        *(unsigned *)(stack_bottom + 8) = 0;
        *(unsigned *)(stack_bottom + 12) = 0;
#else
        *(unsigned *)(stack_bottom + 4) = LTAG_NIL;
        *(stack_bottom + 8) = 0;
        *(stack_bottom + 12) = 0;
#endif
    }
  

    /* initialize bottom area */

    printf("a:thr_data(%d) = %x\n", thr_num, thr_data);
    thr_data[LOCK_STACK] = (unsigned int)mutex_alloc();
    thr_data[PT_2_PAREN] = NULL; /* parent */
    thr_data[CHILD_LIST] = NULL;
    thr_data[PT_2_FRAME] = NULL; /* stolen parent task frame */

    /* SST madatory */

     
    if((tmp_stack = (sst_frame_t **)malloc(sizeof(sst_frame_t *)* SST_SIZE)) == NULL)
        bokkan("memory allocation error at shestack_alloc(): "
               "Can't allocate memory.");

    thr_data[PT_BAS_MAN] = (unsigned int)tmp_stack;
    thr_data[PT_HED_MAN] = (unsigned int)tmp_stack;
    thr_data[PT_TAL_MAN] = (unsigned int)tmp_stack;
       
    /* SST speculative */
    
    if((tmp_stack = (sst_frame_t **)malloc(sizeof(sst_frame_t *)* SST_SIZE)) == NULL)
        bokkan("memory allocation error at shestack_alloc(): "
               "Can't allocate memory.");

    thr_data[PT_BAS_SPE] = (unsigned int)tmp_stack;
    thr_data[PT_HED_SPE] = (unsigned int)tmp_stack;
    thr_data[PT_TAL_SPE] = (unsigned int)tmp_stack;

    if (thrs == pthrs)
        thr_data[THR_STATUS] = PRIMARY;
    else
        thr_data[THR_STATUS] = SUBSIDIARY_S;

    thr_data[THREAD_ID]  = thr_num;
    thr_data[THR_KILLED] = NO;
    thr_data[COND_EXEC]  = NO;
    thr_data[THR_STATE]  = RUN;
    thr_data[KILLED_RES] = LTAG_NIL;

    init_sst_frame_free_list();

#ifdef SHE_ENV

    if((envst = (unsigned int *)malloc(sizeof(unsigned int *) * ENV_STACK_SIZE)) == NULL)
        bokkan("memory allocation error at shestack_alloc(): "
	       "Can't allocate memory.");
    thr_data[PT_BSEV_NC] = (unsigned int) envst;
    thr_data[PT_TPEV_NC] = (unsigned int) envst;

    if((envst = (unsigned int *)malloc(sizeof(unsigned int *) * ENV_STACK_SIZE)) == NULL)
        bokkan("memory allocation error at shestack_alloc(): "
               "Can't allocate memory.");
    thr_data[PT_BSEV_CO] = (unsigned int) envst;
    thr_data[PT_TPEV_CO] = (unsigned int) envst;

    thr_data[PT_ENV_CO] = (unsigned int *)malloc(sizeof(unsigned int) * 20);

#endif

    stack_area_bottom = stack_bottom;  /* what for ?? */

    /* set sp_reg */

    set_top(stack_bottom);

    s_push(iret_she_steal);
    s_push(0);
    s_push(0);

    /* set base & active */
    set_base(top_val());
    set_active(top_val());
    set_a2(LTAG_NIL);
    set_reg(LTAG_NIL);
    more.obj = LTAG_NIL;

    return;
}

/* byte code interpreter for primary thread */

void start_bci_pri(int thr_number)
{

    she_steal();    /*  steal task from other processor's sst stack */
    start_bci_pri_aux(thr_number);     /* interpret byte code */

    return;
}

void start_bci_pri_aux(int thr_number)
{
    int counter;

    /* interpret byte code */
    while (1) {

        counter = 0;
        while (counter++ < interval)
            ((int(*)())(*( bcbody_reg + bcpc_reg )))();

	if (thr_data[PT_2_FRAME] != NULL) {
            /* check if this thread is killed by other thread */
            if (killed_sstfr_p((sst_frame_t *)(thr_data[PT_2_FRAME]))) {
	        sst_frame_t *sstfr;
		sstfr = (sst_frame_t *)thr_data[PT_2_FRAME];
	        kill_my_descendant_with_value(sstfr->result.obj);
		thr_data[KILLED_RES] = sstfr->result.obj;
                thr_data[THR_KILLED] = KILL_BY_PAR;
            }
	}

        /* killed him self if thr_data[THR_KILLED] == KILL_BY_PAR */

        if (thr_data[THR_KILLED] == KILL_BY_PAR) {
            sst_frame_t *sstfr;

            clean_up_sst();

            sstfr = (sst_frame_t *)thr_data[PT_2_FRAME];
            if (sstfr != NULL) {
	        mutex_lock(sstfr->frame_lock);
		set_sstfr_status(sstfr, DETERMINED);
		
		/* wake up thread which is waiting for value */
		if (par_constr(sstfr) == FUTURE_CONST) {
		    ENTITYP fut;
		    fut = (ENTITYP)ADDRESS((unsigned int)(sstfr->result.obj)); 
		    fut->future.sstfr = NULL;
		    if (fut->future.value.obj == LTAG_NIL)
		        fut->future.value.obj = thr_data[KILLED_RES];
		    condition_broadcast(sstfr->cond_obj);
		    free_sst_frame(thr_data[PT_2_FRAME]);
		} else {
		    if (sstfr->result.obj == LTAG_NIL)
		        sstfr->result.obj = thr_data[KILLED_RES];
		    condition_broadcast(sstfr->cond_obj);
		}
		mutex_unlock(sstfr->frame_lock);
	    }


	    free_stolen_frame();
	    clean_up_thr_data();      /* re-initialize thr_data */

            if (thr_data[THREAD_ID] != 0) {
                mutex_lock(pri_act.lock);
		pri_act.active--;
		mutex_unlock(pri_act.lock);
	    }

#if defined(__CYGWIN32__) || defined(linux)
            if (thr_data[THREAD_ID] == 0) siglongjmp(toplevel,1);
#else
            if (thr_data[THREAD_ID] == 0) longjmp(toplevel,1);
#endif
            else  she_steal();
         } 
    }
    return;
}


/* byte code interpreter for subsidiary thread */

void start_bci_sub(int thr_number)
{
    int counter;        /* byte-code step counter */
    int cond_cnt = 0;   /* coundition step counter */

    /*  steal task from other processor's sst stack */
    she_steal();

    /* interpret byte code */

    while (1) {
    cond_check:
        cond_cnt = 0;
        
        do {
	    counter = 0;
	    while (counter++ < interval) {
/*
                if (*(bcbody_reg + bcpc_reg) >= 0x4000000)
                    printf("bcbody_reg = %x, bcpc_reg = %x\n",
                           bcbody_reg, bcpc_reg);
 */
	        ((int(*)())(*( bcbody_reg + bcpc_reg )))();
            }

        } while(thr_data[COND_EXEC] == YES);


    recheck:
        cond_cnt++;

	if (thr_data[PT_2_FRAME] != NULL) {
	    /* check if this thread is killed by other thread */
            if (killed_sstfr_p((sst_frame_t *)(thr_data[PT_2_FRAME]))) {
	        sst_frame_t *sstfr;

		sstfr = (sst_frame_t *)thr_data[PT_2_FRAME];
	        kill_my_descendant_with_value(sstfr->result.obj);
                thr_data[THR_KILLED] = KILL_BY_PAR;
            }
	}

        /* killed him self if thr_data[THR_KILLED] == KILL_BY_PAR */

	if (thr_data[THR_KILLED] == KILL_BY_PAR) {
            sst_frame_t *sstfr;

            clean_up_sst();
	    
            sstfr = (sst_frame_t *)thr_data[PT_2_FRAME];
            mutex_lock(sstfr->frame_lock);
            set_sstfr_status(sstfr, DETERMINED);

            /* wake up thread which is waiting for value */
            if (par_constr(sstfr) == FUTURE_CONST) {
                if (sstfr_status(sstfr) != DETERMINED) {
                    ENTITYP fut;
/*		printf("   killed   \n"); */
                    fut = (ENTITYP)ADDRESS(sstfr->result.obj);
                    fut->future.sstfr = NULL;
                    condition_broadcast(sstfr->cond_obj);
                }
                free_sst_frame(thr_data[PT_2_FRAME]);
            } else {
                condition_broadcast(sstfr->cond_obj);
            }
            mutex_unlock(sstfr->frame_lock);
            
            if (thr_data[THR_STATUS] == SUBSIDIARY_M) {
                mutex_lock(pri_act.lock);
                pri_act.active--;
                mutex_unlock(pri_act.lock);
	    }else {
	          mutex_lock(pri_act.lock);
		  if (thr_data[THR_STATE] == RUN)
		      sub_act.active--;
		  else
		    suspend_act.active--;
		  mutex_unlock(pri_act.lock);
	    }

	    /*	    printf("\n%d is killed ",thr_data[THREAD_ID]);  */

	    thr_data[THR_STATUS] = SUBSIDIARY_S;

	    free_stolen_frame();
 	    clean_up_thr_data();

            she_steal();

        } else {
	    struct timespec r;  

            if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                int idle;
		
		/* check if this thread is promoted by other thread */
		if (promoted_sstfr_p((sst_frame_t *)thr_data[PT_2_FRAME]) == 
		    PROMOTEDP) {
	            cut_connections_with((child_list_t *)(thr_data[CHILD_LIST]),
					 thr_data);
		    thr_data[CHILD_LIST] = NULL;
		    thr_data[THR_STATUS] = SUBSIDIARY_M;

		    mutex_lock(pri_act.lock);
		    pri_act.active++;
		    mutex_unlock(pri_act.lock);

		    if (thr_data[THR_STATE] == RUN) {
		        mutex_lock(pri_act.lock);
		        sub_act.active--;
			mutex_unlock(pri_act.lock);
		    }
		    continue;
		}

                mutex_lock(pri_act.lock);
		if ((idle = no_of_pus - pri_act.active - sub_act.active) > 0) {
                    if (thr_data[THR_STATE] == SUSPEND) {
                        thr_data[THR_STATE]= RUN;
                        sub_act.active++;
                        suspend_act.active--;
                    }
                    mutex_unlock(pri_act.lock);
                } else {

                    if (thr_data[THR_STATE] == RUN) {
                        thr_data[THR_STATE] = SUSPEND;
                        sub_act.active--;
                    } 
                    mutex_unlock(pri_act.lock);
                    nanosleep(&t,&r);

                    if (cond_cnt > 5) {
                        if (thr_data[PT_2_PAREN] == NULL && set_condition()) {
                            thr_data[COND_EXEC] = YES;
                            goto cond_check;
                        } else
                            goto recheck;
                    } else goto recheck;
                }
            }     /*  if(is_subsidiary)                */
	}
    }
    return;
}

int set_condition(void)
{
    sst_frame_t * frame;
    
    frame = (sst_frame_t *)(thr_data[PT_2_FRAME]);
    if (frame->pc2 != 0) {
        if (par_constr(frame) == FUTURE_CONST) {
            open_frame2(iretaddrx_con_she_seq_fut,
                        LTAG_STRING |SSERDDA(bcbody_reg),
                        bcpc_reg);
            s_push(frame->func_obj.obj);
            s_push(reg_reg);
#ifdef SHE_ENV
            copy_env_to_cstack(frame);
            if (frame->env_list.obj != LTAG_NIL)
                s_push(frame->env_list.obj);
#else
            s_push(frame->env_list.obj);
#endif
            active_reg = base_reg;
            bcpc_reg = frame->pc2;
            bcbody_reg =
                (unsigned *) ADDRESS(((ENTITYP)ADDRESS(frame->func_obj.obj))->bcc.body.obj);
        }
        else{
            open_frame2(iretaddrx_con_she_seq_st,
                        LTAG_STRING | SSERDDA(bcbody_reg),
                        bcpc_reg);
            s_push(frame->func_obj.obj);
            s_push(reg_reg);
#ifdef SHE_ENV
            copy_env_to_cstack(frame);
            if (frame->env_list.obj != LTAG_NIL)
                s_push(frame->env_list.obj);
#else
            s_push(frame->env_list.obj);
#endif
            active_reg = base_reg;
            bcpc_reg = frame->pc2;
            bcbody_reg =
                (unsigned *) ADDRESS(((ENTITYP)ADDRESS(frame->func_obj.obj))->bcc.body.obj);
        }
        return 1;
    } else
        return 0;
}

void clean_up_thr_data(void)
{
    thr_data[PT_2_PAREN] = NULL;
    thr_data[CHILD_LIST] = NULL;
    thr_data[PT_2_FRAME] = NULL;
    thr_data[THR_STATE ] = RUN;    
    thr_data[KILLED_RES] = LTAG_NIL;
    thr_data[COND_EXEC ] = NO;
    thr_data[THR_KILLED] = NO;
}

void clean_up_sst(void)
{
  
    unsigned int *headman, *tailman, *headspe, *tailspe;

    /* free-ing stolen task frame */

#ifdef DEBUG_MUTEX
    printf("%d) lock in clean_up_sst()...\n", cthread_self());
#endif
    mutex_lock((mutex_t)thr_data[LOCK_STACK]);
    if(thr_data[PT_HED_MAN] > thr_data[PT_TAL_MAN]){
        sst_frame_t *tframe;

        headman = (unsigned int *)thr_data[PT_HED_MAN];
        tailman = (unsigned int *)thr_data[PT_TAL_MAN];

        /* clean up  mandatory sst */
        /* add 'free task frame' routine here */
        if (thr_data[THR_STATUS] == PRIMARY) {
            if (headman == tailman) {
                thr_data[PT_HED_MAN] = thr_data[PT_TAL_MAN] = thr_data[PT_BAS_MAN];
                thr_data[PT_TPEV_NC] = thr_data[PT_BSEV_NC];
            }
        } else {
            while (--headman >= tailman) {
                tframe = (sst_frame_t *)(*((unsigned int *)(headman)));
                if (tframe != NULL) {
                    if (par_constr(tframe) == FUTURE_CONST) {
                        mutex_lock(tframe->frame_lock);
                        if ((*((unsigned int *)(headspe))) != NULL) {
                            if (killed_sstfr_p(tframe)) {
                                free_sst_frame(tframe);
                            } else {
                                if((sstfr_status(tframe) == DETERMINED) ||
                                   (sstfr_status(tframe) == EXECUTING)) 
                                    printf("oooiiiii\n");
                                ((ENTITYP)ADDRESS(tframe->result.obj))->future.value.obj = thr_data[KILLED_RES];
                                ((ENTITYP)ADDRESS(tframe->result.obj))->future.sstfr = NULL;
                                condition_broadcast(tframe->cond_obj);
                                free_sst_frame(tframe);
                            }
                        }
                        mutex_unlock(tframe->frame_lock);
                    } else {
                        mutex_lock(tframe->frame_lock);
                        if ((*((unsigned int *)(headspe))) != NULL) {
                            if (!killed_sstfr_p(tframe)) {
                                if((sstfr_status(tframe) == DETERMINED) ||
                                   (sstfr_status(tframe) == EXECUTING)) 
                                    printf("oooiiiii\n");
                                tframe->result.obj = thr_data[KILLED_RES];
                                set_sstfr_status(tframe, DETERMINED);
                                condition_broadcast(tframe->cond_obj);
                            }
                        }
                        mutex_unlock(tframe->frame_lock);
                    }
                }
            }
            thr_data[PT_HED_MAN] = thr_data[PT_TAL_MAN] = thr_data[PT_BAS_MAN];
            thr_data[PT_TPEV_NC] = thr_data[PT_BSEV_NC];
        }
    }
#ifdef RMO
# if defined(sparc)
    asm("membar #StoreLoad"); /* for V9 */
# else
    mutex_lock(dummy);
    mutex_unlock(dummy);
# endif
#endif


    if (thr_data[PT_HED_SPE] > thr_data[PT_TAL_SPE]) {
        sst_frame_t *tframe;

        headspe =  (unsigned int *)thr_data[PT_HED_SPE];
	tailspe =  (unsigned int *)thr_data[PT_TAL_SPE];
    
        /* clean up  speculative sst */
        /* add 'free task frame' routine here */
        while (--headspe >= tailspe) {
            tframe = (sst_frame_t *)(*((unsigned int *)(headspe)));
            if (tframe != NULL) {
                if (par_constr(tframe) == FUTURE_CONST) {
                    mutex_lock(tframe->frame_lock);
                    if ((*((unsigned int *)(headspe))) != NULL) {
                        if (killed_sstfr_p(tframe)) {
                            free_sst_frame(tframe);
                        } else {
                            if((sstfr_status(tframe) == DETERMINED) ||
                               (sstfr_status(tframe) == EXECUTING)) 
                                printf("oooiiiii\n");
                            ((ENTITYP)ADDRESS(tframe->result.obj))->future.value.obj = thr_data[KILLED_RES];
                            ((ENTITYP)ADDRESS(tframe->result.obj))->future.sstfr = NULL;
                            set_sstfr_status(tframe, DETERMINED);
                            condition_broadcast(tframe->cond_obj);
                            free_sst_frame(tframe);
                        }
                    }
                    mutex_unlock(tframe->frame_lock);
                } else {
                    mutex_lock(tframe->frame_lock);
                    if ((*((unsigned int *)(headspe))) != NULL) {
                        if (!killed_sstfr_p(tframe)) {
                            if((sstfr_status(tframe) == DETERMINED) ||
                               (sstfr_status(tframe) == EXECUTING)) 
                                printf("oooiiiii\n");
                            tframe->result.obj = thr_data[KILLED_RES];
                            set_sstfr_status(tframe, DETERMINED);
                            condition_broadcast(tframe->cond_obj);
                        }
                    }
                    mutex_unlock(tframe->frame_lock);
                }
            }
        }
    }

    thr_data[PT_HED_SPE] = thr_data[PT_TAL_SPE] = thr_data[PT_BAS_SPE];
    thr_data[PT_TPEV_CO] = thr_data[PT_BSEV_CO];
#ifdef RMO
# if defined(sparc)
    asm("membar #StoreLoad"); /* for V9 */
# else
    mutex_lock(dummy);
    mutex_unlock(dummy);
# endif
#endif

#ifdef DEBUG_MUTEX
    printf("%d) unlock in clean_up_sst()\n", cthread_self());
#endif
    mutex_unlock((mutex_t)thr_data[LOCK_STACK]);

    /* add 'free task frame' routine here */

    if (thr_data[CHILD_LIST] != NULL)
        cut_connections_with((child_list_t *)thr_data[CHILD_LIST], thr_data);

    thr_data[CHILD_LIST] = NULL;
    thr_data[PT_TPEV_CO] = thr_data[PT_BSEV_CO];
    thr_data[PT_TPEV_NC] = thr_data[PT_BSEV_NC];

    thr_data[THR_KILLED] = NO;
}

void free_stolen_frame(void)
{
    if (thr_data[PT_2_FRAME] != NULL) {
        sst_frame_t *tframe;

        tframe = (sst_frame_t *)thr_data[PT_2_FRAME];
        if (par_constr(tframe) == FUTURE_CONST)
            free_sst_frame(tframe);
        thr_data[PT_2_FRAME] = NULL;
    }
}

void iret_she_steal(void)
{
    sst_frame_t *sstfr,*frame;

    if (GETtag(sp_reg[-5]) == LTAG_FUTURE) {
        ENTITYP fut;

        fut = (ENTITYP)ADDRESS(sp_reg[-5]);
        sstfr = fut->future.sstfr;

        if (sstfr != NULL) {    /* not killed */
            mutex_lock(sstfr->frame_lock);
            fut->future.value.obj = reg_reg;
            fut->future.sstfr = NULL;
            condition_broadcast(sstfr->cond_obj);
            mutex_unlock(sstfr->frame_lock);
            free_sst_frame(sstfr);
        }
        reg_reg = sp_reg[-6];       /* restore reg_reg */
	
    } else if (GETtag(sp_reg[-5]) == LTAG_STEALABLE) {
        sstfr = (sst_frame_t *)ADDRESS(sp_reg[-5]);
        mutex_lock(sstfr->frame_lock);
        sstfr->result.obj = reg_reg;
        reg_reg = sp_reg[-6];       /* restore reg_reg */
        set_sstfr_status(sstfr, DETERMINED);
        condition_broadcast(sstfr->cond_obj);
        mutex_unlock(sstfr->frame_lock);
    } else
        fprintf(stderr, "Error in iret_she_steal()\n");

    switch (thr_data[THR_STATUS]) {
    case PRIMARY:
    case SUBSIDIARY_M:
        /*      printf("%d is pop-ing his own stack\n",thr_data[THREAD_ID]);  */
        frame = pop_sst_frame(MANDATORY_SST);
	if ((!(frame)) && (thr_data[THR_STATUS] == SUBSIDIARY_M)) {
	    frame = pop_sst_frame(SPECULATIVE_SST); 
	} 
        break;
    case SUBSIDIARY_S:
        /*      printf("%d is pop-ing his own stack\n",thr_data[THREAD_ID]);  */
        frame = pop_sst_frame(MANDATORY_SST);
	if (!(frame))
	    frame = pop_sst_frame(SPECULATIVE_SST); 
        break;
    default:
        bokkan("Fatal error in she_steal()");
    }

    if (frame) {
        set_top(stack_bottom);
	sp_reg -= 3;
	set_base(top_val());
	set_active(top_val());
	if (par_constr(frame) == FUTURE_CONST){
	    s_push(frame->func_obj.obj);
	    s_push(frame->result.obj);
	    s_push(reg_reg);
#ifdef SHE_ENV
	    pop_env_to_cstack(frame);
	    if (frame->env_list.obj != LTAG_NIL)
	        s_push(frame->env_list.obj);
#else
                s_push(frame->env_list.obj);
#endif
                active_reg = base_reg;
                bcpc_reg = frame->pc1;
                bcbody_reg =
		   (unsigned *) ADDRESS(((ENTITYP)ADDRESS(frame->func_obj.obj))->bcc.body.obj);
	} else{
	    s_push(frame->func_obj.obj);
	    s_push(LTAG_STEALABLE | SSERDDA(frame));
	    s_push(reg_reg);
#ifdef SHE_ENV
	    pop_env_to_cstack(frame);
	    if (frame->env_list.obj != LTAG_NIL)
	        s_push(frame->env_list.obj);
#else
	    s_push(frame->env_list.obj);
#endif
	    active_reg = base_reg;
	    bcpc_reg = frame->pc1;
	    bcbody_reg =
	       (unsigned *) ADDRESS(((ENTITYP)ADDRESS(frame->func_obj.obj))->bcc.body.obj);
	}
	return;
    }

    switch (thr_data[THR_STATUS]) {
    case SUBSIDIARY_M :
        thr_data[THR_STATUS] == SUBSIDIARY_S;
        mutex_lock(pri_act.lock);
        pri_act.active--;
        mutex_unlock(pri_act.lock);
	break;
    case SUBSIDIARY_S :
        if (thr_data[THR_STATE] == RUN) {
            mutex_lock(pri_act.lock);
            sub_act.active--;
            mutex_unlock(pri_act.lock);
        }
	break;
    case PRIMARY :
        mutex_lock(pri_act.lock);
        pri_act.active--;
        mutex_unlock(pri_act.lock);
	break;
    default : break;
    }

    clean_up_sst();
    free_stolen_frame();
    clean_up_thr_data();
    she_steal();
}

void she_steal(void)
{

    struct timespec r;

    /* set sp_reg */
    set_top(stack_bottom);
    sp_reg -= 3;

    /* set base & active */
    set_base(top_val());
    set_active(top_val());



    if (thr_data[THR_STATUS] == SUBSIDIARY_M)  
        thr_data[THR_STATUS] = SUBSIDIARY_S;
    clean_up_thr_data();      /* re-initialize thr_data */
    thr_data[THR_STATE] = SLEEP;

    while (steal_sst() == 0){
      int c;
      nanosleep(&t,&r); 
      c = 0;
    }
    /*
    if(thr_data[THR_STATUS] == PRIMARY)
      printf("\nprimary ");
    else
      printf("\nsubsidiary ");
    printf(" %d stealing ",thr_data[THREAD_ID]); fflush(stdout);
    */
    thr_data[THR_STATE] = RUN;


    if (thr_data[THR_STATUS] == PRIMARY) {
        mutex_lock(pri_act.lock);
        pri_act.active++;
        mutex_unlock(pri_act.lock);
    } else {
      /*        printf("\n%d start working",thr_data[THREAD_ID]); */
        mutex_lock(pri_act.lock);
        sub_act.active++;
        mutex_unlock(pri_act.lock);
    }
}

/* steal_sst() will return 0 if it fail to steal frame */
/* and return 1 if it is succeed to steal frame      */

int steal_sst(void)
{
    sst_frame_t * frame;  
    int rtrn = 0;

    frame = steal_sst_frame();

    if (frame != NULL) {

#ifdef DEBUG_STEAL
        if (thr_data[THREAD_ID] != 0) {
           printf("thread %d had stole task\n",thr_data[THREAD_ID]);
	   fflush(stdout);
	}
#endif
        rtrn = 1;
        if (par_constr(frame) == FUTURE_CONST) {
            s_push(frame->func_obj.obj);
            s_push(frame->result.obj);
            s_push(reg_reg);
#ifdef SHE_ENV
            copy_env_to_cstack(frame);
            if (frame->env_list.obj != LTAG_NIL)
                s_push(frame->env_list.obj);
#else
            s_push(frame->env_list.obj);
#endif
            active_reg = base_reg;
            bcpc_reg = frame->pc1;
            bcbody_reg =
                (unsigned *) ADDRESS(((ENTITYP)ADDRESS(frame->func_obj.obj))->bcc.body.obj);
        } else {
            s_push(frame->func_obj.obj);
            s_push(LTAG_STEALABLE | SSERDDA(frame));
            s_push(reg_reg);
#ifdef SHE_ENV
            copy_env_to_cstack(frame);
            if (frame->env_list.obj != LTAG_NIL)
                s_push(frame->env_list.obj);
#else
            s_push(frame->env_list.obj);
#endif
            active_reg = base_reg;
            bcpc_reg = frame->pc1;
            bcbody_reg =
                (unsigned *) ADDRESS(((ENTITYP)ADDRESS(frame->func_obj.obj))->bcc.body.obj);
        }
    }
    return rtrn;
}

void push_sst_frame(sst_frame_t *object,int mode)
{
    sst_frame_t ** sst_head;
    int thead,tbase;

    if(mode == MANDATORY_SST){
        thead = PT_HED_MAN;
        tbase = PT_BAS_MAN;
    } else {
        thead = PT_HED_SPE;
        tbase = PT_BAS_SPE;
    }
/*
  mutex_lock((mutex_t)(thr_data[tlock]));
  */
    sst_head = (sst_frame_t **)(thr_data[thead]);   

    sst_head[0] = object; 
    object->pstack = (unsigned int)sst_head;

    if((sst_head + 1 > ((sst_frame_t **)thr_data[tbase] + SST_SIZE)))
        bokkan("fatal error at push_sst_frame(): "
               "sst stack overflow \n");
    thr_data[thead] = (unsigned int)(sst_head + 1);
/*
  mutex_unlock((mutex_t)(thr_data[tlock]));
  */
}

/* lock - lock */

sst_frame_t * pop_sst_frame(int mode)
{

    unsigned int * rstk;
    int            thread_id;
    sst_frame_t ** sst_head;
    sst_frame_t  * return_pointer = NULL;
    sst_frame_t *sstfr;
    int            tbase,thead,ttail;
    int            i, top, base;

    if(mode == MANDATORY_SST){
        tbase = PT_BAS_MAN;
        thead = PT_HED_MAN;
        ttail = PT_TAL_MAN;
        base = PT_BSEV_NC;
        top  = PT_TPEV_NC; 
    } else {
        tbase = PT_BAS_SPE;
        thead = PT_HED_SPE;
        ttail = PT_TAL_SPE;
        base = PT_BSEV_CO;
        top  = PT_TPEV_CO; 
    }

    mutex_lock((mutex_t)thr_data[LOCK_STACK]);

#ifdef DEBUG_MUTEX
	printf("%d) lock in pop_sst_frame()...\n", cthread_self());
#endif
    while (thr_data[thead] > thr_data[ttail]){

        ((sst_frame_t **)thr_data[thead])--;
#ifdef RMO
# if defined(sparc)
        asm("membar #StoreLoad"); /* for V9 */
# else
        mutex_lock(dummy);
        mutex_unlock(dummy);
# endif
#endif

	if (thr_data[thead] >= thr_data[ttail]) {

            sstfr = *(sst_frame_t **)thr_data[thead];
	    if (sstfr == NULL) continue;

            mutex_lock((mutex_t)(sstfr)->frame_lock);

	    if (*(sst_frame_t **)thr_data[thead] == NULL) {
                mutex_unlock((mutex_t)(sstfr)->frame_lock);
                continue;
            }

	    if (killed_sstfr_p(sstfr) != NOT_NEEDED) {
		if ((sstfr_status(sstfr) == EXECUTING) ||
		    (sstfr_status(sstfr) == DETERMINED)) {
		    mutex_unlock((mutex_t)(sstfr)->frame_lock);
		    continue;
		}

		set_sstfr_status(sstfr, EXECUTING);
                mutex_unlock((mutex_t)(sstfr)->frame_lock);
		return_pointer = sstfr;
		break;
	    }
            mutex_unlock((mutex_t)(sstfr)->frame_lock);
	} else {
	        thr_data[thead] = thr_data[ttail] = thr_data[tbase];
		thr_data[top] = thr_data[base];
		break;
	}

    }
    mutex_unlock((mutex_t)thr_data[LOCK_STACK]);

#ifdef DEBUG_MUTEX
	printf("%d) unlock in pop_sst_frame()\n", cthread_self());
#endif

    return return_pointer;
}



child_list_t * add_child_list(child_list_t * root, unsigned int * child)
{
    child_list_t *tmp;

    tmp = (child_list_t *)malloc(sizeof(child_list_t));
    printf("\n%x\n",tmp);

    if(tmp == NULL){
        bokkan("memory allocation error at add_child_list(): "
               "Can't allocate memory.");
    } else {
        tmp->next = root;
        tmp->child = child;
        return tmp; 
    }
}

/* attention to cut_connections_with */

void free_child_list(child_list_t *point)
{
    child_list_t *tmp;
    for (; point != NULL;) {
        tmp = point;
        point = point->next;
        free(tmp);
    }
}

sst_frame_t * steal_sst_frame(void)
{  
    sst_frame_t *tmp;

    if(thr_data[THR_STATUS] == PRIMARY){
        tmp = steal_sst_frame_aux(MAN_MAN_NO_COND, 1);
	if(tmp == NULL)
	  tmp = steal_sst_frame_aux(MAN_SUB_MAN, 1);
    } else {
        tmp = steal_sst_frame_aux(SUB_MAN_COND, 0);
        if(tmp == NULL){
	    tmp = steal_sst_frame_aux(SUB_SUB_NO_COND, 1);
            if(tmp == NULL)
                tmp = steal_sst_frame_aux(SUB_SUB_COND, 1);
        }
    }
    return tmp;
}

#define KILL_DEBUG
#ifdef KILL_DEBUG

unsigned int *de_stack;
unsigned int *de_stack2;
unsigned int *de_st;
sst_frame_t  *de_frame;
child_list_t *de_child;

void set_debug(int thr_num)
{
   de_stack = sthrs[thr_num]->stack;
}


#endif

sst_frame_t * get_sst_frame_aux(sst_frame_t *frame, unsigned int fut)
{
    mutex_lock(frame->frame_lock);
    if (fut) {
        if (*(unsigned int *)(frame->pstack) == NULL ||
            ((ENTITYP)fut)->future.sstfr == NULL ||
            sstfr_status(frame) == EXECUTING) {
            mutex_unlock(frame->frame_lock);
            return NULL;
        }
    } else if (*(unsigned int *)(frame->pstack) == NULL ||
               (sstfr_status(frame) == DETERMINED) ||
               (sstfr_status(frame) == EXECUTING)){
        mutex_unlock(frame->frame_lock);
        return NULL;
    }

    *((unsigned int *)frame->pstack) = NULL;
    set_sstfr_status(frame,EXECUTING);
    mutex_unlock(frame->frame_lock);

    return frame;
}


sst_frame_t * steal_sst_frame_aux(int mode, int same_class_p)
{
    int j, proc_num;
    int tbase,thead,ttail;

    unsigned int * rstk;
    sst_frame_t ** sst_tail;
    sst_frame_t  * return_pointer = NULL;
    sst_frame_t *sstfr;
    thread_data_t ** thrs;

    switch(mode){
    case MAN_MAN_NO_COND  :     /* mandatory -> mandatory no condition */
        thrs  = pthrs;
        proc_num = processors;
        tbase = PT_BAS_MAN;
        thead = PT_HED_MAN;
        ttail = PT_TAL_MAN;
        break;
    case MAN_SUB_MAN      :    /* mandatory  -> subsidiary_m no condition */
        thrs  = sthrs;
        proc_num = sprocessors; 
        tbase = PT_BAS_MAN;
        thead = PT_HED_MAN;
        ttail = PT_TAL_MAN;
        break;
    case SUB_MAN_COND     :    /* subsidiary -> mandatory condition */
        thrs  = pthrs;
        proc_num = processors;
        tbase = PT_BAS_SPE;
        thead = PT_HED_SPE;
        ttail = PT_TAL_SPE;
        break;
    case SUB_SUB_NO_COND  :   /* subsidiary -> subsidiary no condition */
        thrs  = sthrs;
        proc_num = sprocessors; 
        tbase = PT_BAS_MAN;
        thead = PT_HED_MAN;
        ttail = PT_TAL_MAN;
        break;
    case SUB_SUB_COND     :   /* subsidiary -> subsidiary condition */
        thrs  = sthrs;
        proc_num = sprocessors; 
        tbase = PT_BAS_SPE;
        thead = PT_HED_SPE;
        ttail = PT_TAL_SPE;
        break;
    default   : 
        bokkan("Unknown steal mode error at steal_sst_frame_aux(): ");
        break;
    }

    for (j = 1; j <= proc_num - same_class_p; j++) {

        rstk = thrs[(thr_data[THREAD_ID] + j) % proc_num]->stack;

	if(rstk[THR_KILLED] == KILL_BY_PAR) continue;

        if (mode == MAN_SUB_MAN && rstk[THR_STATUS] != SUBSIDIARY_M)
            continue;
        else if (mode == SUB_SUB_NO_COND && rstk[THR_STATUS] == SUBSIDIARY_M)
            continue;

        if (rstk[thead] > rstk[ttail]) {
#ifdef DEBUG_MUTEX
            printf("%d) try lock ...\n", cthread_self());
#endif
            if (! mutex_try_lock((mutex_t)rstk[LOCK_STACK])) {
#ifdef DEBUG_MUTEX
                printf("%d) fail trylock\n", cthread_self());
#endif
                continue;
            }
            while (rstk[thead] > rstk[ttail]) {
                ((sst_frame_t **)rstk[ttail])++;
#ifdef RMO
# if defined(sparc)
                asm("membar #StoreLoad"); /* for V9 */
# else
                mutex_lock(dummy);
                mutex_unlock(dummy);
# endif
#endif
                if (rstk[thead] >= rstk[ttail]) {
                    sst_tail = (sst_frame_t **)((sst_frame_t **)rstk[ttail] - 1);
                    sstfr = *sst_tail;
                    if (sstfr == NULL) continue;

		    mutex_lock((mutex_t)sstfr->frame_lock);

                    if (*sst_tail == NULL) {
                        mutex_unlock((mutex_t)sstfr->frame_lock);
                        continue;
                    }

                    if (killed_sstfr_p(sstfr) == NOT_NEEDED) {
		      /* free_sst_frame(*sst_tail); */
                        mutex_unlock((mutex_t)sstfr->frame_lock);
                        continue;
                    }

		    if ((sstfr_status(sstfr) == DETERMINED) ||
			(sstfr_status(sstfr) == EXECUTING)) {
                        mutex_unlock((mutex_t)sstfr->frame_lock);
                        continue;
                    }

		    set_sstfr_status(sstfr, EXECUTING);
		    mutex_unlock((mutex_t)sstfr->frame_lock);
		    return_pointer = sstfr;
		    thr_data[PT_2_FRAME] = (unsigned int)sstfr;
		    break;
                    
                } else {
                    ((sst_frame_t **)rstk[ttail])--;
#ifdef RMO
# if defined(sparc)
                    asm("membar #StoreLoad"); /* for V9 */
# else
                    mutex_lock(dummy);
                    mutex_unlock(dummy);
# endif
#endif
                    break;
                }
            } /* while */

            if (return_pointer != NULL) {
#ifdef DEBUG_MUTEX
                printf("%d) lock thr_data[] in steal_sst_frame_aux()...\n", cthread_self());
#endif

                mutex_lock((mutex_t)thr_data[LOCK_STACK]);
                if (rstk[THR_STATUS] == SUBSIDIARY_S) {

                    thr_data[PT_2_PAREN] = (unsigned int)rstk; /* set speculative parnt */
	            {
                       child_list_t *tmp = (child_list_t *)malloc(sizeof(child_list_t));

		       if(tmp == NULL){
			 bokkan("memory allocation error at add_child_list(): "
				"Can't allocate memory.");
		       } 
		       tmp->next = (child_list_t *)rstk[CHILD_LIST];
		       tmp->child = (unsigned int *)thr_data;
		       rstk[CHILD_LIST] = (unsigned int) tmp;
                    }
                }

#ifdef DEBUG_MUTEX
                printf("%d) unlock thr_data[] in steal_sst_frame_aux()\n", cthread_self());
#endif
                mutex_unlock((mutex_t)thr_data[LOCK_STACK]);
/*                mutex_unlock((mutex_t)rstk[tlock]); *//* unlock other task */
#ifdef DEBUG_MUTEX
                printf("%d) unlock in steal_sst_frame_aux()\n", cthread_self());
#endif
                mutex_unlock((mutex_t)rstk[LOCK_STACK]);
#ifdef SHE_ENV
                if (tbase == PT_BAS_SPE) /* stolen frame has a condition */
                    copy_env_to_con_stack(return_pointer);
#endif
                break;
            }
/*            mutex_unlock((mutex_t)rstk[tlock]); *//* unlock other task */
#ifdef DEBUG_MUTEX
            printf("%d) unlock in steal_sst_frame_aux()\n", cthread_self());
#endif
            mutex_unlock((mutex_t)rstk[LOCK_STACK]);

        }
    } /* for j */

    return return_pointer;
}

int Sprint_stat()
{
    print_stat(getarg(1) == LTAG_FALSE ? sthrs : pthrs, fix_to_int(getarg(2)));
    set_reg(LTAG_T);
}

int print_stat_of_sst_c(unsigned int *head, unsigned int *tail)
{
    int killed, gotten;

    killed = gotten = 0;
    while (head > tail) {
        if (*tail == NULL)
            gotten++;
        else if (killed_sstfr_p((sst_frame_t *)*tail))
            killed++;
        tail++;
    }
    printf("    killed/gotten = %d/%d\n", killed, gotten);
    return 0;
}

int print_stat(thread_data_t ** thrs, int thr_num)
{
    printf("pthread's ID = %d\n", cthread_self());
    printf("current threads' thr_data = %x\n", thr_data);
    printf("%s[%d]: ", thrs == pthrs ? "pthrs" : "sthrs", thr_num);
    printf("PT_2_PAREN = %x, ", thrs[thr_num]->stack[PT_2_PAREN]);
    printf("CHILD_LIST = %x, ", thrs[thr_num]->stack[CHILD_LIST]);
    printf("PT_2_FRAME = %x\n", thrs[thr_num]->stack[PT_2_FRAME]);
    printf("PT_BAS_MAN = %x, ", thrs[thr_num]->stack[PT_BAS_MAN]);
    printf("PT_HED_MAN = %x, ", thrs[thr_num]->stack[PT_HED_MAN]);
    printf("PT_TAL_MAN = %x, ", thrs[thr_num]->stack[PT_TAL_MAN]);
    printf("Task Man = %d\n",  (thrs[thr_num]->stack[PT_HED_MAN] -
				thrs[thr_num]->stack[PT_TAL_MAN]));
    printf("PT_BAS_SPE = %x, ", thrs[thr_num]->stack[PT_BAS_SPE]);
    printf("PT_HED_SPE = %x, ", thrs[thr_num]->stack[PT_HED_SPE]);
    printf("PT_TAL_SPE = %x, ", thrs[thr_num]->stack[PT_TAL_SPE]);
    printf("Task Spe = %d\n",  (thrs[thr_num]->stack[PT_HED_SPE] -
				thrs[thr_num]->stack[PT_TAL_SPE]));
    print_stat_of_sst_c(thrs[thr_num]->stack[PT_HED_SPE],
                        thrs[thr_num]->stack[PT_TAL_SPE]);
    printf("PT_BSEV_NC = %x, ", thrs[thr_num]->stack[PT_BSEV_NC]);
    printf("PT_TPEV_NC = %x, ", thrs[thr_num]->stack[PT_TPEV_NC]);
    printf("Gomi in NC = %d\n", thrs[thr_num]->stack[PT_TPEV_NC] -
	                        thrs[thr_num]->stack[PT_BSEV_NC]);
    printf("PT_BSEV_CO = %x, ", thrs[thr_num]->stack[PT_BSEV_CO]);
    printf("PT_TPEV_CO = %x, ", thrs[thr_num]->stack[PT_TPEV_CO]); 
    printf("Gomi in CO = %d\n", thrs[thr_num]->stack[PT_TPEV_CO] -
	                        thrs[thr_num]->stack[PT_BSEV_CO]);
    printf("PT_2_HEAD  = %x, ", thrs[thr_num]->stack[PT_2_HEAD]); 
    printf("PT_2_TAIL  = %x\n", thrs[thr_num]->stack[PT_2_TAIL]); 
    printf("processor  = %d, ", processors); 
    printf("sprocessor = %d\n", sprocessors); 
    printf("primary    = %d, ", pri_act.active); 
    printf("subdidiary = %d, ", sub_act.active); 
    printf("suspended  = %d\n", suspend_act.active); 

    switch ((int)thrs[thr_num]->stack[THR_STATUS]) {
    case PRIMARY:
        printf("PRIMARY, ");
        break;
    case SUBSIDIARY_M:
        printf("SUBSIDIARY_M, ");
        break;
    case SUBSIDIARY_S:
        printf("SUBSIDIARY_S, ");
        break;
    }
    printf("THREAD_ID = %x, ",  thrs[thr_num]->stack[THREAD_ID]);
    printf("killed? = %x, ", thrs[thr_num]->stack[THR_KILLED] == KILL_BY_PAR);
    printf("cond_exec? = %x\n",  thrs[thr_num]->stack[COND_EXEC] == YES);
    fflush(stdout);
}

#ifdef SHE_ENV

void push_env(unsigned int env,int mode)
{
    int top;
    int base;

    switch (mode) {
    case NO_COND :  
        base = PT_BSEV_NC;
	top  = PT_TPEV_NC; 
	break;
    case COND    :  
        base = PT_BSEV_CO;
	top  = PT_TPEV_CO; 
	break;
    default:
        bokkan("Error in push_env()\n");
    }

    if(((unsigned int *)(thr_data[top]) + 1)
       > ((unsigned int *)(thr_data[base]) + ENV_STACK_SIZE))
        bokkan("fatal error at push_env(): "
               "environment stack overflow \n");

    *(((unsigned int *)(thr_data[top]))++) = env;
}

void copy_env_to_cstack(sst_frame_t *sstfr)
{
    int lp;
    unsigned int * arg = sstfr->env_array - 1;

    for (lp = sstfr->narg; lp > 0; lp--) s_push(*arg--); 
}

void pop_env_to_cstack(sst_frame_t *sstfr)
{
    int lp;
    unsigned int * arg = sstfr->env_array - 1;

    for (lp = sstfr->narg; lp > 0; lp--) s_push(*arg--); 

    if (sstfr->env_array == thr_data[PT_TPEV_NC])
       (unsigned int *)(thr_data[PT_TPEV_NC]) -= sstfr->narg;
    else if (sstfr->env_array == thr_data[PT_TPEV_CO])
             (unsigned int *)(thr_data[PT_TPEV_CO]) -= sstfr->narg;
}

void copy_env_to_con_stack(sst_frame_t *sstfr)
{
    int lp;
    unsigned int * env = (unsigned int *)thr_data[PT_ENV_CO],
                 * arg = sstfr->env_array - sstfr->narg;

    for (lp = sstfr->narg; lp > 0; lp--) *env++ = *arg++;

    sstfr->env_array = env;
}

#endif

#endif
