/*
 * structure.c - 
 */

/*
 * $Id: structure.c,v 1.3 2014/10/28 02:22:55 komiya Exp $
 *
 *  ---------
 *  STRUCTURE
 *  ---------
 *  defstruct-type, structure?, structure-type,
 *  structure, structure-ref, structure-set!
 *
 *  File Name : structure.c
 *  
 *  Mikio Yamamoto. 1990.
 *
 */

static char *RCSid="$Id: structure.c,v 1.3 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"


int eq();

struct_size(sym)
int sym;
{
    SOBJ prop;
    prop.obj = assoc(eq, struct_sym.obj, 
        ((SYMBOL * )ADDRESS(sym))->proplist.obj);
    if (prop.obj == LTAG_FALSE)
        return(-1);
    else
        return(fix_to_int(CDRobj(prop.obj)));
}


/*
 *  scheme system function
 *
 */

void Sdefstruct_type()
{ /* (defstruct-type name size) ==> name */
    SOBJ name;
    SOBJ strusize;
    int old_size, new_size, tag;

#ifdef TOUCH_ARG
    name.obj = touch_obj(getarg(1));
#else
    name.obj = getarg(1);
#endif
    if (GETsym_tag(name.obj) != LTAG_SYMBOL){
        err_illegal_argument_type(name.obj, "symbol");
	return ;
    }

#ifdef TOUCH_ARG
    strusize.obj = touch_obj(getarg(2));
#else
    strusize.obj = getarg(2);
#endif
    if ((tag = GETtag(strusize.obj)) != LTAG_FIXNUM && tag != LTAG_FIXNUM1){
        err_illegal_argument_type(strusize.obj, "non-negative fixnum");
	return ;
    }

    old_size = struct_size(name.obj);
    if (old_size >= 0) {
        new_size = fix_to_int(strusize.obj);
        if (new_size != old_size){
            bokkan("Structure definition can't change to different size.");
	    return;
	}
        else {
            print_con("Warning: Structure ");
            printobj_con(name.obj, PRINT, 5);
            print_con(" is being redefined.\n");
#ifdef TOUCH_ARG
            set_reg(touch_obj(getarg(1)));
#else
            set_reg(getarg(1));
#endif
            return;
        }
    }
    reg1 = ((SYMBOL * )ADDRESS(name.obj))->proplist.obj;
    reg2 = mk_cons(struct_sym.obj, strusize.obj);
    reg1 = mk_cons(reg2, reg1);
#ifdef TOUCH_ARG
    ((SYMBOL * )ADDRESS(touch_obj(getarg(1))))->proplist.obj = reg1;
#else
    ((SYMBOL * )ADDRESS(getarg(1)))->proplist.obj = reg1;
#endif
    reg1 = LTAG_NIL;
    reg2 = LTAG_NIL;
#ifdef TOUCH_ARG
    set_reg(touch_obj(getarg(1)));
#else
    set_reg(getarg(1));
#endif
}




void Sstructp() 
{    /* (structure?  scheme-obj) ==> #f or #t */
#ifdef TOUCH_ARG
    if (GETtag(touch_obj(getarg(1))) == LTAG_STRUCTURE)
#else
    if (GETtag(getarg(1)) == LTAG_STRUCTURE)
#endif
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
}


void Sstruct_type()
{  /* (structure-type structure) ==> structure-type */
#ifdef TOUCH_ARG
    if (GETtag(touch_obj(getarg(1))) != LTAG_STRUCTURE) {
#else
    if (GETtag(getarg(1)) != LTAG_STRUCTURE){
#endif
        err_illegal_argument_type(getarg(1), "structure");
	return ;
    }
    else
#ifdef TOUCH_ARG
        set_reg( ((ENTITYP)ADDRESS(touch_obj(getarg(1))))->struc.type.obj );
#else
        set_reg( ((ENTITYP)ADDRESS(getarg(1)))->struc.type.obj );
#endif
}


void Sstructure() 
{   /* (structure name ele1 ele2 ...) ==> new-structure-obj */
    SOBJP result;
    int len, i;
    int size;
    int sym;

    len = getargnum();
#ifdef TOUCH_ARG
    sym = touch_obj(getarg(1));
#else
    sym = getarg(1);
#endif
    if (GETsym_tag(sym) != LTAG_SYMBOL){
        err_illegal_argument_type(sym, "symbol");
	return ;
    }
    size = struct_size(sym);
    if (size < 0){
        err_illegal_argument_type(sym, "structure name");
	return;
    }
    if ( (len - 1) != size ){
        bokkan("Structure length and arguments length is mis-matched.");
	return;
    }

    result = (SOBJP)mk_estructp(sym, len - 1);
    for (i = 2; i <= len; i++)
        (result + i)->obj = getarg(i);
    set_reg(LTAG_STRUCTURE | SSERDDA(result));
}


void Sstructref()
{  /* (structure-ref name structure nth) ==> nth-element */
    SOBJ obj1, obj2;
    int index, tag, size, sym;

#ifdef TOUCH_ARG
    sym = touch_obj(getarg(1));
#else
    sym = getarg(1);
#endif
    if (GETsym_tag(sym) != LTAG_SYMBOL){
        err_illegal_argument_type(sym, "symbol");
	return ;
    }
    size = struct_size(sym);
    if (size < 0){
        err_illegal_argument_type(sym, "structure name");
	return ;
    }

#ifdef TOUCH_ARG
    obj1.obj = touch_obj(getarg(2));
#else
    obj1.obj = getarg(2);
#endif
    if (GETtag(obj1.obj) != LTAG_STRUCTURE){
        err_illegal_argument_type(obj1.obj, "structure");
	return ;
    }

#ifdef TOUCH_ARG
    if ( ((ENTITYP)ADDRESS(obj1.obj))->struc.type.obj != touch_obj(getarg(1))){
#else
    if ( ((ENTITYP)ADDRESS(obj1.obj))->struc.type.obj != getarg(1) ){
#endif
        bokkan("different structure type.");
	return;
    }

#ifdef TOUCH_ARG
    obj2.obj = touch_obj(getarg(3));
#else
    obj2.obj = getarg(3);
#endif
    if ((tag = GETtag(obj2.obj)) != LTAG_FIXNUM && tag != LTAG_FIXNUM1){
        err_illegal_argument_type(obj2.obj, "non-negative fixnum");
	return ;
    }

    index = fix_to_int(obj2.obj);
    if (size <= index){
        err_illegal_argument("3rd argument is out of the structure length.");
	return ;
    }

    set_reg(*(((unsigned *)ADDRESS(obj1.obj)) + index + 2));
}


void Sstructset() 
{  /* (structure-set! name structure nth val) ==> val */
    SOBJ obj1, obj2;
    int index, tag, size, sym;

#ifdef TOUCH_ARG
    sym = touch_obj(getarg(1));
#else
    sym = getarg(1);
#endif
    if (GETsym_tag(sym) != LTAG_SYMBOL){
        err_illegal_argument_type(sym, "symbol");
	return ;
    }
    size = struct_size(sym);
    if (size < 0){
        err_illegal_argument_type(sym, "structure name");
	return ;
    }

#ifdef TOUCH_ARG
    obj1.obj = touch_obj(getarg(2));
#else
    obj1.obj = getarg(2);
#endif
    if (GETtag(obj1.obj) != LTAG_STRUCTURE){
        err_illegal_argument_type(obj1.obj, "structure");
	return ;
    }

#ifdef TOUCH_ARG
    if ( ((ENTITYP)ADDRESS(obj1.obj))->struc.type.obj != touch_obj(getarg(1)))
#else
    if ( ((ENTITYP)ADDRESS(obj1.obj))->struc.type.obj != getarg(1) )
#endif
        bokkan("different structure type.");

#ifdef TOUCH_ARG
    obj2.obj = touch_obj(getarg(3));
#else
    obj2.obj = getarg(3);
#endif
    if ((tag = GETtag(obj2.obj)) != LTAG_FIXNUM && tag != LTAG_FIXNUM1){
        err_illegal_argument_type(obj2.obj, "non-negative fixnum");
	return ;
    }

    index = fix_to_int(obj2.obj);
    if (size <= index){
        err_illegal_argument("3rd argument is out of the structure length.");
	return ;
    }

    *(((unsigned *)ADDRESS(obj1.obj)) + index + 2) = getarg(4);
    set_reg(getarg(4));
}
