/*
 * symbol.c - 
 */

/*
 * $Id: symbol.c,v 1.3 2014/10/28 02:22:55 komiya Exp $
 *
 *  ----------------
 *  SYMBOL OPERATION
 *  ----------------
 *
 *  File Name : symbol.c
 *  written by Mikio Yamamoto. 1990
 *  fair-copied by Yass.
 *
 */

static char *RCSid="$Id: symbol.c,v 1.3 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"

#ifdef CTHREAD
mutex_t addsymbol_lock;
#endif

/*
 *  hash functions
 *  basic symbol manipulation functions
 *
 */
int hash(str_obj)
int str_obj;
{
    char    *strp;
    int length;

    if (GETtag(((STRING * )ADDRESS(str_obj))->head) == LTAG_BLOCK1)
        str_obj = ((STRING * )ADDRESS(str_obj))->str_len;
    length = ((STRING * )ADDRESS(str_obj))->str_len;
    strp = (char *)(ADDRESS(str_obj) + 12);

#ifdef NIHONGO
    if (((STRING * )ADDRESS(str_obj))->str_kind == K_STRING) /* Kanji symbol */
	    return(k_hashing((short *)strp, length));
    else
	    return(a_hashing(strp, length));
#else
    if (((STRING * )ADDRESS(str_obj))->str_kind == K_STRING) /* Kanji symbol */
        length *= 2;
    return(hashing(strp, length));
#endif
}

#ifdef NIHONGO

int k_hashing(str, len)
short	*str;
int	len;
{
	int	hash_val = 0;

	for (; 0 < len; len--)
		hash_val += (int)*str++;

	return(hash_val & hash_bit_mask);
}

int a_hashing(str, len)
char	*str;
int	len;
{
	int	hash_val = 0;

	for (; 0 < len; len--)
		hash_val += (int)*str++;

	return(hash_val & hash_bit_mask);
}

#endif /* NIHONGO */

int hashing(str, len)
char    *str;
int len;
{ 
    int hash_val = 0,
    flag = TRUE;

    for (; 0 < len ; len--) { 
        if (flag) { 
            hash_val += (int)*(str++);
            flag = FALSE;
        } else    { 
            hash_val += ((int)*(str++)) << 8;
            flag = TRUE;
        }
    }
    return(hash_val & hash_bit_mask);
}


int addsymbol(str_obj, sym_kind)
int str_obj, sym_kind;
{
    SOBJ    tmp;
    SOBJP   obj, t_obj;
    ENTITYP body;

#ifdef CTHREAD
    mutex_lock(addsymbol_lock);
#endif
    t_obj = obj = (SOBJP)(hash_table + hash(str_obj));
    while (obj->obj != LTAG_NIL) { 
        body = (ENTITYP)ADDRESS(obj->obj);
        if (str_compare(str_obj, body->sym.sym_name.obj) == STR_EQUAL) {
#ifdef CTHREAD
            mutex_unlock(addsymbol_lock);
#endif
            return(obj->obj);
        } else
            obj = &(body->sym.hash_lk);
    }
    if (t_obj->obj == LTAG_NIL)
        t_obj->obj = mk_sym(str_obj, sym_kind);
    else {
        tmp.obj = mk_sym(str_obj, sym_kind);
        ((ENTITYP)ADDRESS(tmp.obj))->sym.hash_lk.obj = t_obj->obj;
        t_obj->obj = tmp.obj;
    }
#ifdef CTHREAD
    mutex_unlock(addsymbol_lock);
#endif
    return(t_obj->obj);
}


int sym_name(sym_obj)
int sym_obj;
{
    return(((ENTITYP)ADDRESS(sym_obj))->sym.sym_name.obj);
}


set_sval(sym_obj, value)
int sym_obj, value;
{
    ((ENTITYP)ADDRESS(sym_obj))->sym.sym_val.obj = value;
}


set_sval_with_check(sym_obj, value)
int sym_obj, value;
{
    if (GETtag(sym_obj) != LTAG_ORDINARY) {
	/* T.Komiya */
	print_con("Warning: ");
	printobj_con(sym_obj, PRINT, 5);
	print_con(" is being redefined.\n");
        ((ENTITYP)ADDRESS(sym_obj))->sym.sym_val.obj = value;
    } else if (((ENTITYP)ADDRESS(sym_obj))->sym.sym_val.obj
	       == unbound_obj.obj ) {
	err_unbound_symbol4(sym_obj);
	return NULL;
    } else
        ((ENTITYP)ADDRESS(sym_obj))->sym.sym_val.obj = value;
}


int get_sval(sym_obj)
int sym_obj;
{
    return(((ENTITYP)ADDRESS(sym_obj))->sym.sym_val.obj);
}


int cp_str_obj(str_obj)
int str_obj;
{ 
    ENTITYP new_obj;
    int b_len, s_len, tag, i;
    unsigned *str_body;

    str_body = (unsigned *)ADDRESS(str_obj);
    if (GETtag(*str_body) == LTAG_BLOCK1)
        str_body = (unsigned *)ADDRESS(*(str_body + 1));
    b_len = IMM(*str_body) + 4;
    s_len = *(str_body + 1);
    s_push(str_obj);
    s_push(LTAG_STRING | SSERDDA(str_body));
    new_obj = (ENTITYP)memget(b_len);
    str_body = (unsigned *)ADDRESS(s_pop());
    str_obj = s_pop();
    new_obj->str.head    = LTAG_BLOCK2 | (unsigned)(b_len - 4);
    new_obj->str.str_len = s_len;
    new_obj->str.str_kind = *(str_body + 2);
    /* Myama myama */
    /* bug! bug! */
    /*
  if((tag = GETtag(str_obj)) == LTAG_KSTRING)
    b_len = (s_len + 1) * 2;
  else 
*/

#ifdef NIHONGO
    if (new_obj->str.str_kind == K_STRING)
	b_len = (s_len + 1) * 2;
    else
	b_len = s_len + 1;
#else
    b_len = s_len + 1;
#endif

    for (i = 0 ; i < b_len ; i++)
        *(((char *)new_obj) + 12 + i) = *(((char *)str_body) + 12 + i);
    return(SSERDDA(new_obj) | LTAG_STRING);
}


/*
 *   --------------------------
 *   SCHEME FUNCTION for SYMBOL
 *   --------------------------
 *   symbol?
 *   symbol->string
 *   string->symbol
 *   getprop, putprop, symbol-plist
 *   getsval
 *   symbol-globalization
 *
 */


void Ssymbolp()
{
#ifdef TOUCH_ARG
    if (GETsym_tag(touch_obj(getarg(1))) == LTAG_SYMBOL)
#else
    if (GETsym_tag(getarg(1)) == LTAG_SYMBOL)
#endif
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
}


void Ssymbolstring()
{ 
    SOBJ symbol;

#ifdef TOUCH_ARG
    symbol.obj = touch_obj(getarg(1));
#else
    symbol.obj = getarg(1);
#endif
    if (GETsym_tag(symbol.obj) != LTAG_SYMBOL){
        err_illegal_argument_type(symbol.obj, "symbol");
	return;
    }
    set_reg(((ENTITYP)ADDRESS(symbol.obj))->sym.sym_name.obj);
}


void Sstringsymbol()
{ 
    SOBJ string;
    int tag;

#ifdef TOUCH_ARG
    string.obj = touch_obj(getarg(1));
#else
    string.obj = getarg(1);
#endif
    if (GETstr_tag(string.obj) != LTAG_STRING){
        err_illegal_argument_type(string.obj, "string");
	return;
    }

    if (GETtag(((STRING * )ADDRESS(string.obj))->head) == LTAG_BLOCK1) /* Added by Ushijima */
        string.obj = ((STRING * )ADDRESS(string.obj))->str_len;

    /* Myama myama */
    /*
  if((tag = GETtag(string.obj)) == LTAG_ASSTRING || tag == LTAG_KSSTRING)
*/
    if (GETtag(string.obj) == LTAG_SSTRING)
        set_reg(addsymbol(string.obj, LTAG_ORDINARY));
    else
        set_reg(addsymbol((cp_str_obj(string.obj) | SYMBOL_STRING), LTAG_ORDINARY));
}

void Sstringsymbol_no_intern()    /* by T.Komiya */
{ 
    SOBJ string;
    int tag;

#ifdef TOUCH_ARG
    string.obj = touch_obj(getarg(1));
#else
    string.obj = getarg(1);
#endif
    if (GETstr_tag(string.obj) != LTAG_STRING){
        err_illegal_argument_type(string.obj,"string");
	return;
    }
    if (GETtag(((STRING * )ADDRESS(string.obj))->head) == LTAG_BLOCK1) /* Added by Ushijima */
        string.obj = ((STRING * )ADDRESS(string.obj))->str_len;

    if (GETtag(string.obj) == LTAG_SSTRING)
    	set_reg(mk_sym(string.obj, LTAG_ORDINARY));
    else
    	set_reg(mk_sym((cp_str_obj(string.obj) | SYMBOL_STRING), LTAG_ORDINARY));
}

/*
 *   property list operation
 *
 */

int eq();

void Sgetprop()
{
    SOBJ element_pair;
    int sym;
#ifdef TOUCH_ARG
    sym = touch_obj(getarg(1));
#else
    sym = getarg(1);
#endif
    if (GETsym_tag(sym) != LTAG_SYMBOL){
        err_illegal_argument_type(sym, "symbol");
	return;
    }
    element_pair.obj = assoc(eq, getarg(2),
        ((SYMBOL * )ADDRESS(sym))->proplist.obj);
    if (element_pair.obj == LTAG_FALSE)
        set_reg(LTAG_FALSE);
    else 
        set_reg(CDRobj(element_pair.obj));
}


void Sputprop()
{
    SOBJ element_pair;
    int new_element;

#ifdef TOUCH_ARG
    reg1 = touch_obj(getarg(1)); /* reg1 & reg2 will be gabage-collected */
#else
    reg1 = getarg(1); /* reg1 & reg2 will be gabage-collected */
#endif
    if (GETsym_tag(reg1) != LTAG_SYMBOL){
        err_illegal_argument_type(reg1, "symbol");
	return;
    }
    element_pair.obj = assoc(eq, getarg(2),
        ((SYMBOL * )ADDRESS(reg1))->proplist.obj);
    if (element_pair.obj == LTAG_FALSE) {
        reg2 = mk_cons(getarg(2), getarg(3));
        reg2 = mk_cons(reg2, ((SYMBOL * )ADDRESS(reg1))->proplist.obj);
        ((SYMBOL * )ADDRESS(reg1))->proplist.obj = reg2;
        reg2 = LTAG_NIL;
    } else {
        set_cdr(element_pair.obj, getarg(3));
    }
}

void Ssymbol_plist()
{
    SOBJ element_pair;
    int sym;
#ifdef TOUCH_ARG
    sym = touch_obj(getarg(1));
#else
    sym = getarg(1);
#endif
    if (GETsym_tag(sym) != LTAG_SYMBOL){
        err_illegal_argument_type(sym, "symbol");
	return;
    }
    set_reg(((SYMBOL * )ADDRESS(sym))->proplist.obj);
}


/*  other */

void Sgetsval()
{
    SOBJ sym;
    SOBJ value;

#ifdef TOUCH_ARG
    sym.obj = touch_obj(getsv(-2));
#else
    sym.obj = getsv(-2);
#endif
    if (GETsym_tag(sym.obj) != LTAG_SYMBOL){
        err_illegal_argument_type(sym.obj, "symbol");
	return;
    }
    value.obj = get_sval(sym.obj);
    if (value.obj == unbound_obj.obj){
        err_unbound_symbol(sym.obj);
	return;
    }
    set_reg(value.obj);
}


void Ssymbol_globalization()
{
    SOBJ sym;
#ifdef TOUCH_ARG
    sym.obj = touch_obj(getsv(-2));
#else
    sym.obj = getsv(-2);
#endif
    if (GETtag(sym.obj) != LTAG_ORDINARY){
        err_not_ordinary_symbol(sym.obj);
	return;
    }
    set_sval(sym.obj, LTAG_NIL);
    set_reg(sym.obj);
}


void Sboundp()
{
    SOBJ sym;

#ifdef TOUCH_ARG
    sym.obj = touch_obj(getarg(1));
#else
    sym.obj = getarg(1);
#endif
    if (GETsym_tag(sym.obj) != LTAG_SYMBOL){
        err_illegal_argument_type(sym.obj, "symbol");
	return;
    }
    if (get_sval(sym.obj) == unbound_obj.obj)
        set_reg(LTAG_FALSE);
    else
        set_reg(LTAG_T);
}


void Sunboundp()
{
    SOBJ sym;

#ifdef TOUCH_ARG
    sym.obj = touch_obj(getarg(1));
#else
    sym.obj = getarg(1);
#endif
    if (GETsym_tag(sym.obj) != LTAG_SYMBOL){
        err_illegal_argument_type(sym.obj, "symbol");
	return;
    }
    if (get_sval(sym.obj) == unbound_obj.obj)
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
}


void Ssysfunp()
{
    SOBJ sym, val;

#ifdef TOUCH_ARG
    sym.obj = touch_obj(getarg(1));
#else
    sym.obj = getarg(1);
#endif
    if (GETsym_tag(sym.obj) != LTAG_SYMBOL){
        err_illegal_argument_type(sym.obj, "symbol");
	return;
    }
    if ((val.obj = get_sval(sym.obj)) == unbound_obj.obj)
        set_reg(LTAG_FALSE);
    else if ((GETtag(val.obj) == LTAG_SFUNCTION)
             || (GETtag(val.obj) == LTAG_MSFUNCTION))
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
}

void Ssymbol_sensitive()
{
    extern int symbol_sensitive;
    int n = symbol_sensitive;

    if (getarg(1) == LTAG_FALSE)
        symbol_sensitive = 0;
    else
        symbol_sensitive = 1;
    
    set_reg(n = 0 ? LTAG_FALSE : LTAG_T);
}
