;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;    
;;;    debuger and tracer
;;;         filename : debugfun.lsp
;;;         author : M.Yamamoto
;;;	    date : 1990. 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    debuger
;;;

(define (debuger) #f)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    gc
;;;

(define *gc-message-level* 0) 	;;; 0: no message
			      	;;; 1: normal message
			      	;;; 2: for gc debug

(define *msgc2copygc* #t)     	;;; #f: forbid to change mode of msgc to copygc.
				;;; other: allow to change.

(define aaa (protect-symbol '*gc-message-level*))
(define aaa (protect-symbol '*msgc2copygc*))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    tracer
;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;      global variable
;;;        (hidden or visible)
;;;

(define *debug-level* 1)  ;;; assign error handling procedure.(visible)
			  ;;; 0: nothing is happened at error. 
			  ;;; 1: back-trace is happened at error.
			  ;;; 2: debuger is invoked at error.
(define *trace-level* 0)  ;;; clear by topfun.(hidden) 
(define *max-back-trace-depth* 30)
			  ;;; max-level of back-trace at error(visible) 
			  ;;; -1: infinite depth.
(define *max-trace-depth* 0)
			  ;;; max-level of trace at execution.(visible)
			  ;;; max-level < 0 or = no-number: no cut trace.
(define *trace-indent-length* 2)
(define *trace-indent-rate* 1)
(define *max-trace-indent-length* 50)
(define *trace-level-bar-rate* 3)
(define *trace-level-bar* "|") 

(define aaa (protect-symbol '*debug-level*))
(define aaa (protect-symbol '*max-back-trace-depth*))
(define aaa (protect-symbol '*max-trace-depth*))
(define aaa (protect-symbol '*trace-indent-length*))
(define aaa (protect-symbol '*trace-indent-rate*))
(define aaa (protect-symbol '*max-trace-indent-length*))
(define aaa (protect-symbol '*trace-level-bar*))
(define aaa (protect-symbol '*trace-level-bar-rate*))

;;;(define aaa (protect-symbol '**))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  prototype of a trace function
;;;

(define (bc-trace-proc)
  (let ((target nil)(prologue nil)(epilogue nil))
    (lambda args
      (set! *trace-level* (1+ *trace-level*))
      (let ((prologue-result (user-tracer-apply prologue
						`(,*trace-level* ,target ,@args))))
        (case prologue-result
	  ((#t) (let ((result (user-function-apply target args)))
	          (set! result (user-tracer-apply
				 epilogue
				 `(,*trace-level* #t ,result nil ,target ,@args)))
		  (set! *trace-level* (1- *trace-level*))
		  result))
	  ((#f) (set! *trace-level* (1- *trace-level*))
		(user-function-apply target args))
	  (else (let
		  ((result 
		   (user-tracer-apply
		     epilogue
		     `(,*trace-level* #f nil ,prologue-result ,target ,@args))))
		  (set! *trace-level* (1- *trace-level*))
		  result)))))))

(define bc-trace-proc (bc-trace-proc))

(define (user-function-apply proc args)
  (let ((result (apply proc args)))
    (block-tail-recursive-call)
    result))

(define (user-tracer-apply proc args)
  (let ((result (apply proc args)))
    (block-tail-recursive-call)
    result))

(define (block-tail-recursive-call) nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;    
;;;   prototype of prologue&epilogue for normal trace function
;;;    (hidden function)
;;; 

(define (prologue depth fun . args)
  (indent depth)
  (display depth *stdout*)
  (display ">" *stdout*)
  (display "(" *stdout*)
  (display (let ((x (get-procedure-name fun)))
                (if (null? x) "#<function>" x))
           *stdout*)
  (dolist (x args) (display " " *stdout*) (write x *stdout*))
  (display ")" *stdout*)
  (newline *stdout*)
  #t)

(define (epilogue depth flag result prologue-result fun .  args)
  (indent depth)
  (display depth *stdout*)
  (display "<" *stdout*)
  (display "(" *stdout*)
  (display (let ((x (get-procedure-name fun)))
                (if (null? x) "#<function>" x))
           *stdout*)
  (display " " *stdout*)
  (write result *stdout*)
  (display ")" *stdout*)
;  (display " " *stdout*)
  (newline *stdout*)
  result)

(define (indent depth)
  (let ((length (* (quotient (1- depth) *trace-indent-rate*)
				   *trace-indent-length*)))
    (if (> length *max-trace-indent-length*)
	(write-space *max-trace-indent-length*)
	(write-space length))))

(define (write-space num)
  (if (number? *trace-level-bar-rate*)
      (let* ((interval (* *trace-level-bar-rate* *trace-indent-length*))
	     (interval-1indent (- interval *trace-indent-length*)))
        (do ((cu 1 (1+ cu)))
            ((> cu num))
            (if (eqv? (remainder cu interval) interval-1indent)
    	        (display *trace-level-bar* *stdout*)
	        (display " " *stdout*))))
      (do ((cu 1 (1+ cu)))
	  ((> cu num))
          (display " " *stdout*))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;    
;;;    trace & untrace function
;;;    
;;;    

;(define (trace proc)
;  (set-trace-function proc prologue epilogue))
;(define (untrace proc)
;  (detouch-trace-function proc))

(define _trace '())
(define _untrace '())

(let ((flist '()))
  (define (remove f)
    (define (aux curr prev)
      (cond ((null? curr) #f)
	    ((eq? (car curr) f) (set-cdr! prev (cdr curr)))
	    (else (aux (cdr curr) curr))))
    (cond ((null? flist) #f)
	  ((eq? (car flist) f) (set! flist (cdr flist)))
	  (else (aux flist (cdr flist)))))
  (set! _trace
	(lambda (proc)
	  (if (not (memq proc flist))
	      (begin
	        (set-trace-function proc prologue epilogue)
	        (set! flist (cons proc flist))))
	  #t))
  (set! _untrace
	(lambda proc
	  (for-each
	    (lambda (x)
	      (remove x)
	      (detouch-trace-function x))
	    (if (null? proc) flist proc))
	  #t)))

(define (trace proc) (_trace proc))
(define (untrace . proc) (apply _untrace proc))
